program ex1_hous2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of HOUS2 and APPLY_HOUS2
!   in module Hous_Procedures. 
!                                                                              
! LATEST REVISION : 18/03/2022
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, half, safmin, machhuge, &
                         hous2, apply_hous2,  norm, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE SIZE OF THE LINEAR SYSTEM TO BE SOLVED.
!
    integer(i4b), parameter :: prtunit=6, n=1000
!
    character(len=*), parameter :: name_proc='Example 1 of hous2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: err, eps, tau, tmp, d1, normx, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a
    real(stnd), dimension(:),  allocatable  :: b, x, res
!
    integer(i4b) :: i
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : SOLVING A LINEAR SYSTEM WITH A REAL n-BY-n MATRIX
!               AND ONE RIGHT HAND-SIDE WITH THE QR DECOMPOSITION.
!               THE QR DECOMPOSITION IS COMPUTED BY THE HOUSEHOLDER
!               METHOD.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = real( n, stnd)*sqrt( epsilon( err ) )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), b(n), x(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-BY-n RANDOM DATA MATRIX a .
!
    call random_number( a )
!
    a = a - half
!
    call random_number( tmp )
!
    if ( tmp>safmin ) then
        a = a/tmp
    end if
!
!   GENERATE A n-ELEMENT RANDOM SOLUTION VECTOR x .
!
    call random_number( x )
!
!   COMPUTE THE MATRIX-VECTOR PRODUCT b = a*x .
!
    b = matmul( a, x )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE SOLUTION VECTOR FOR LINEAR SYSTEM
!
!         a(:n,:n)*x(:n) = b(:n) .
!    
!   TRANSFORM THE MATRIX a TO UPPER TRIANGULAR FORM BY APPLYING 
!   A SERIE OF HOUSEHOLDER REFLECTIONS ON THE COLUMNS OF a AND APPLY
!   THE TRANSFORMATIONS TO b .
!
    d1 = machhuge
!
    do i = 1_i4b, n
!
        call hous2( a(i,i), a(i+1_i4b:n,i), tau )
        call apply_hous2( a(i+1_i4b:n,i), tau, a(i,i+1_i4b:n), a(i+1_i4b:n,i+1_i4b:n), left=true )
        call apply_hous2( a(i+1_i4b:n,i), tau, b(i), b(i+1_i4b:n) )
!
        d1 = min( d1, abs(a(i,i)) )
!
    end do
!    
    if ( d1>safmin ) then
!    
!       SOLVE THE n-BY-n RESULTING UPPER TRIANGULAR SYSTEM IF THE LINEAR SYSTEM
!       IS NOT SINGULAR.
!    
        do i = n, 1_i4b, -1_i4b
!    
            b(i) = b(i)/a(i,i)
            b(1_i4b:i-1_i4b) = b(1_i4b:i-1_i4b) - b(i)*a(1_i4b:i-1_i4b,i)
!    
        end do
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( d1>safmin .and. do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( res(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK THE RESULTS FOR SMALL RESIDUALS.
!
        res(:n) = b(:n) - x(:n)
        err     = norm(res)/norm(x)
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, b, x, res )
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, b, x )
!
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. d1>safmin ) then
!
        write (prtunit,*) name_proc//' is correct'
!
        if ( do_test ) then
!
            write (prtunit,*) 
!
!           PRINT RELATIVE ERROR OF COMPUTED SOLUTION.
!
            write (prtunit,*) 'Relative error of the computed solution = ', err
!
        end if
!
    else
!
        write (prtunit,*) name_proc//' is incorrect'
!
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the solution of a linear real system of size ', &
       n, ' by ', n, ' is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_hous2
! ========================
!
end program ex1_hous2
