program ex1_lapack_gesvdx
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of the driver subroutine GESVDX
!   in LAPACK software. 
!                                                                              
! LATEST REVISION : 12/01/2022
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : gesvdx
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, zero, one, seven, c30, c50, c1_e6,    &
                         norm, unit_matrix, random_seed_, random_number_, gen_random_mat, merror, &
                         allocate_error
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE GENERATED MATRIX,
! nsvd0 IS THE NUMERICAL RANK OF THE GENERATED MATRIX (WHICH MUST BE LESS OR EQUAL TO k) FOR CASES GREATER THAN 0,
! nsing IS THE TARGET RANK OF THE APPROXIMATE PARTIAL SVD, WHICH IS SOUGHT.
!
    integer(i4b), parameter :: prtunit = 6, m=3000, n=3000, k=min(m,n), nsvd0=3000, nsing=3000
!   
! conda IS THE CONDITION NUMBER OF THE GENERATED MATRIX IN CASES 5, 6 and 7.
!   
    real(stnd), parameter  :: fudge=c50, conda=c1_e6
!
    character(len=*), parameter :: name_proc='Example 1 of gesvdx'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, tmp, tmp2, anorm, &
                                               elapsed_time, vl, vu
    real(stnd)                              :: work2(1)
    real(stnd), allocatable, dimension(:)   :: s, work
    real(stnd), allocatable, dimension(:,:) :: a, a2, u, vt
!
    integer                            :: info, lwork, il, iu, ns, iok,     &
                                          istart, iend, irate, imax, itime
    integer, allocatable, dimension(:) :: iwork
    integer(i4b)                       :: i, mat_type
!
    logical(lgl)   :: do_test
!
    character      :: jobu, jobvt, range
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : FULL OR PARTIAL SVD OF A REAL MATRIX USING AN ASSOCIATED EIGENVALUE PROBLEM
!               AND THE BISECTION AND INVERSE ITERATION METHODS.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type < 1  -> RANDOM MATRIX FROM THE UNIFORM DISTRIBUTION
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type = 4  -> VERY SLOW DECAY OF SINGULAR VALUES
!   mat_type = 5  -> STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda
!                    SPECTRUM WITH FEW DEFLATIONS FOR LAPACK GESDD ROUTINE
!   mat_type = 6  -> GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH MODERATE DEFLATION FOR LAPACK GESDD ROUTINE
!   mat_type = 7  -> ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK GESDD ROUTINE
!   mat_type > 7  -> UNIFORM DISTRIBUTION OF SINGULAR VALUES
!
    mat_type = 2_i4b
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS MUST BE PERFORMED.
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), u(m,k), vt(k,n), s(k), iwork(12*k), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE A RANDOM UNIFORM MATRIX OR SINGULAR VALUES.
!
    select case( mat_type )
!
        case( :0_i4b )
!
!           RANDOM UNIFORM MATRIX.
!
            call random_number_( a )
!
!           COMPUTE THE FROBENIUS NORM OF THE RANDOM MATRIX.
!
            anorm = norm( a )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                s(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                s(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                s(i) = 0.0001_stnd + one/( one + exp( tmp - c30 ) )
!        
            end do
!
        case( 4_i4b )
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            tmp = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                s(i) = exp( -tmp2/tmp )
!        
            end do
!
        case( 5_i4b )
!
!           STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH FEW DEFLATIONS FOR LAPACK GESDD ROUTINE.
!
            s(:nsvd0-1_i4b) = one
            s(nsvd0) = one/conda
!
        case( 6_i4b )
!
!           GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH MODERATE DEFLATION FOR LAPACK GESDD ROUTINE.
!
            tmp  = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                s(i) = conda**( -tmp2/tmp )
!        
            end do
!
        case( 7_i4b )
!
!           ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK GESDD ROUTINE.
!
            tmp  = real( nsvd0 - 1_i4b, stnd )
            tmp2 = one - one/conda
!
            do i = 1_i4b, nsvd0
!
                s(i) = one - (real( i - 1_i4b, stnd )/tmp)*tmp2
!        
            end do
!
        case default
!
!           UNIFORM DISTRIBUTION OF SINGULAR VALUES.
!
            call random_number_( s(:nsvd0) )
!
    end select
!
    if ( mat_type>0_i4b ) then
!
#ifdef _F2003
        if ( ieee_support_datatype( s ) ) then
!
            if ( .not.all( ieee_is_normal( s(:nsvd0) ) ) ) then
!
                call merror( name_proc//' : Exceptions occurred when generating the input matrix !'  )                    
!
            end if
!
        end if
#endif
!
!       GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A SPECIFIED DISTRIBUTION
!       OF SINGULAR VALUES AND A RANK EQUAL TO nsvd0.
!
        call gen_random_mat( s(:nsvd0), a )
!
!       COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
        anorm = norm( s(:nsvd0) )
!
    end if
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( a2(m,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX .
!
        a2(:m,:n) = a(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   DETERMINE THE WORK TO DO AND HOW MANY SINGULAR TRIPLETS
!   WILL BE COMPUTED AND STORED.
!
!   IF jobu  = 'V', LEFT SINGULAR VECTORS WILL BE COMPUTED.
!   IF jobu  = 'N', LEFT SINGULAR VECTORS WILL NOT BE COMPUTED.
!   IF jobvt = 'V', RIGHT SINGULAR VECTORS WILL BE COMPUTED.
!   IF jobvt = 'N', RIGHT SINGULAR VECTORS WILL NOT BE COMPUTED.
!
    jobu  = 'V'
    jobvt = 'V'
!
!   range DETERMINED HOW MANY SINGULAR VALUES AND VECTORS WILL BE COMPUTED.
!
!   IF range='A' ALL SINGULAR VALUES WILL BE FOUND,
!   IF range='V' ALL SINGULAR VALUES IN THE HALF OPEN INTERVAL (vl,vu] WILL BE FOUND,
!   IF range='I' THE il-TH THROUGH iu-TH SINGULAR VALUES IN ASCENDING ORDER WILL BE FOUND.
!
!   IN EACH CASE, THE ASSOCIATED SINGULAR VECTORS WILL BE ALSO COMPUTED.
!
!    range = 'A'
    range = 'I'
!
    vl = zero
    vu = zero
!
    il = k - nsing + 1
    iu = k
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR gesvdx SUBROUTINE.
!
    lwork = -1
!
    call gesvdx( jobu, jobvt, range, m, n, a, m, vl, vu, il, iu, ns,   &
                 s, u, m, vt, k, work2, lwork, iwork, info )
!
    if ( info==0 ) then
!
        lwork = int(work2(1))
!
!       ALLOCATE OPTIMAL WORK VARIABLE NEEDED BY gesvdx SUBROUTINE.
!
        allocate( work(lwork), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       gesvdx COMPUTES THE SINGULAR VALUE DECOMPOSITION (SVD) OF A REAL
!       m-BY-n MATRIX a BY SOLVING AN ASSOCIATE EIGENVALUE PROBLEM. THE SVD
!       IS WRITTEN
!
!                       a = U * S * V**(t)
!
!       WHERE S IS AN m-BY-n MATRIX WHICH IS ZERO EXCEPT FOR ITS
!       min(m,n) DIAGONAL ELEMENTS, U IS AN m-BY-m ORTHOGONAL MATRIX, AND
!       V IS AN n-BY-n ORTHOGONAL MATRIX.  THE DIAGONAL ELEMENTS OF S
!       ARE THE SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!       THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
!       THE ADVANTAGE OF gesvdx SUBROUTINE COMPARED TO OTHER SVD SUBROUTINES
!       AVAILABLE IN LAPACK (E.G. gesvd OR gesdd) IS THAT gesvdx IS ABLE TO COMPUTE
!       A PARTIAL SVD DECOMPOSITION OR SELECTED SINGULAR TRIPLETS OF THE INPUT
!       REAL MATRIX WHILE THE OTHER SUBROUTINES DO NOT OFFER THIS POSSIBILITY.
!       HOWEVER, BEWARE THAT THE gesvdx SUBROUTINE IS AVAILABLE ONLY IN LAPACK 3.6.0
!       AND ABOVE.
!
        call gesvdx( jobu, jobvt, range, m, n, a, m, vl, vu, il, iu, ns,   &
                     s, u, m, vt, k, work, lwork, iwork, info )
!
!       THE ROUTINE RETURNS THE SINGULAR VALUES, THE LEFT AND RIGHT
!       SINGULAR VECTORS. THE RIGHT SINGULAR VECTORS ARE RETURNED ROWWISE.
!
!       ON EXIT OF gesvdx :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT.
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE.
!         info=i  > 0 :  INDICATES THAT i EIVENVECTORS FAILED TO CONVERGE.
!         info=n*2+1  :  INDICATES THAT AN INTERNAL ERROR OCCURRED.
!
!       ON EXIT OF gesvdx WITH THE VALUES SPECIFIED ABOVE FOR jobu ('V'),
!       jobvt ('V') AND range ('A'):
!
!         u IS OVERWRITTEN WITH THE FIRST  min(m,n)
!         COLUMNS OF U (THE LEFT SINGULAR VECTORS,
!         STORED COLUMNWISE);
!
!         vt IS OVERWRITTEN WITH THE FIRST min(m,n)
!         ROWS OF V**(t) (THE RIGHT SINGULAR VECTORS,
!         STORED ROWWISE);
!
!         s IS OVERWRITTEN WITH THE SINGULAR VALUES OF a IN DECREASING ORDER.
!
!       FOR range='I' OR range='V', ns IS THE NUMBER OF SINGULAR TRIPLETS
!       COMPUTED AND STORED BY THE SUBROUTINE IN ARGUMENTS u, vt AND s.
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate ( work )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
 !
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( info/=0 ) then
!
!       ANORMAL EXIT FROM gesvdx SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to GESVDX subroutine, Info=', info
!
        if ( info<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value'')') - info
!
        end if
!                
    else if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:m,:n)*V(:n,:ns) - U(:m,:ns)*S(:ns,:ns).
!
        a(:m,:ns)     = matmul(a2(:m,:n),transpose(vt(:ns,:n))) - u(:m,:ns)*spread(s(:ns),dim=1,ncopies=m)
        a2(:ns,1_i4b) = norm( a(:m,:ns), dim=2_i4b )
!
        err1          =  maxval( a2(:ns,1_i4b) )/( anorm*real(m,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
        call unit_matrix( a2(:ns,:ns) )
!
        a(:ns,:ns) = abs( a2(:ns,:ns) - matmul( transpose(u(:m,:ns)), u(:m,:ns) ) )
!
        err2       = maxval( a(:ns,:ns) )/real(m,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V**(t)*V.
!
        a(:ns,:ns) = abs( a2(:ns,:ns) - matmul( vt(:ns,:n), transpose(vt(:ns,:n)) ) )
!
        err3 = maxval( a(:ns,:ns) )/real(n,stnd)
!
        err = max( err1, err2, err3 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, s, u, vt, iwork, a2 )
    else
        deallocate( a, s, u, vt, iwork )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. info==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix_type < 1  -> random matrix from the uniform distribution'
    write (prtunit,*) ' Matrix type = 1  -> slow decay of singular values'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of singular values'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of singular values'
    write (prtunit,*) ' Matrix type = 4  -> very slow decay of singular values'
    write (prtunit,*) ' Matrix type = 5  -> strongly clustered singular values at 1'
    write (prtunit,*) '                     spectrum with few deflations (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type = 6  -> geometric distribution of singular values'
    write (prtunit,*) '                     spectrum with moderate deflation (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type = 7  -> arithmetic distribution of singular values'
    write (prtunit,*) '                     spectrum with complete deflation (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type > 7  -> uniform distribution of singular values'
!
    write (prtunit,*) 
    write (prtunit,*) ' INFO ( from gesvdx() ) = ', info
!
    if ( do_test .and. info==0 ) then
!
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
!
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', ns,' singular values and vectors of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lapack_gesvdx
! ================================
!
end program ex1_lapack_gesvdx
