program ex1_lapack_syevr
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SYEVR
!   in LAPACK software. 
!                                                                              
! LATEST REVISION : 12/01/2022
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : syevr
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, two, seven, c30, c50, c1_e6,  &
                         lamch, norm, unit_matrix, random_seed_, random_number_,               &
                         gen_random_sym_mat, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE DIMENSION OF THE GENERATED SYMMETRIC MATRIX,
! neig0 IS THE NUMERICAL RANK OF THE GENERATED SYMMETRIC MATRIX (WHICH MUST BE LESS OR EQUAL TO n)
! FOR CASES GREATER THAN 0.
!
    integer(i4b), parameter :: prtunit=6, n=3000, neig0=3000
!   
! conda IS THE CONDITION NUMBER OF THE GENERATED MATRIX IN CASES 5, 6 and 7.
!   
    real(stnd), parameter  :: fudge=c50, conda=c1_e6
!
    character(len=*), parameter :: name_proc='Example 1 of syevr'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, eps, abstol, vl, vu, tmp, tmp2, &
                                               ulp, anorm, elapsed_time
    real(stnd)                              :: work2(1)
    real(stnd), dimension(:),   allocatable :: w, work, resid2
    real(stnd), dimension(:,:), allocatable :: a, a2, z
!
    integer                            :: info, lwork, liwork, iok, il, iu, m,    &
                                          istart, iend, irate, imax, itime
    integer                            :: iwork2(1)
    integer, dimension(:), allocatable :: isuppz, iwork
    integer(i4b)                       :: i, mat_type
!
    logical(lgl)   :: do_test
!
    character      :: jobz, uplo, range
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : EIGENVALUES AND, OPTIONALLY, EIGENVECTORS OF A REAL SYMMETRIC
!               MATRIX USING THE MRRR ALGORITHM.
!
!   SPECIFY THE TYPE OF INPUT SYMMETRIC MATRIX:
!
!   mat_type < 1  -> RANDOM SYMMETRIC MATRIX FROM THE UNIFORM DISTRIBUTION
!   mat_type = 1  -> SLOW DECAY OF EIGENVALUES
!   mat_type = 2  -> FAST DECAY OF EIGENVALUES
!   mat_type = 3  -> S-SHAPED DECAY OF EIGENVALUES
!   mat_type = 4  -> VERY SLOW DECAY OF EIGENVALUES
!   mat_type = 5  -> STRONGLY CLUSTERED EIGENVALUES AT 1 WITH CONDITION NUMBER conda
!                    SPECTRUM WITH FEW DEFLATIONS FOR LAPACK SYESVD ROUTINE
!   mat_type = 6  -> GEOMETRIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH MODERATE DEFLATION FOR LAPACK SYESVD ROUTINE
!   mat_type = 7  -> ARITHMETIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK SYESVD ROUTINE
!   mat_type > 7  -> UNIFORM DISTRIBUTION OF EIGENVALUES
!
    mat_type = 6_i4b
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    abstol = two*lamch('s')
    ulp = epsilon( err )
    eps = fudge*ulp
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS MUST BE PERFORMED.
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), z(n,n), w(n), isuppz(2*n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE A RANDOM UNIFORM SYMMETRIC MATRIX OR EIGENVALUES.
!
    select case( mat_type )
!
        case( :0_i4b )
!
!           RANDOM UNIFORM SYMMETRIC MATRIX.
!
            call random_number_( z )
!
            a = z + transpose( z )
!
!           COMPUTE THE FROBENIUS NORM OF THE RANDOM SYMMETRIC MATRIX.
!
            anorm = norm( a )
!
        case( 1_i4b )
!
!           SLOW DECAY OF EIGENVALUES.
!
            do i = 1_i4b, neig0
!
                tmp = real( i, stnd )
!
                w(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF EIGENVALUES.
!
            do i = 1_i4b, neig0
!
                tmp = real( i, stnd )
!
                w(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF EIGENVALUES.
!
            do i = 1_i4b, neig0
!
                tmp = real( i, stnd )
!
                w(i) = 0.0001_stnd + one/( one + exp( tmp - c30 ) )
!        
            end do
!
        case( 4_i4b )
!
!           VERY SLOW DECAY OF EIGENVALUES.
!
            tmp = real( neig0 - 1_i4b, stnd )
!
            do i = 1_i4b, neig0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                w(i) = exp( -tmp2/tmp )
!        
            end do
!
        case( 5_i4b )
!
!           STRONGLY CLUSTERED EIGENVALUES AT 1 WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH FEW DEFLATIONS FOR LAPACK SYESVD ROUTINE.
!
            w(:neig0-1_i4b) = one
            w(neig0) = one/conda
!
        case( 6_i4b )
!
!           GEOMETRIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH MODERATE DEFLATION FOR LAPACK SYESVD ROUTINE.
!
            tmp  = real( neig0 - 1_i4b, stnd )
!
            do i = 1_i4b, neig0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                w(i) = conda**( -tmp2/tmp )
!        
            end do
!
        case( 7_i4b )
!
!           ARITHMETIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK SYESVD ROUTINE.
!
            tmp  = real( neig0 - 1_i4b, stnd )
            tmp2 = one - one/conda
!
            do i = 1_i4b, neig0
!
                w(i) = one - (real( i - 1_i4b, stnd )/tmp)*tmp2
!        
            end do
!
        case default
!
!           UNIFORM DISTRIBUTION OF EIGENVALUES.
!
            call random_number_( w(:neig0) )
!
    end select
!
    if ( mat_type>0_i4b ) then
!
#ifdef _F2003
        if ( ieee_support_datatype( w ) ) then
!
            if ( .not.all( ieee_is_normal( w(:neig0) ) ) ) then
!
                call merror( name_proc//' : Exceptions occurred when generating the input symmetric matrix !'  )                    
!
            end if
!
        end if
#endif
!
!       GENERATE A n-BY-n RANDOM REAL SYMMETRIC MATRIX a WITH A SPECIFIED DISTRIBUTION
!       OF EIGENVALUES AND A RANK EQUAL TO neig0.
!
        call gen_random_sym_mat( w(:neig0), a )
!
!       COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
        anorm = norm( w(:neig0) )
!
    end if
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,n), resid2(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM SELF-ADJOINT MATRIX.
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   FIRST DETERMINE THE WORK TO DO.
!
!   IF jobz='N' COMPUTE EIGENVALUES ONLY,
!   IF jobz='V' COMPUTE EIGENVALUES AND EIGENVECTORS.
!
    jobz  = 'V'
!
!   IF range='A' ALL EIGENVALUES WILL BE FOUND,
!   IF range='V' ALL EIGENVALUES IN THE HALF OPEN INTERVAL (vl,vu) WILL BE FOUND,
!   IF range='I' THE il-TH THROUGH iu-TH EIGENVALUES IN ASCENDING ORDER WILL BE FOUND.
!
    range = 'A'
!
!   IF uplo='U' UPPER TRIANGLE OF a IS STORED,
!   IF uplo='L' LOWER TRIANGLE OF a IS STORED.
!
    uplo  = 'U'
!
    vl = zero
    vu = zero
!
    il = 1
    iu = n
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR syevr SUBROUTINE.
!
    lwork  = -1
    liwork = -1
!
    call syevr( jobz, range, uplo, n, a, n, vl, vu, il, iu, abstol,        &
                m, w, z, n, isuppz, work2, lwork, iwork2, liwork, info     )
!
    if ( info==0 ) then
!
        lwork  = int(work2(1)) + 10
        liwork = iwork2(1)
!
!       ALLOCATE WORK VARIABLES NEEDED BY syevr SUBROUTINE.
!
        allocate( work(lwork), iwork(liwork), stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE EIGENVALUES AND EIGENVECTORS OF THE SELF-ADJOINT MATRIX a
!       WITH SUBROUTINE syevr FROM LAPACK.
!       THE EIGENVALUE DECOMPOSITION OF A REAL n-BY-n SYMMETRIC MATRIX a
!       IS WRITTEN
!
!                       a = U * D * U**(t)
!
!       WHERE D IS AN n-BY-n DIAGONAL MATRIX AND U IS AN n-BY-n ORTHOGONAL MATRIX.
!       THE DIAGONAL ELEMENTS OF D ARE THE EIGENVALUES OF a; THEY ARE REAL.
!       THE COLUMNS OF U ARE THE EIGENVECTORS OF a.
!
        call syevr( jobz, range, uplo, n, a, n, vl, vu, il, iu, abstol,      &
                    m, w, z, n, isuppz, work, lwork, iwork, liwork, info     )
!
!       ON EXIT OF syevr :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT.
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE.
!         info=i  > 0 :  INDICATES THAT THE ALGORITHM FAILED TO CONVERGE.
!
!       ON EXIT OF syevr :
!
!         THE LOWER (IF uplo='L') OR UPPER (IF uplo='U') TRIANGLE
!         OF a IS DESTROYED, INCLUDING THE DIAGONAL.
!
!         m GIVES THE TOTAL NUMBER OF EIGENVALUES/EIGENVECTORS FOUND.
!
!         z IS OVERWRITTEN WITH THE FIRST m EIGENVECTORS
!         OF a (THE EIGENVECTORS ARE STORED COLUMNWISE).
!
!         w IS OVERWRITTEN WITH THE EIGENVALUES OF a IN ASCENDING ORDER,
!         CORRESPONDING TO THE EIGENVECTORS IN z .
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate ( work, iwork )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( info/=0 ) then
!
!       ANORMAL EXIT FROM syevr SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to SYEVR subroutine, Info=', info
!
        if ( info<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value'')') - info
!
        end if
!                
    else if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*U - U*D
!
        a(:n,:m) = matmul(a2(:n,:n),z(:n,:m)) - z(:n,:m)*spread(w(:m),dim=1,ncopies=n)
        resid2(:m) = norm( a(:n,:m), dim=2_i4b )
        err1 =  maxval( resid2(:m) )/( anorm*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
        call unit_matrix( a2(:m,:m) )
!
        a(:m,:m) = abs( a2(:m,:m) - matmul( transpose(z(:n,:m)), z(:n,:m) ) )
        err2 = maxval( a(:m,:m) )/real(n,stnd)
!
        err = max( err1, err2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, z, w, isuppz, a2, resid2 )
    else
       deallocate( a, z, w, isuppz )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. info==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix_type < 1  -> random symmetric matrix from the uniform distribution'
    write (prtunit,*) ' Matrix type = 1  -> slow decay of eigenvalues'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of eigenvalues'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of eigenvalues'
    write (prtunit,*) ' Matrix type = 4  -> very slow decay of eigenvalues'
    write (prtunit,*) ' Matrix type = 5  -> strongly clustered eigenvalues at 1'
    write (prtunit,*) '                     spectrum with few deflations (for LAPACK SYESVD routine)'
    write (prtunit,*) ' Matrix type = 6  -> geometric distribution of eigenvalues'
    write (prtunit,*) '                     spectrum with moderate deflation (for LAPACK SYESVD routine)'
    write (prtunit,*) ' Matrix type = 7  -> arithmetic distribution of eigenvalues'
    write (prtunit,*) '                     spectrum with complete deflation (for LAPACK SYESVD routine)'
    write (prtunit,*) ' Matrix type > 7  -> uniform distribution of eigenvalues'
!
    write (prtunit,*) 
    write (prtunit,*) ' INFO ( from syevr() ) = ', info
!
    if ( do_test .and. info==0  ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', m, ' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lapack_syevr
! ===============================
!
end program ex1_lapack_syevr
