program ex1_lin_lu_solve
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine LIN_LU_SOLVE
!   in module Lin_Procedures. 
!                                                                              
! LATEST REVISION : 21/02/2022
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, safmin, zero, half, lin_lu_solve, &
                         norm, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE SIZE OF THE LINEAR SYSTEM TO BE SOLVED.
!
    integer(i4b), parameter :: prtunit=6, n=4001
!
    character(len=*), parameter :: name_proc='Example 1 of lin_lu_solve'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, tmp, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a
    real(stnd), dimension(:), allocatable   :: b, x, res
!
    integer :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: failure, do_test
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : SOLVING A LINEAR SYSTEM WITH A REAL n-BY-n MATRIX
!               AND ONE RIGHT HAND-SIDE WITH THE LU DECOMPOSITION.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = real( n, stnd)*sqrt( epsilon( err ) )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), b(n), x(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM DATA MATRIX a .
!
    call random_number( a )
!
    a = a - half
!
    call random_number( tmp )
!
    if ( tmp>safmin ) then
        a = a/tmp
    end if
!
!   GENERATE A n-ELEMENT RANDOM SOLUTION VECTOR x .
!
    call random_number( x )
!
!   COMPUTE THE MATRIX-VECTOR PRODUCT b = a*x .
!
    b = matmul( a, x )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE SOLUTION VECTOR FOR LINEAR SYSTEM
!
!            a*x = b .
!
!   BY COMPUTING THE LU DECOMPOSITION WITH PARTIAL PIVOTING AND
!   IMPLICIT ROW SCALING OF MATRIX a. IF ON OUTPUT OF lin_lu_solve
!   failure IS SET TO FALSE THEN THE LINEAR SYSTEM IS NOT SINGULAR
!   AND THE SOLUTION VECTOR HAS BEEN COMPUTED.
!
    call lin_lu_solve( a, b, failure )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( failure ) then
!
!       ANORMAL EXIT FROM lin_lu_solve SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to LIN_LU_SOLVE subroutine, failure=', failure
!
    else if ( do_test ) then
!
!       ALLOCATE WORK ARRAY .
!
        allocate( res(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK THE RESULTS FOR SMALL RESIDUALS.
!
        res(:n) = b(:n) - x(:n)
        err     = norm(res)/norm(x)
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( res )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, b, x )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure ) then
!
        write (prtunit,*) name_proc//' is correct'
!
        if ( do_test ) then
!
            write (prtunit,*) 
!
!           PRINT RELATIVE ERROR OF COMPUTED SOLUTION.
!
            write (prtunit,*) 'Relative error of the computed solution = ', err
!
        end if
!
    else
!
        write (prtunit,*) name_proc//' is incorrect'
!
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the solution of a linear real system of size ', &
       n, ' by ', n, ' is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lin_lu_solve
! ===============================
!
end program ex1_lin_lu_solve
