program ex1_llsq_svd_solve
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine LLSQ_SVD_SOLVE
!   in module LLSQ_Procedures. 
!                                                                              
! LATEST REVISION : 20/01/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c100, lamch, norm,        &
                         print_array, llsq_svd_solve, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT.
! m IS THE NUMBER OF ROWS OF THE RANDOM MATRIX.
! n IS THE NUMBER OF COLUMNS OF THE RANDOM MATRIX.
!
    integer(i4b), parameter :: prtunit=6, m=8000, n=4000, mn=min(m,n)
!   
    real(stnd), parameter  :: fudge=c100
!
    character(len=*), parameter :: name_proc='Example 1 of llsq_svd_solve'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: err, err1, err2, ulp, eps, tol, anorm, rnorm, bnorm, cond, sfmin, elapsed_time
    real(stnd), allocatable, dimension(:)   :: b, b2, res, res2, x, sing_values
    real(stnd), allocatable, dimension(:,:) :: a, a2
!
    integer(i4b) :: krank, j
!
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: failure, do_test, do_print
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : SOLVE A LINEAR LEAST SQUARES SYSTEM WITH ONE RIGHT HAND-SIDE
!               BY THE SINGULAR VALUE DECOMPOSITION OF THE COEFFICIENT MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( ulp )
    eps = fudge*ulp
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE SUBROUTINE
!   AND IF DETAILED RESULTS MUST BE PRINTED.
!
    do_test  = true
    do_print = false
!
!   SET TOLERANCE FOR DETERMINING THE RANK OF THE MATRIX IN THE SUBROUTINE.
!
!    tol = 0.0000001_stnd
    tol = sqrt( ulp )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), b(m), x(n), sing_values(mn), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-by-n REAL RANDOM COEFFICIENT MATRIX a .
!
    call random_number( a )
!
!   GENERATE A DEFICIENT COEFFICIENT MATRIX a(:m,:n) WITH RANK min(mn-1,n) .
!
    j = min( n, 5_i4b ) 
    a(:m,j) = a(:m,1_i4b) + a(:m,2_i4b)
!
!   GENERATE A RANDOM RIGHT HAND-SIDE VECTOR b .
!
    call random_number( b )
!
!   COMPUTE THE NORM OF DEPENDENT VARIABLE b .
!
    bnorm = norm( b(:m) )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(m,n), b2(m), res(m), res2(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE FROBENIUS NORM OF THE MATRIX.
!
        anorm = norm( a(:m,:n) )
!
!       SAVE DATA MATRIX.
!
        a2(:m,:n) = a(:m,:n)
!
!       SAVE RIGHT HAND SIDE VECTOR.
!
        b2(:m) = b(:m)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   llsq_svd_solve COMPUTES THE MINIMUM NORM SOLUTION TO A REAL LINEAR LEAST
!   SQUARES PROBLEM :
!
!                       Minimize || b - a*x ||_2
!
!   USING THE SINGULAR VALUE DECOMPOSITION (SVD) OF a. A IS AN m-BY-n MATRIX
!   WHICH MAY BE RANK-DEFICIENT. b AND x CAN BE VECTORS OF MATRICES, BUT THEIR
!   SHAPES MUST BE CONFORMABLE WITH THE SHAPE OF a.
!
!   IN OTHER WORDS, IF b AND x ARE MATRICES, SEVERAL RIGHT HAND SIDE VECTORS b
!   AND SOLUTION VECTORS x CAN BE HANDLED IN A SINGLE CALL; THEY ARE STORED AS
!   THE COLUMNS OF THE m-BY-nrhs RIGHT HAND SIDE MATRIX b AND THE n-BY-nrhs SOLUTION
!   MATRIX b, RESPECTIVELY.
!
!   THE EFFECTIVE RANK OF a, krank,IS DETERMINED BY TREATING AS ZERO THOSE
!   SINGULAR VALUES WHICH ARE LESS THAN tol TIMES THE LARGEST SINGULAR VALUE.
!
    call llsq_svd_solve( a, b, failure, x,                                            &
                         singvalues=sing_values, krank=krank, rnorm=rnorm, tol=tol )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK THAT THE RESIDUAL VECTOR IS ORTHOGONAL TO THE RANGE OF a .
!
        res(:m)  = b2(:m) - matmul( a2(:m,:n), x(:n) )
        res2(:n) = matmul( res(:m), a2(:m,:n) )
!
        err1 = maxval( abs(res2(:n)) )/anorm
!
!       CHECK THE NORM OF THE RESIDUAL VECTOR.
!
        err2 = abs( norm( res(:m) ) - rnorm )
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, b2, res, res2 )
!
    end if
!
!   PRINT THE RESULT OF THE TESTS.
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
!   GET MACHINE CONSTANT sfmin, SUCH THAT 1/sfmin DOES NOT OVERFLOW.
!
    sfmin = lamch( 's' )
!
!   COMPUTE THE CONDITION NUMBER OF a(:m,:n) IN THE 2-NORM
!
!            singvalues(1)/singvalues(min(m,n)) .
!
    if ( sing_values(mn)/sing_values(1_i4b)<=sfmin ) then
        cond = huge( cond )
    else
        cond = sing_values(1_i4b)/sing_values(mn)
    end if
!
!   PRINT RESULTS .
!
    write (prtunit,*)
    write (prtunit,*)
    write (prtunit,*) 'Least squares solution via Singular Value Decomposition'
    write (prtunit,*)
    write (prtunit,*) '    min of ||a(:,:)*x(:)-b(:)||**2 for vector x(:)     '
    write (prtunit,*)
    write (prtunit,*) 'Tolerance for zero singular values (tol*sing_values(1)):',tol*sing_values(1)
    write (prtunit,*)
    write (prtunit,*) 'Condition number (in the 2-norm) of a :',cond
    write (prtunit,*) 'Rank of a                             :',krank
    write (prtunit,*)
    write (prtunit,*) 'Residual sum of squares     ||a*x-b||**2          :',rnorm**2
    write (prtunit,*) 'Residual sum of squares (%) ||a*x-b||**2/||b||**2 :',(rnorm/bnorm)**2
    write (prtunit,*)
!
    if ( do_print ) then
!
!       PRINT DETAILED RESULTS.
!
        call print_array( sing_values, title=' Singular values of a ' )
!
        write (prtunit,*)
!
        call print_array( x, title=' Least squares solution vector x ' )
!
    end if
!
    write (prtunit,*)
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for solving a linear least squares problem with a ', &
       m, ' by ', n,' real coefficient matrix is ', elapsed_time, ' seconds'
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, b, x, sing_values )
!
!
! END OF PROGRAM ex1_llsq_svd_solve
! =================================
!
end program ex1_llsq_svd_solve
