program ex1_lq_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines LQ_CMP and ORTHO_GEN_LQ
!   in module QR_Procedures.
!                                                                              
! LATEST REVISION : 19/01/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, zero, one, c50, true, false, lq_cmp, ortho_gen_lq, norm, &
                         merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT.
! m IS THE NUMBER OF ROWS OF THE RANDOM MATRIX.
! n IS THE NUMBER OF COLUMNS OF THE RANDOM MATRIX.
!
    integer(i4b), parameter :: prtunit=6, m=3000, n=4000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of lq_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, ulp, elapsed_time
    real(stnd), dimension(:),   allocatable :: diagl, tau, resid2, norma
    real(stnd), dimension(:,:), allocatable :: a, q, l, resid
!
    integer(i4b) :: k, j, p
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : COMPUTING A FULL LQ DECOMPOSITION OF A MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = fudge*ulp
!
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE SUBROUTINE.
!
    do_test = true
!
    k = min( m, n )
    p = max( m, n )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), diagl(k), tau(k), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM COEFFICIENT MATRIX a(:m,:n).
!
    call random_number( a(:m,:n) )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( l(m,k), q(p,n), resid(p,n), resid2(m), norma(m), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX.
!
        resid(:m,:n) = a(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE LQ DECOMPOSITION OF RANDOM DATA MATRIX.
!
    call lq_cmp( a(:m,:n), diagl(:k), tau(:k) )
!
!   lq_cmp COMPUTES AN ORTHOGONAL FACTORIZATION OF A REAL m-BY-n MATRIX
!   BY THE HOUSEOLDER METHOD. THE MATRIX IS ASSUMED OF FULL RANK. THE ROUTINE
!   COMPUTES A LQ FACTORIZATION OF a AS:
!
!                     a = L * Q
!
!   Q IS A n-BY-n ORTHOGONAL MATRIX AND L IS A m-BY-n LOWER TRIANGULAR OR
!   TRAPEZOIDAL MATRIX.
!
!   ON EXIT, a HAS BEEN OVERWRITTEN BY DETAILS OF ITS ORTHOGONAL
!   FACTORIZATION. 
!
!   THE MATRIX Q IS REPRESENTED AS A PRODUCT OF ELEMENTARY REFLECTORS
!
!            q = h(k)*h(k-1)* ... *h(1), WHERE k = min( size(a,1) , size(a,2) )
!
!   EACH h(i) HAS THE FORM
!
!            h(i) = I + TAU * ( V * V' ) ,
!                      
!   WHERE TAU IS A REAL SCALAR AND V IS A REAL n-ELEMENT VECTOR WITH V(1:i-1) = 0.
!   V(i:n) IS STORED ON EXIT IN a(i,i:n) AND TAU IN tau(i).
!                      
!   THE ELEMENTS BELOW THE DIAGONAL OF THE ARRAY a CONTAIN THE
!   CORRESPONDING ELEMENTS OF THE TRIANGULAR MATRIX L. THE ELEMENTS
!   OF THE DIAGONAL OF L ARE STORED IN THE ARRAY diagl ON EXIT. 
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!                
    if ( do_test ) then
!
!       NOW, RESTORE TRIANGULAR FACTOR L OF LQ DECOMPOSITION OF RANDOM DATA MATRIX a
!       IN MATRIX l(:m,:k).
!
        do j = 1, k
!
            l(1:j-1,j) = zero
            l(j,j)     = diagl(j)
            l(j+1:m,j) = a(j+1:m,j)
!
        end do
!
        q(:k,:n) = a(:k,:n)
!
!       GENERATE ORTHOGONAL MATRIX Q OF LQ DECOMPOSITION OF RANDOM DATA MATRIX a
!       AND AN ORTHOGONAL BASIS FOR THE ORTHOGONAL COMPLEMENT TO THE RANGE OF a'.
!       a IS ASSUMED OF FULL RANK.
!
        call ortho_gen_lq( q(:n,:n), tau(:k) )
!
!       ortho_gen_lq GENERATES AN n-BY-n REAL MATRIX WITH ORTHONORMAL ROWS, WHICH IS
!       DEFINED AS THE PRODUCT OF k ELEMENTARY REFLECTORS OF ORDER n
!
!            q = h(k)*h(k-1)* ... *h(1)
!
!       AS RETURNED BY lq_cmp.
!
!       THE SIZE OF tau DETERMINES THE NUMBER k OF ELEMENTARY REFLECTORS
!       WHOSE PRODUCT DEFINES THE MATRIX Q.
!
!       NOW, THE ROWS OF q(:k,:n) ARE AN ORTHOGONAL BASIS FOR THE RANGE OF a'
!       AND THE ROWS OF q(k+1:n,:n) ARE AN ORTHOGONAL BASIS FOR THE ORTHOGONAL
!       COMPLEMENT TO THE RANGE OF a'.
!
!       RESTORE THE INPUT MATRIX IN a(:m,:n) .
!
        a(:m,:n) = resid(:m,:n)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:m,:n) - l(:m,:k)*q(:k,:n).
!
        resid(:m,:n) = a(:m,:n) - matmul( l(:m,:k), q(:k,:n) )
        resid2(:m)   = norm( resid(:m,:n), dim=1_i4b )
        norma(:m)    = norm( a(:m,:n), dim=1_i4b )
        err1         = maxval( resid2(:m) / norma(:m) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q*Q'.
!
        resid(:n,:n) = matmul( q(:n,:n), transpose(q(:n,:n)) )
!
        do j = 1, n
            resid(j,j) =  resid(j,j) - one
        end do
!
        err2 = maxval( abs(resid(:n,:n)) )/real(n,stnd)
!
!       CHECK ORTHOGONALITY OF a(:m,:n) AND ITS ORTHOGONAL COMPLEMENT q(m+1:n,:n).
!
        if ( m<n ) then
!
            resid(:m,m+1_i4b:n) = matmul( a(:m,:n), transpose(q(m+1_i4b:n,:n) ) )
            err3 = maxval( abs( resid(:m,m+1_i4b:n) ) )/real(n,stnd)
!
        else
!
            err3 = zero
!
        end if
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( l, q, resid, resid2, norma )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, diagl, tau )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
!
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the LQ decomposition     &
                          &                                      = ', err1
        write (prtunit,*) 'Orthogonality of the Q matrix        &
                          &                                      = ', err2
!
        if ( m<n ) then
            write (prtunit,*) 'Orthogonality of the row-space of the matrix&
                              & and its orthogonal complement = ', err3
        end if
!
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing a LQ decomposition of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lq_cmp
! =========================
!
end program ex1_lq_cmp
