program ex1_partial_rtqr_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines PARTIAL_RTQR_CMP in 
!   module Random and ORTHO_GEN_QR in module QR_Procedures.
!                                                                              
! LATEST REVISION : 05/02/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, seven, c30, c50, partial_rtqr_cmp, &
                         ortho_gen_qr, norm, merror, allocate_error, gen_random_mat, random_seed_
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!    
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT.
! m IS THE NUMBER OF ROWS OF THE RANDOM MATRIX.
! n IS THE NUMBER OF COLUMNS OF THE RANDOM MATRIX.
! nsvd0 IS THE NUMERICAL RANK OF THE GENERATED MATRIX,
! nqr IS THE TARGET RANK OF THE PARTIAL QR APPROXIMATION, WHICH IS SOUGHT.
!
    integer(i4b), parameter :: prtunit = 6, m=4000, n=3000, nsvd0=2000, nqr=100
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of partial_rtqr_cmp'
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, tmp, norma, eps, ulp, tol, elapsed_time
    real(stnd), allocatable, dimension(:)   :: diagr, beta, singval0
    real(stnd), allocatable, dimension(:,:) :: a, q, r, resid
!
    integer(i4b)                            :: i, j, krank, niter, nover, mat_type
    integer(i4b), allocatable, dimension(:) :: ip
    integer                                 :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : COMPUTE A RANDOMIZED PARTIAL AND TRUNCATED QR DECOMPOSITION WITH COLUMN PIVOTING
!               OF A DATA MATRIX.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type > 3  -> VERY SLOW DECAY OF SINGULAR VALUES
!
    mat_type = 4_i4b
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE SUBROUTINE.
!
    do_test = true
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp  = epsilon( ulp )
    eps  = fudge*ulp
    err3 = zero
!
!   SET TOLERANCE FOR DETERMINING THE EXACT RANK OF THE PARTIAL QR APPROXIMATION.
!
!    tol = 0.000001_stnd
    tol = sqrt( ulp )
!
!   DETERMINE THE NUMBER OF SUBSPACE ITERATIONS PERFORMED FOR IMPROVING THE QUALITY
!   OF THE COMPRESSION MATRIX USED IN THE RANDOMIZED PARTIAL QR ALGORITHM.
!
    niter = 2_i4b
!
!   DETERMINE THE OVERSAMPLING SIZE PARAMETER FOR IMPROVING THE QUALITY
!   OF THE COMPRESSION MATRIX USED IN THE RANDOMIZED PARTIAL QR ALGORITHM.
!
    nover = max( 10_i4b, nqr/2_i4b )
!
!   ALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
!
        i = max( m, n )
!
        allocate( a(m,n), diagr(nqr), beta(nqr), ip(n), singval0(nsvd0),  &
                  r(nqr,n), q(m,i), resid(m,i), stat=iok )
!
    else
!
        allocate( a(m,n), diagr(nqr), beta(nqr), ip(n), singval0(nsvd0),  &
                  r(nqr,n), q(m,nqr), resid(m,n), stat=iok )
!
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A SPECIFIED DISTRIBUTION
!   OF SINGULAR VALUES.
!
!   GENERATE SINGULAR VALUES.
!
    select case( mat_type )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = 0.0001_stnd + one/( one + exp( tmp - c30 ) )
!        
            end do
!
        case default
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            norma = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i - 1_i4b, stnd )
!
                singval0(i) = exp( -tmp/norma )
!        
            end do
!
    end select
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE A m-BY-n MATRIX a OF RANK nsvd0 .
!
    call gen_random_mat( singval0(:nsvd0), a )
!
!   COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
!    norma = norm( a(:m,:n) )
    norma = sqrt(sum( singval0(:nsvd0)**2 ) )
!
!   MAKE A COPY OF THE DATA MATRIX FOR LATER USE.
!
    resid(:m,:n) = a(:m,:n)
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE A RANDOMIZED PARTIAL AND TRUNCATED QR DECOMPOSITION WITH COLUMN PIVOTING OF A DATA
!   MATRIX a WITH SUBROUTINE partial_rtqr_cmp. THE RANK OF THE PARTIAL QR DECOMPOSITION IS
!   DETERMINED BY THE SIZE OF THE ARGUMENT diagr, nqr = size(diagr) .
!
    call partial_rtqr_cmp( a(:m,:n), diagr(:nqr), beta(:nqr), ip(:n), krank, tol=tol,  &
                           niter=niter, nover=nover )
!
!   THE ROUTINE COMPUTES AN APPROXIMATE RANDOMIZED PARTIAL AND TRUNCATED QR FACTORIZATION
!   WITH COLUMN PIVOTING OF a AS:
!
!                     a * P ≈ Q * R = Q * [ R11  R12 ]
!
!   WHERE Q IS A m-BY-nqr MATRIX WITH ORTHOGONAL COLUMNS, R IS A nqr-BY-n UPPER OR TRAPEZOIDAL
!   MATRIX AND R11 IS A nqr-BY-nqr UPPER TRIANGULAR MATRIX.
!
!   ON EXIT, a HAS BEEN OVERWRITTEN BY DETAILS OF ITS PARTIAL QR FACTORIZATION:
!                      
!   - THE ELEMENTS BELOW AND ON THE DIAGONAL OF THE ARRAY a(:m,:nqr) AND THE ARRAY
!     beta(:nqr) STORED Q IN FACTORED FORM.
!                      
!   - THE ELEMENTS ABOVE THE DIAGONAL OF THE ARRAY a CONTAIN THE
!     CORRESPONDING ELEMENTS OF THE TRIANGULAR MATRIX R. THE ELEMENTS
!     OF THE DIAGONAL OF R ARE STORED IN THE ARRAY diagr. 
!
!   - ip STORES THE PERMUTATION MATRIX P IN THE QR DECOMPOSITION OF a.
!     IF ip(j)=k, THEN THE jTH COLUMN OF a*P WAS THE kTH COLUMN OF a.
!     THE MATRIX P IS REPRESENTED IN THE ARRAY ip AS FOLLOWS:
!     IF ip(j) = i THEN THE jTH COLUMN OF P IS THE iTH CANONICAL UNIT VECTOR.
!                      
!   - krank CONTAINS THE EFFECTIVE RANK OF THE QR APPROXIMATION (e.g. THE RANK OF R11),
!     WHICH IS EQUAL TO nqr IF THE RANK OF a IS GREATER THAN nqr .
!                      
!   IF tol IS PRESENT AND IS IN ]0,1[, THEN :
!       CALCULATIONS TO DETERMINE THE CONDITION NUMBER OF THE SUBMATRIX R11
!       ARE PERFORMED. THEN, tol IS USED TO DETERMINE THE EFFECTIVE RANK
!       OF R11, WHICH IS DEFINED AS THE ORDER OF THE LARGEST LEADING
!       TRIANGULAR SUBMATRIX IN THE PARTIAL QR FACTORIZATION WITH PIVOTING OF a,
!       WHOSE ESTIMATED CONDITION NUMBER IS LESS THAN 1/tol.
!       ON EXIT, THE RECIPROCAL OF THE CONDITION NUMBER IS RETURNED IN tol.
!
!   IF tol IS PRESENT AND IS EQUAL TO 0, THEN :
!       THE NUMERICAL RANK OF R IS DETERMINED, E.G., CRUDE TESTS ON R(j,j) ARE
!       DONE TO DETERMINE THE NUMERICAL RANK OF R11 AND tol IS NOT CHANGED ON EXIT.
!
!   IF tol IS NOT SPECIFIED OR IS OUTSIDE [0,1[ :
!       THE CALCULATIONS TO DETERMINE THE CONDITION NUMBER OF R11 ARE NOT
!       PERFORMED AND THE RANK OF R11 IS ASSUMED TO BE EQUAL TO nqr.
!
!   IF IT IS POSSIBLE THAT a MAY NOT BE FULL RANK (I.E., CERTAIN COLUMNS OF a ARE
!   LINEAR COMBINATIONS OF OTHER COLUMNS), THEN THE LINEARLY DEPENDENT COLUMNS CAN
!   USUALLY BE EXCLUDED FROM THE QR APPROXIMATION BY USING tol=RELATIVE PRECISION
!   OF THE ELEMENTS IN a.
!   IF EACH ELEMENT IS CORRECT TO, SAY, 5 DIGITS THEN tol=0.00001 SHOULD BE USED.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   NOW, RESTORE TRIANGULAR FACTOR R OF PARTIAL QR DECOMPOSITION OF DATA MATRIX a
!   IN MATRIX r(:nqr,:n) .
!
    do j = 1_i4b, nqr
!
        r(1_i4b:j-1_i4b,j) = a(1_i4b:j-1_i4b,j)
        r(j,j)             = diagr(j)
        r(j+1_i4b:nqr,j)   = zero
!
    end do
!
    do j = nqr+1_i4b, n
!
        r(1_i4b:nqr,j) = a(1_i4b:nqr,j)
!
    end do
!
!   GENERATE ORTHOGONAL MATRIX q OF PARTIAL QR DECOMPOSITION OF DATA MATRIX a .
!
    q(:m,:krank) = a(:m,:krank)
!                
    if ( do_test ) then
!
        i = m
!
    else
!
        i = krank
!
    end if
!
    call ortho_gen_qr( q(:m,:i), beta(:krank) )
!
!   HERE ortho_gen_qr GENERATES AN m-BY-l REAL MATRIX WITH ORTHONORMAL COLUMNS, WHICH IS
!   DEFINED AS THE FIRST l COLUMNS OF A PRODUCT OF krank ELEMENTARY REFLECTORS OF ORDER m
!
!            Q = h(1)*h(2)* ... *h(krank)
!
!   AS RETURNED BY qr_cmp, qr_cmp2, partial_qr_cmp, partial_rqr_cmp, partial_rqr_cmp2
!   AND partial_rtqr_cmp SUBROUTINES.
!
!   THE SIZE OF beta DETERMINES THE NUMBER OF ELEMENTARY REFLECTORS WHOSE PRODUCT
!   DEFINES THE MATRIX Q.
!
!   APPLY PERMUTATION TO a .
!
    do j = 1_i4b, n
!
        a(:m,j) = resid(:m,ip(j))
!
    end do
!
!   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:m,:n)*P(:n,:n) - Q(:m,:krank)*R(:krank,:n).
!
    resid(:m,:n) = a(:m,:n) - matmul( q(:m,:krank), r(:krank,:n) )
!
!   COMPUTE RELATIVE ERROR OF THE QR APPROXIMATION.
!
    err1 = norm( resid(:m,:n) )/norma
!                
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION R(:krank,:n) - Q(:m,:krank)'*a*P .
!
        resid(:krank,:n) = abs( r(:krank,:n) - matmul( transpose(q(:m,:krank)), a(:m,:n) ) )
        err2 = maxval( resid(:krank,:n) )/real(m,stnd)
!
!       CHECK ORTHOGONALITY OF (a*P)(:m,:krank) AND ITS ORTHOGONAL COMPLEMENT Q(:m,krank+1:m).
!
        if ( m>krank ) then
!
            resid(:krank,krank+1:m) = matmul( transpose(a(:m,:krank)), q(:m,krank+1:m) )
            err3 = maxval( abs( resid(:krank,krank+1:m) ) )/real(m,stnd)
!
        else
!
            err3 = zero
!
        end if
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, diagr, beta, ip, singval0, r, q, resid )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err3<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = 1  -> slow decay of singular values'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of singular values'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of singular values'
    write (prtunit,*) ' Matrix type > 3  -> very slow decay of singular values'
!
    write (prtunit,*) 
    write (prtunit,*) 'Rank of the partial QR approximation     &
                       &                                        = ', krank
!        
    write (prtunit,*) 'Accuracy of the partial QR decomposition &
                       &||A - Q*R||_F/||A||_F                   = ', err1
!
    if ( do_test ) then
!        
        write (prtunit,*) 'Accuracy of the range of the partial QR &
                          &approximation                            = ', err2
!
        if ( m>krank ) then
            write (prtunit,*) 'Orthogonality of the range of the QR approximation&
                               & and its orthogonal complement = ', err3
        end if
!
    end if
!
    write (prtunit,*)
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing a randomized partial and truncated QR decomposition with column pivoting of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_partial_rtqr_cmp
! ===================================
!
end program ex1_partial_rtqr_cmp
