program ex1_qr_solve
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine QR_CMP in module QR_Procedures
!   and QR_SOLVE in module LLSQ_Procedures.
!                                                                              
! LATEST REVISION : 21/01/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, lgl, stnd, zero, true, false, c500, allocate_error, merror, &
                         qr_cmp, qr_solve, norm
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT.
! m IS THE NUMBER OF ROWS OF THE RANDOM MATRIX.
! n IS THE NUMBER OF COLUMNS OF THE RANDOM MATRIX.
! nrhs IS THE NUMBER OF COLUMNS OF THE RIGHT HAND SIDE MATRIX.
!
    integer(i4b), parameter :: prtunit=6, m=4000, n=3000, mn=min( m, n ), nrhs=10
!   
    real(stnd), parameter  :: fudge=c500
!
    character(len=*), parameter :: name_proc='Example 1 of qr_solve'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: ulp, eps, anorm, err1, err2, err, elapsed_time
    real(stnd), allocatable, dimension(:)   :: diagr, beta, rnorm
    real(stnd), allocatable, dimension(:,:) :: a, a2, b, x, res
!
    integer :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: comp_resid, do_test
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : SOLVING LINEAR LEAST SQUARES PROBLEMS WITH A m-BY-n REAL COEFFICIENT
!               MATRIX USING A QR DECOMPOSITION OF THE COEFFICIENT MATRIX
!               AND SEVERAL RIGHT HAND-SIDES. THE COEFFICIENT MATRIX IS
!               ASSUMED OF FULL RANK, BUT BOTH m>=n OR m<n ARE PERMITTED.
!
!               COMPUTE SOLUTION MATRIX x FOR LINEAR LEAST SQUARES SYSTEM
!
!                              a(:m,:n)*x(:n,:nrhs) ≈ b(:m,:nrhs) .
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = fudge*ulp
!
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE SUBROUTINE.
!
    do_test = true
!
!   DECIDE IF RESIDUAL MATRIX MUST BE COMPUTED.
!
    comp_resid = do_test
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), b(m,nrhs), x(n,nrhs), diagr(mn), beta(mn), rnorm(nrhs), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM COEFFICIENT MATRIX a(:m,:n) .
!
    call random_number( a(:m,:n) )
!
!   GENERATE A RANDOM RIGHT HAND-SIDE MATRIX b(:m,:nrhs) .
!
    call random_number( b(:m,:nrhs) )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS FOR LATER USE.
!
        allocate( a2(m,n), res(nrhs,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE FROBENIUS NORM OF THE DATA MATRIX.
!
        anorm = norm( a(:m,:n) )
!
!       MAKE A COPY OF THE DATA MATRIX FOR LATER USE.
!
        a2(:m,:n) = a(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE A QR FACTORIZATION OF RANDOM DATA MATRIX WITH SUBROUTINE qr_cmp.
!
    call qr_cmp( a(:m,:n), diagr(:mn), beta(:mn) )
!
!   qr_cmp COMPUTES AN ORTHOGONAL FACTORIZATION OF A REAL m-BY-n MATRIX
!   a . a IS ASSUMED OF FULL RANK. THE ROUTINE COMPUTES A QR FACTORIZATION
!   OF a AS:
!
!                     a = Q * R
!
!   ON EXIT, a HAS BEEN OVERWRITTEN BY DETAILS OF ITS ORTHOGONAL
!   FACTORIZATION. 
!
!   THE MATRIX Q IS REPRESENTED AS A PRODUCT OF ELEMENTARY REFLECTORS
!
!            Q = h(1)*h(2)* ... *h(k), WHERE k = min( size(a,1) , size(a,2) )
!
!   EACH h(i) HAS THE FORM
!
!            h(i) = I + BETA * ( V * V' ) ,
!                      
!   WHERE BETA IS A REAL SCALAR AND V IS A REAL m-ELEMENTS VECTOR WITH V(1:i-1) = 0.
!   V(i:m) IS STORED ON EXIT IN a(i:m,i) AND BETA IN beta(i).
!                      
!   THE ELEMENTS ABOVE THE DIAGONAL OF THE ARRAY a CONTAIN THE
!   CORRESPONDING ELEMENTS OF THE TRIANGULAR MATRIX R. THE ELEMENTS
!   OF THE DIAGONAL OF R ARE STORED IN THE ARRAY diagr. 
!
!   NEXT, COMPUTE THE SOLUTION MATRIX FOR LINEAR LEAST SQUARES SYSTEM
!
!                   a(:m,:n)*x(:n,:nrhs) ≈ b(:m,:nrhs) .
!
!   WITH SUBROUTINE qr_solve AND THE QR DECOMPOSITION COMPUTED BY qr_cmp.
!
    call qr_solve( a(:m,:n), diagr(:mn), beta(:mn), b(:m,:nrhs), x(:n,:nrhs),  &
                   rnorm=rnorm(:nrhs), comp_resid=comp_resid )
!
!   qr_solve COMPUTES SOLUTION(S) TO REAL LINEAR LEAST SQUARES PROBLEMS 
!   OF THE FORM:
!
!                       Minimize || b - a*x ||_2
!
!   USING A QR FACTORIZATION COMPUTED BY qr_cmp. a IS AN m-BY-n MATRIX
!   WHICH IS ASSUMED OF FULL RANK, BUT BOTH m>=n OR n>m ARE PERMITTED.
!
!   SEVERAL RIGHT HAND SIDE VECTORS b AND SOLUTION VECTORS x CAN BE
!   HANDLED IN A SINGLE CALL; THEY ARE STORED AS THE COLUMNS OF THE
!   m-BY-nrhs RIGHT HAND SIDE MATRIX b AND THE n-BY-nrhs SOLUTION
!   MATRIX x, RESPECTIVELY. IN OTHER WORDS, THE ARGUMENTS b AND x MAY
!   BE VECTORS OR MATRICES. b IS OVERWRITTEN BY qr_solve.
!   
!   ON EXIT, IF THE OPTIONAL LOGICAL PARAMETER comp_resid IS PRESENT AND IS SET TO true,
!   THE RESIDUAL VECTOR OR MATRIX b - a*x is COMPUTED AND OVERWRITES b.
!
!   THE 2-NORM OF THE RESIDUAL VECTOR(S) FOR THE SOLUTION VECTOR(S) x, CAN ALSO BE COMPUTED
!   DIRECTLY BY SPECIFYING THE OPTIONAL ARRAY ARGUMENT rnorm IN THE CALL OF qr_solve.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK THAT THE RESIDUALS ARE ORTHOGONAL TO THE RANGE OF THE COEFFICIENT MATRIX a .
!
        res(:nrhs,:n) = matmul( transpose( b(:m,:nrhs) ), a2(:m,:n) )
!
        err1 = maxval( abs(res(:nrhs,:n)) )/anorm
!
!       CHECK THE NORMS OF THE COLUMNS OF RESIDUAL MATRIX.
!
        err2 = maxval( abs( norm( b(:m,:nrhs), dim=2_i4b ) - rnorm(:nrhs) ) )
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, res )
!
    end if
!
!   PRINT THE RESULTS OF THE TESTS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*)
    write (prtunit,*) '2-norm of residual vectors ||a*x(:,i)-b(:,i)|| = ', rnorm(:nrhs)
!
    write (prtunit,*)
    write (*,'(a,i5,a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for solving a linear least squares problem with a ', &
      m, ' by ', n,' real coefficient matrix and a ', m, ' by ', nrhs,       &
      ' right hand side matrix is ', elapsed_time, ' seconds'
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, b, x, diagr, beta, rnorm )
!
!
! END OF PROGRAM ex1_qr_solve
! ===========================
!
end program ex1_qr_solve
