program ex1_qr_solve2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine QR_CMP2 in module QR_Procedures
!   and QR_SOLVE2 in module LLSQ_Procedures.
!                                                                              
! LATEST REVISION : 27/01/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, lgl, stnd, zero, true, false, c500, allocate_error, merror, &
                         qr_cmp2, qr_solve2, norm
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT.
! m IS THE NUMBER OF ROWS OF THE RANDOM MATRIX.
! n IS THE NUMBER OF COLUMNS OF THE RANDOM MATRIX.
! nrhs IS THE NUMBER OF COLUMNS OF THE RIGHT HAND SIDE MATRIX.
!
    integer(i4b), parameter :: prtunit=6, m=4000, n=3000, mn=min( m, n ), nrhs=10
!   
    real(stnd), parameter  :: fudge=c500
!
    character(len=*), parameter :: name_proc='Example 1 of qr_solve2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: ulp, eps, tol, anorm, err1, err2, err, elapsed_time
    real(stnd), allocatable, dimension(:)   :: diagr, beta, tau, rnorm
    real(stnd), allocatable, dimension(:,:) :: a, a2, b, x, res
!
    integer(i4b)                            :: krank, j, l, idep
    integer(i4b), allocatable, dimension(:) :: ip
    integer                                 :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: comp_resid, do_test, test_lin
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : SOLVING LINEAR LEAST SQUARES PROBLEMS WITH A m-BY-n REAL COEFFICIENT
!               MATRIX USING A COMPLETE ORTHOGONAL DECOMPOSITION OF THE COEFFICIENT MATRIX
!               AND SEVERAL RIGHT HAND-SIDES. THE COEFFICIENT MATRIX CAN BE RANK DEFICIENT
!               AND BOTH m>=n OR m<n ARE PERMITTED.
!
!               COMPUTE SOLUTION MATRIX x FOR LINEAR LEAST SQUARES SYSTEM
!
!                              a(:m,:n)*x(:n,:nrhs) ≈ b(:m,:nrhs) .
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = fudge*ulp
!
    err = zero
    test_lin = true
!
!   SET TOLERANCE FOR DETERMINING THE RANK OF THE MATRIX IN THE SUBROUTINE.
!
!    tol = 0.0000001_stnd
    tol = sqrt( ulp )
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE SUBROUTINE.
!
    do_test = true
!
!   DECIDE IF COLUMN PIVOTING MUST BE PERFORMED AND
!   IF RESIDUAL MATRIX MUST BE COMPUTED.
!
    krank = 0
!
    comp_resid = do_test
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), b(m,nrhs), x(n,nrhs), diagr(mn), beta(mn), tau(mn),  &
              rnorm(nrhs), ip(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM COEFFICIENT MATRIX a(:m,:n) .
!
    call random_number( a(:m,:n) )
!
!   GENERATE A DEFICIENT COEFFICIENT MATRIX a(:m,:n) WITH RANK min(mn-1,n) .
!
    idep = min( n, 5_i4b ) 
    a(:m,idep) = a(:m,1_i4b) + a(:m,n)
!
!   GENERATE A RANDOM RIGHT HAND-SIDE MATRIX b(:m,:nrhs) .
!
    call random_number( b(:m,:nrhs) )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS FOR LATER USE.
!
        allocate( a2(m,n), res(nrhs,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE FROBENIUS NORM OF THE DATA MATRIX.
!
        anorm = norm( a(:m,:n) )
!
!       MAKE A COPY OF THE DATA MATRIX FOR LATER USE.
!
        a2(:m,:n) = a(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE A COMPLETE ORTHOGONAL DECOMPOSITION OF RANDOM DATA MATRIX WITH SUBROUTINE qr_cmp2.
!
    call qr_cmp2( a(:m,:n), diagr(:mn), beta(:mn), ip(:n), krank, tol=tol, tau=tau(:mn) )
!
!   qr_cmp2 COMPUTES A QR OR COMPLETE ORTHOGONAL FACTORIZATION OF A REAL m-BY-n MATRIX.
!   THE MATRIX MAY BE RANK-DEFICIENT.
!
!   HERE THE ROUTINE COMPUTES A COMPLETE ORTHOGONAL FACTORIZATION OF a.
!
!   A QR FACTORIZATION WITH COLUMN PIVOTING OF a IS FIRST COMPUTED AS:
!
!                     a * P = Q * R = Q * [ R11 R12 ]
!                                         [  0  R22 ]
!
!   P IS A n-BY-n PERMUTATION MATRIX, Q IS A m-BY-m ORTHOGONAL MATRIX AND
!   R IS A m-BY-n UPPER TRIANGULAR OR TRAPEZOIDAL MATRIX.
!   
!   R11 IS DEFINED AS THE LARGEST LEADING SUBMATRIX OF R WHOSE ESTIMATED CONDITION
!   NUMBER  IS LESS THAN 1/tol IF tol IS IN ]0,1[ OR SUCH THAT ABS(R11[j,j])>0 IF
!   tol IS EQUAL TO 0 (SEE BELOW FOR DETAILS).
!   
!   THE ORDER OF R11 IS AN ESTIMATE OF THE RANK OF a AND IS STORED
!   IN THE ARGUMENT krank ON EXIT OF qr_cmp2.
!
!   IN OTHER WORDS, ON EXIT, krank CONTAINS THE EFFECTIVE RANK OF a. I.E., THE NUMBER
!   OF INDEPENDENT COLUMNS IN MATRIX a.
!
!   ON INPUT, IF krank=k, THIS IMPLIES THAT THE FIRST k COLUMNS OF a ARE TO BE FORCED
!   INTO THE BASIS. PIVOTING IS PERFORMED ON THE LAST n-k COLUMNS OF a.
!   
!   WHEN krank>=min(m,n) IS USED, PIVOTING IS NOT PERFORMED. THIS IS
!   APPROPRIATE WHEN a IS KNOWN TO BE FULL RANK. krank<=0 CAN BE USED
!   WHEN PIVOTING IS DESIRED ON ALL COLUMNS OF a.
!                      
!   IF tol IS PRESENT AND IS IN [0,1[, THEN :
!       ON ENTRY, CALCULATIONS TO DETERMINE THE CONDITION NUMBER OF a
!       ( IN THE 1-NORM) ARE PERFORMED. tol IS THEN USED TO DETERMINE THE EFFECTIVE RANK
!       OF a, WHICH IS DEFINED AS THE ORDER OF THE LARGEST LEADING
!       TRIANGULAR SUBMATRIX R11 IN THE QR FACTORIZATION WITH PIVOTING OF a,
!       WHOSE ESTIMATED CONDITION NUMBER IS LESS THAN 1/tol. IF tol=0 IS SPECIFIED
!       THE NUMERICAL RANK OF a IS DETERMINED.
!
!   IF tol IS NOT SPECIFIED OR IS OUTSIDE [0,1[ :
!       THE CALCULATIONS TO DETERMINE THE CONDITION NUMBER OF a ARE NOT
!       PERFORMED AND CRUDE TESTS ON R(j,j) ARE DONE TO DETERMINE
!       THE NUMERICAL RANK OF a.
!
!   FURTHERMORE, IF tol IS PRESENT AND IS IN [0,1[ IN THE CALL OF qr_cmp2, THE CONDITION
!   NUMBER OF a IS COMPUTED IN ALL CASES AND ON EXIT, THE RECIPROCAL OF THE CONDITION NUMBER
!   IS RETURNED IN tol.
!
!   IF IT IS POSSIBLE THAT a MAY NOT BE OF FULL RANK (I.E., CERTAIN COLUMNS OF a ARE
!   LINEAR COMBINATIONS OF OTHER COLUMNS), THEN THE LINEARLY DEPENDENT COLUMNS CAN
!   USUALLY BE DETERMINED BY USING tol=RELATIVE PRECISION OF THE ELEMENTS IN a.
!   IF EACH ELEMENT IS CORRECT TO, SAY, 5 DIGITS THEN tol=0.00001 SHOULD BE USED.
!   
!   IN A SECOND STEP, IF THE OPTIONAL PARAMETER tau IS PRESENT IN THE CALL OF qr_cmp2,
!   THEN R22 IS CONSIDERED TO BE NEGLIGIBLE AND THE SUBMATRIX R12 IS ANNIHILATED BY
!   ORTHOGONAL TRANSFORMATIONS FROM THE RIGHT, ARRIVING AT THE COMPLETE ORTHOGONAL
!   FACTORIZATION:
!  
!                     a * P ≈ Q * [ T11 0 ] * Z
!                                 [  0  0 ]
!   
!   WHERE P IS A n-BY-n PERMUTATION MATRIX, Q IS A m-BY-m ORTHOGONAL MATRIX,
!   T11 IS A krank-BY-krank UPPER TRIANGULAR MATRIX AND Z IS A n-BY-n
!   ORTHOGONAL MATRIX. P AND Q ARE ARE ALREADY COMPUTED IN THE QR FACTORIZATION
!   WITH COLUM PIVOTING.
!
!   ON EXIT, a HAS BEEN OVERWRITTEN BY DETAILS OF ITS QR OR COMPLETE 
!   ORTHOGONAL FACTORIZATION, DEPENDING IF ARGUMENT tau IS ABSENT OR PRESENT.
!    
!   IF THE OPTIONAL PARAMETER tau IS ABSENT, qr_cmp2 COMPUTES ONLY
!   A FULL QR FACTORIZATION WITH COLUMN PIVOTING OF a AND:
!
!   - THE ELEMENTS BELOW AND ON THE DIAGONAL OF THE ARRAY a(:m,:mn) AND THE ARRAY
!     beta(:mn) STORED Q IN FACTORED FORM.
!    
!   - THE ELEMENTS ABOVE THE DIAGONAL OF THE ARRAY a CONTAIN THE
!     CORRESPONDING ELEMENTS OF THE TRIANGULAR OR TRAPEZOIDAL MATRIX R.
!     THE ELEMENTS OF THE DIAGONAL OF R ARE STORED IN THE ARRAY diagr(:mn). 
!
!   - ip STORES THE PERMUTATION MATRIX P IN THE QR DECOMPOSITION OF a.
!     IF ip(j)=k, THEN THE jTH COLUMN OF a*P WAS THE kTH COLUMN OF a.
!     THE MATRIX P IS REPRESENTED IN THE ARRAY ip AS FOLLOWS:
!     IF ip(j) = i THEN THE jTH COLUMN OF P IS THE iTH CANONICAL UNIT VECTOR.
!                      
!   - krank CONTAINS THE EFFECTIVE RANK OF a (E.G., THE RANK OF R11).
!                      
!   ON THE OTHER HAND, IF THE OPTIONAL PARAMETER tau IS PRESENT, qr_cmp2 COMPUTES A
!   COMPLETE ORTHOGONAL FACTORIZATION FROM THE QR FACTORIZATION WITH COLUMN PIVOTING OF a
!   AND:
!                      
!   - THE ELEMENTS BELOW AND ON THE DIAGONAL OF THE ARRAY a(:m,:mn) AND THE ARRAY
!     beta(:mn) STORED Q IN FACTORED FORM.
!                      
!   - THE ELEMENTS ABOVE THE DIAGONAL OF THE ARRAY SECTION a(1:krank,1:krank)
!     CONTAIN THE CORRESPONDING ELEMENTS OF THE TRIANGULAR MATRIX T11. THE ELEMENTS OF
!     THE DIAGONAL OF T11 ARE STORED IN THE ARRAY SECTION diagr(1:krank).
!
!   - ip STORES THE PERMUTATION MATRIX P IN THE COMPLETE DECOMPOSITION OF a.
!     IF ip(j)=k, THEN THE jTH COLUMN OF a*P WAS THE kTH COLUMN OF a.
!     THE MATRIX P IS REPRESENTED IN THE ARRAY ip AS FOLLOWS:
!     IF ip(j) = i THEN THE jTH COLUMN OF P IS THE iTH CANONICAL UNIT VECTOR.
!
!   - THE ORTHOGONAL MATRIX Z IS STORED IN FACTORED FORM IN THE ARRAY SECTIONS
!     a(:krank,krank+1:n) AND tau(:krank).
!                      
!   - krank CONTAINS THE EFFECTIVE RANK OF a, I.E., THE ORDER OF
!     THE SUBMATRIX T11.
!                      
!   THIS COMPLETE ORTHOGONAL FACTORIZATION CAN THEN BE USED TO COMPUTE THE MINIMAL 2-NORM
!   SOLUTIONS FOR RANK DEFICIENT LINEAR LEAST SQUARES PROBLEMS WITH SUBROUTINE qr_solve2.
!                      
!
!   NEXT, COMPUTE THE SOLUTION MATRIX FOR LINEAR LEAST SQUARE SYSTEM
!
!                   a(:m,:n)*x(:n,:nrhs) ≈ b(:m,:nrhs) .
!
!   WITH SUBROUTINE qr_solve2 AND THE COMPLETE ORTHOGONAL DECOMPOSITION COMPUTED BY qr_cmp2.
!
    call qr_solve2( a(:m,:n), diagr(:mn), beta(:mn), ip(:n), krank, b(:m,:nrhs), x(:n,:nrhs),  &
                    tau=tau(:mn), rnorm=rnorm(:nrhs), comp_resid=comp_resid )
!
!   qr_solve2 COMPUTES SOLUTION(S) TO REAL LINEAR LEAST SQUARES PROBLEMS 
!   OF THE FORM:
!
!                       Minimize || b - a*x ||_2
!
!   USING A QR FACTORIZATION WITH COLUMNS PIVOTING OR A COMPLETE
!   ORTHOGONAL FACTORIZATION OF a COMPUTED BY qr_cmp2. a IS AN m-BY-n MATRIX
!   WHICH MAY BE RANK-DEFICIENT. m>=n OR n>m IS PERMITTED.
!
!   SEVERAL RIGHT HAND SIDE VECTORS b AND SOLUTION VECTORS x CAN BE
!   HANDLED IN A SINGLE CALL; THEY ARE STORED AS THE COLUMNS OF THE
!   m-BY-nrhs RIGHT HAND SIDE MATRIX b AND THE n-BY-nrhs SOLUTION
!   MATRIX x, RESPECTIVELY. IN OTHER WORDS, THE ARGUMENTS b AND x MAY
!   BE VECTORS OR MATRICES. b IS OVERWRITTEN BY qr_solve2.
!   
!   ON EXIT, IF THE OPTIONAL LOGICAL PARAMETER comp_resid IS PRESENT AND IS SET TO true,
!   THE RESIDUAL VECTOR OR MATRIX b - a*x is COMPUTED AND OVERWRITES b.
!
!   THE 2-NORM OF THE RESIDUAL VECTOR(S) FOR THE SOLUTION VECTOR(S) x, CAN ALSO BE COMPUTED
!   DIRECTLY BY SPECIFYING THE OPTIONAL ARRAY ARGUMENT rnorm IN THE CALL OF qr_solve2.
!
!   THE MINIMUN 2-NORM SOLUTION(S) ARE COMPUTED IF THE OPTIONAL ARRAY PARAMETER tau IS
!   PRESENT IN BOTH THE CALLS OF qr_cmp2 AND qr_solve2 SUBROUTINES. OTHERWISE, SOLUTION(S) ARE
!   COMPUTED SUCH THAT IF THE jTH COLUMN OF a IS OMITTED FROM THE BASIS, x[j] O
!   x[j,:nrhs] IS SET TO ZERO.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   CHECK IF THE DEPENDENT COLUMNS IN MATRIX a HAVE BEEN IDENTIFIED CORRECTLY.
!
    if ( m>=n .and. idep<n ) then
!
        do l = krank+1_i4b, n
!
            j = ip(l)
!
            if ( j==idep .or. j==1_i4b .or. j==n ) exit
!
        end do
!
        test_lin = l/=n+1_i4b
!
    end if
!
    if ( do_test ) then
!
!       CHECK THAT THE RESIDUALS ARE ORTHOGONAL TO THE RANGE OF THE COEFFICIENT MATRIX a .
!
        res(:nrhs,:n) = matmul( transpose( b(:m,:nrhs) ), a2(:m,:n) )
!
        err1 = maxval( abs(res(:nrhs,:n)) )/anorm
!
!       CHECK THE NORMS OF THE COLUMNS OF RESIDUAL MATRIX.
!
        err2 = maxval( abs( norm( b(:m,:nrhs), dim=2_i4b ) - rnorm(:nrhs) ) )
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, res )
!
    end if
!
!   PRINT THE RESULTS OF THE TESTS.
!
    if ( err<=eps .and. test_lin ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*)
    write (prtunit,*) 'Estimated rank of the coefficient matrix = ', krank
!
    if ( krank/=mn ) then
        write (prtunit,*) 'Indices of linearly dependent columns    = ', ip(krank+1:n)
    end if
    write (prtunit,*) '2-norm of residual vectors ||a*x(:,i)-b(:,i)|| = ', rnorm(:nrhs)
!
    write (prtunit,*)
    write (*,'(a,i5,a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for solving a linear least squares problem with a ', &
      m, ' by ', n,' real coefficient matrix and a ', m, ' by ', nrhs,       &
      ' right hand side matrix is ', elapsed_time, ' seconds'
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, b, x, diagr, beta, tau, rnorm, ip )
!
!
! END OF PROGRAM ex1_qr_solve2
! ============================
!
end program ex1_qr_solve2
