program ex1_random_svd_fixed_precision_with_blas
!
!
! Purpose
! =======
!
!   This program illustrates how to compute an approximate reduced SVD with randomized
!   subspace iterations, which fullfills a given relative error in Frobenius norm using
!   STATPACK and BLAS subroutines.
!                                                                              
!                                                     
! Further Details
! ===============
!
!   The program shows the use of subroutines SVD_CMP in module SVD_Procedures,
!   QR_CMP and ORTHO_GEN_QR in module QR_Procedures, NORMAL_RANDOM_NUMBER3_ and
!   GEN_RANDOM_MAT in module Random and GEMM generic interface in module BLAS_interfaces.
!    
!                                                                              
! LATEST REVISION : 23/10/2020
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, two, three, seven, c30, allocate_error,  &
                         merror, norm, unit_matrix, random_seed_, normal_random_number3_, triang_solve,   &
                         singval_sort, qr_cmp, ortho_gen_qr, svd_cmp, gen_random_mat
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE GENERATED MATRIX,
! nsvd0 IS THE RANK OF THE GENERATED MATRIX,
! relerr0 IS THE REQUESTED RELATIVE ERROR OF THE RESTRICTED SVD IN FROBENIUS NORM.
!
    integer(i4b), parameter :: prtunit=6, m=2000, n=2000, mn=min(m,n), nsvd0=1000
!   
    real(stnd), parameter  :: relerr0=0.10_stnd
!
    character(len=*), parameter :: name_proc='Example 1 of random_svd_fixed_precision_with_blas'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: err, err1, err2, err3, eps, elapsed_time, threshold, &
                  norma, erra, relerr, relerr2, tmp
#ifdef _BLAS
    real(stnd), dimension(:,:), allocatable :: a, q, b, h, r, r2, ti, ti2, yi, bi, v, leftvec, rightvec
#else
    real(stnd), dimension(:,:), allocatable :: a, q, qt, b, bt, r, r2, ti, ti2, yi, bi, leftvec, rightvec
#endif
    real(stnd), dimension(:),   allocatable :: diagr, beta, s, s0
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: np, blksz, niter, maxiter, i, i0, i1, i2, j, nsvd, mat_type
!
    logical(lgl) :: failure_qb, failure_svd, do_test, ortho, reortho
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PARTIAL SVD OF A m-BY-n REAL MATRIX USING RANDOMIZED
!               POWER SUBSPACE ITERATIONS.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type > 3  -> VERY SLOW DECAY OF SINGULAR VALUES
!
    mat_type = 2_i4b
!
!   SET THE TOLERANCE FOR THE ERROR OF THE RELATIVE ERROR IN FROBENIUS NORM.
!
    eps = 0.01_stnd
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF SINGULAR TRIPLETS.
!
    do_test = true
!
!   CHOOSE TUNING PARAMETERS FOR THE QB ALGORITHM.
!   blksz*maxiter IS THE MAXIMUM ALLOWABLE RANK OF THE
!   PARTIAL SVD, WHICH IS SOUGHT.
!
    blksz   = 10_i4b
    maxiter = 20_i4b
!
    tmp = real( mn, stnd )/(three*real( blksz, stnd ))
    maxiter = min( maxiter, int( tmp, i4b ) )
!
    np = blksz*maxiter
!
!   DETERMINE THE NUMBER OF POWER ITERATIONS niter TO BE PERFORMED.
!
    niter = 1_i4b
!
!   SPECIFY IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE (POWER) SUBSPACE ITERATIONS, TO AVOID LOSS OF ACCURACY DUE
!   TO ROUNDING ERRORS.
!
    ortho = true
!
!   SPECIFY IF REORTHONORMALIZATION IS CARRIED OUT TO AVOID LOSS OF ORTHOGONALITY
!   IN THE BLOCK GRAM-SCHMIDT ORTHOGONALISATION SCHEME USED TO BUILD THE ORTHOGONAL MATRIX
!   OF THE QB DECOMPOSITION OF THE INPUT MATRIX.
!
    reortho = false
!
!   ALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    allocate( a(m,n), q(m,np), b(n,np), diagr(mn), beta(mn), s0(nsvd0),  &
              h(n,np), r(blksz,blksz), yi(m,blksz), ti(np,blksz),        &
              bi(blksz,n),  stat=iok     )
#else
    allocate( a(m,n), q(m,np), qt(np,m), b(n,np), bt(np,n), diagr(mn),        &
              beta(mn), s0(nsvd0), r(blksz,blksz), yi(m,blksz), ti(np,blksz), &
              bi(blksz,n),  stat=iok      )
#endif
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    if ( reortho ) then
!
        allocate( r2(blksz,blksz), ti2(np,blksz), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
    end if
!
!   GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A SPECIFIED DISTRIBUTION
!   OF SINGULAR VALUES.
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE SINGULAR VALUES.
!
    select case( mat_type )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                s0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                s0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                s0(i) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
            end do
!
        case default
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            norma = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i - 1_i4b, stnd )
!
                s0(i) = exp( -tmp/norma )
!        
            end do
!
    end select
!
!   SORT THE SINGULAR VALUES.
!
    call singval_sort( sort, s0(:nsvd0) )
!
!   GENERATE A m-BY-n MATRIX a OF RANK nsvd0 .
!
    call gen_random_mat( s0(:nsvd0), a )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   TEST VALIDITY OF THE PRESET RELATIVE ACCURACY TOLERANCE relerr .
!
    tmp = two*sqrt( epsilon( relerr0 )/relerr0 )
    relerr = max( tmp, relerr0 )
!
!   COMPUTE THE FROBENIUS NORM OF THE INPUT MATRIX AND THE REQUIRED THRESHOLD.
!
!    norma = norm( a(1_i4b:m,1_i4b:n) )
    norma = sqrt(sum( s0(:nsvd0)**2 ) )
!
    erra = norma*norma
!
    threshold = erra*(relerr*relerr)
!
!        write (prtunit,*) erra, threshold
!
!   GENERATE A n-BY-np IID GAUSSIAN MATRIX .
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING MATRIX
!   FORM OF SUBROUTINE normal_random_number3_.
!
    call normal_random_number3_( b(:n,:np) )
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING VECTOR
!   FORM OF SUBROUTINE normal_random_number_.
!
!    do i = 1_i4b, np
!
!        call normal_random_number3_( b(:n,i) )
!
!    end do
!
!   COMPUTE RANDOM SAMPLE MATRIX.
!
#ifdef _BLAS
    call gemm( 'N', 'N', m, np, n, one, a(1_i4b:m,1_i4b:n), m,       &
               b(1_i4b:n,1_i4b:np), n, zero, q(1_i4b:m,1_i4b:np), m  )
#else
    q(:m,:np) = matmul( a(:m,:n), b(:n,:np) )
#endif
!
!   DO SUBSPACE ITERATIONS.
!
    do i = 1_i4b, niter
!
        if ( ortho ) then
!
!           COMPUTE QR DECOMPOSITION OF RANDOM MATRIX TO OBTAIN AN
!           ORTHONORMAL BASIS OF RANDOM SUBSPACE.
!
            call qr_cmp( q(:m,:np), diagr(:np), beta(:np) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( q(:m,:np), beta(:np) )
!
        end if
!
!       COMPUTE RANDOM SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'T', 'N', n, np, m, one, a(1_i4b:m,1_i4b:n), m,       &
                   q(1_i4b:m,1_i4b:np), m, zero, b(1_i4b:n,1_i4b:np), n  )
#else
        qt(:np,:m) = transpose( q(:m,:np) )
!
        bt(:np,:n) = matmul( qt(:np,:m), a(:m,:n) )
!
        b(:n,:np) = transpose( bt(:np,:n) )
#endif
!
        if ( ortho ) then
!
!           COMPUTE QR DECOMPOSITION OF RANDOM MATRIX TO OBTAIN AN
!           ORTHONORMAL BASIS OF RANDOM SUBSPACE.
!
            call qr_cmp( b(:n,:np), diagr(:np), beta(:np) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( b(:n,:np), beta(:np) )
!
        end if
!
!       COMPUTE RANDOM SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, np, n, one, a(1_i4b:m,1_i4b:n), m,       &
                   b(1_i4b:n,1_i4b:np), n, zero, q(1_i4b:m,1_i4b:np), m  )
#else
        q(:m,:np) = matmul( a(:m,:n), b(:n,:np) )
#endif
!
    end do
!
#ifdef _BLAS
    call gemm( 'T', 'N', n, np, m, one, a(1_i4b:m,1_i4b:n), m,       &
               q(1_i4b:m,1_i4b:np), m, zero, h(1_i4b:n,1_i4b:np), n  )
#else
    qt(:np,:m) = transpose( q(:m,:np) )
!
    bt(:np,:n) = matmul( qt(:np,:m), a(:m,:n) )
#endif
!
    i0   = 0_i4b
    nsvd = 0_i4b
!
    r(:blksz,:blksz)         = zero
    ti(1_i4b:np,1_i4b:blksz) = zero
!
    if ( reortho ) then
!
        r2(:blksz,:blksz) = zero
        ti2(1_i4b:np,1_i4b:blksz) = zero
!
    end if
!
    failure_qb = true
!
!   COMPUTE THE QB FACTORIZATION.
!
    iter : do i = 1_i4b, maxiter
!
        i1 = i0 + 1_i4b
        i2 = i0 + blksz
!
        if ( i==1_i4b ) then
!
            yi(:m,:blksz) = q(:m,i1:i2)
!
        else
!
!           PERFORM BLOCK GRAM-SCHMIDT ORTHOGONALISATION STEP.
!
#ifdef _BLAS
            call gemm( 'T', 'N', i0, blksz, n, one, b(1_i4b:n,1_i4b:i0), n,     &
                       b(1_i4b:n,i1:i2), n, zero, ti(1_i4b:np,1_i4b:blksz), np  )
!
            call gemm( 'N', 'N', m, blksz, i0, -one, q(1_i4b:m,1_i4b:i0), m,   &
                       ti(1_i4b:np,1_i4b:blksz), np, one, q(1_i4b:m,i1:i2), m  )
#else
            ti(:i0,:blksz) = matmul( bt(:i0,:n), b(:n,i1:i2) )
!
            q(:m,i1:i2) = q(:m,i1:i2) - matmul( q(:m,:i0), ti(:i0,:blksz) )
#endif
!            
            yi(:m,:blksz) = q(:m,i1:i2)
!
        end if
!
        call qr_cmp( q(:m,i1:i2), diagr(:blksz), beta(:blksz) )
!
!       SAVE UPPER TRIANGULAR MATRIX FOR LATER USE.
!
        if ( i==1_i4b .or. .not.reortho ) then
!
            do j = 1, blksz
!
                r(1_i4b:j-1_i4b,j) = q(1_i4b:j-1_i4b,i0+j)
                r(j,j)             = diagr(j)
!              r(j+1_i4b:blksz,j) = zero
!
            end do
!
        else
!
            do j = 1, blksz
!
                r2(1_i4b:j-1_i4b,j) = q(1_i4b:j-1_i4b,i0+j)
                r2(j,j)             = diagr(j)
!              r2(j+1_i4b:blksz,j) = zero
!
            end do
!
        end if
!
!       GENERATE COLUMNS i1 TO i2 OF THE ORTHOGONAL MATRIX.
!
        call ortho_gen_qr( q(:m,i1:i2), beta(:blksz) )
!
!       SAVE UPPER TRIANGULAR MATRIX FOR LATER USE.
!
        if ( i/=1_i4b ) then
!
            if ( reortho ) then
!
!               REORTHONORMALIZATION IS CARRIED OUT TO AVOID LOSS OF ORTHOGONALITY
!               IN THE BLOCK GRAM-SCHMIDT ORTHOGONALISATION SCHEME USED TO BUILD
!               THE ORTHOGONAL MATRIX OF THE QB DECOMPOSITION.
!
#ifdef _BLAS
                call gemm( 'T', 'N', i0, blksz, m, one, q(1_i4b:m,1_i4b:i0), m,     &
                           q(1_i4b:m,i1:i2), m, zero, ti2(1_i4b:np,1_i4b:blksz), np )
!
                call gemm( 'N', 'N', m, blksz, i0, -one, q(1_i4b:m,1_i4b:i0), m,    &
                           ti2(1_i4b:np,1_i4b:blksz), np, one, q(1_i4b:m,i1:i2), m  )
#else
                ti2(:i0,:blksz) = matmul( qt(:i0,:m), q(:m,i1:i2) )
!
                q(:m,i1:i2) = q(:m,i1:i2) - matmul( q(:m,:i0), ti2(:i0,:blksz) )
#endif
!
!               IMPROVE ESTIMATES OF COLUMNS i1 TO i2 OF THE ORTHOGONAL MATRIX.
!
                call qr_cmp( q(:m,i1:i2), diagr(:blksz), beta(:blksz) )
!
!               SAVE UPPER TRIANGULAR MATRIX FOR LATER USE.
!
                do j = 1, blksz
!
                    bi(1_i4b:j-1_i4b,j) = q(1_i4b:j-1_i4b,i0+j)
                    bi(j,j)             = diagr(j)
                    bi(j+1_i4b:blksz,j) = zero
!
                end do
!
!               UPDATE COLUMNS i1 TO i2 OF THE ORTHOGONAL MATRIX.
!
                call ortho_gen_qr( q(:m,i1:i2), beta(:blksz) )
!
#ifdef _BLAS
                call gemm( 'N', 'N', blksz, blksz, blksz, one, bi(1_i4b:blksz,1_i4b:blksz), blksz,     &
                           r2(1_i4b:blksz,1_i4b:blksz), blksz, zero, r(1_i4b:blksz,1_i4b:blksz), blksz )
!
                call gemm( 'T', 'N', i0, blksz, m, one, q(1_i4b:m,1_i4b:i0), m,          &
                           yi(1_i4b:m,1_i4b:blksz), m, one, ti(1_i4b:np,1_i4b:blksz), np )
#else
                r(:blksz,:blksz) = matmul( bi(:blksz,:blksz), r2(:blksz,:blksz) )
!
                ti(:i0,:blksz) = ti(:i0,:blksz) + matmul( qt(:i0,:m), yi(:m,:blksz) )
#endif
!
            end if
!
#ifdef _BLAS
!
            call gemm( 'N', 'N', n, blksz, i0, -one, b(1_i4b:n,1_i4b:i0), n,  &
                       ti(1_i4b:np,1_i4b:blksz), np, one, h(1_i4b:n,i1:i2), n )
#else
            bi(:blksz,:i0) = transpose( ti(:i0,:blksz) )
            bt(i1:i2,:n)   = bt(i1:i2,:n) - matmul( bi(:blksz,:i0), bt(:i0,:n) )
#endif
!
        end if
!
!       COMPUTE ROWS i1 TO i2 OF THE B FACTOR OF QB DECOMPOSITION.
!
#ifdef _BLAS
        bi(:blksz,:n) = transpose( h(:n,i1:i2) )
!
        call triang_solve( r(:blksz,:blksz), bi(:blksz,:n), upper=true, trans=true )
!
        b(:n,i1:i2)  = transpose( bi(:blksz,:n) )
!
#else
        bi(:blksz,:n) = bt(i1:i2,:n)
!
        call triang_solve( r(:blksz,:blksz), bi(:blksz,:n), upper=true, trans=true )
!
        bt(i1:i2,:n) = bi(:blksz,:n)
        qt(i1:i2,:m) = transpose( q(:m,i1:i2) )
#endif
!
!       COMPUTE CURRENT ERROR OF THE QB FACTORIZATION.
!
        tmp = norm( bi(:blksz,:n) )
        tmp = erra - tmp*tmp
!
!       CHECK IF THE THRESHOLD IS SATISFIED WITH INCLUSION OF THE CURRENT BLOCK.
!
        if ( tmp<threshold  ) then
!
            failure_qb = false
!
!           DETREMINE THE PRECISE RANK OF THE ORTHOGONAL MATRIX OF THE QB DECOMPOSITION.
!
            do j = 1, blksz
!
#ifdef _BLAS
                tmp = norm( b(:n,i0+j) )
#else
                tmp = norm( bi(j,:n) )
#endif
                erra = erra - tmp*tmp
!
                if ( erra<threshold  ) then
!
!                   nsvd IS THE NUMBER OF COLUMNS OF THE ORTHOGONAL MATRIX OF THE QB DECOMPOSITION.
!
                    nsvd = i0 + j
!
                    exit iter
!
                end if
!
            end do
!
        else
!
            erra = tmp
!
        end if
!
!        write (prtunit,*) erra, threshold
!
        i0 = i0 + blksz
!
    end do iter
!
    if ( failure_qb ) then
!
        nsvd = np
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    deallocate( h, r, yi, ti, bi, diagr )
#else
    deallocate( qt, r, yi, ti, bi, diagr )
#endif
!
    if ( reortho ) then
!
        deallocate(  r2, ti2 )
!
    end if
!
!   ALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    allocate( v(nsvd,nsvd), s(nsvd), leftvec(m,nsvd), rightvec(n,nsvd), stat=iok )
#else
    allocate( s(nsvd), leftvec(m,nsvd), rightvec(n,nsvd), stat=iok )
#endif
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   COMPUTE SVD OF THE FINAL QB FACTORIZATION.
!
#ifdef _BLAS
    call svd_cmp( b(:n,:nsvd), s(:nsvd), failure_svd, v=v(:nsvd,:nsvd), sort=sort, max_francis_steps=10_i4b )
!
!   COMPUTE THE APPROXIMATE TOP nsvd LEFT SINGULAR VECTORS OF a .
!
    call gemm( 'N', 'N', m, nsvd, nsvd, one, q(1_i4b:m,1_i4b:nsvd), m,               &
               v(1_i4b:nsvd,1_i4b:nsvd), nsvd, zero, leftvec(1_i4b:m,1_i4b:nsvd), m  )
!
!   EXTRACT THE APPROXIMATE TOP nsvd RIGHT SINGULAR VECTORS OF a .
!
    rightvec(:n,:nsvd) = b(:n,:nsvd)
#else
    call svd_cmp( bt(:nsvd,:n), s(:nsvd), failure_svd, v=rightvec(:n,:nsvd), sort=sort, max_francis_steps=10_i4b )
!
!   COMPUTE THE APPROXIMATE TOP nsvd LEFT SINGULAR VECTORS OF a .
!
    leftvec(:m,:nsvd) = matmul( q(:m,:nsvd), bt(:nsvd,:nsvd) )
#endif
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   COMPUTE THE ESTIMATED RELATIVE ERROR.
!
    relerr = sqrt( erra )/norma
!
!   COMPUTE THE TRUE RELATIVE ERROR.
!
    relerr2 = sqrt(sum( (s0(nsvd+1_i4b:nsvd0)/norma)**2 ))
!
!   COMPUTE ERROR BETWEEN THE TRUE AND ESTIMATED RELATIVE ERRORS.
!
    err = abs( relerr - relerr2 )
!
!   TEST ACCURACY OF SINGULAR TRIPLETS IF REQUIRED.
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v(:m,:nsvd) - u(:n,:nsvd)*diag(s(:nsvd)),
!       WHERE u AND v ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
        q(:m,:nsvd) = leftvec(:m,:nsvd)*spread( s(:nsvd), dim=1, ncopies=m)
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsvd, n, one, a(1_i4b:m,1_i4b:n), m,                &
                   rightvec(1_i4b:n,1_i4b:nsvd), n, -one, q(1_i4b:m,1_i4b:nsvd), m  )
#else
        q(:m,:nsvd) = matmul( a(:m,:n), rightvec(:n,:nsvd) ) - q(:m,:nsvd)
#endif
!
        beta(:nsvd) = norm( q(:m,:nsvd), dim=2_i4b )
!
        if ( norma==zero ) then
            norma = one
        end if
!
        err1 = maxval( beta(:nsvd) )/( norma*real(mn,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u(:n,:nsvd)**(t)*u(:n,:nsvd).
!
        call unit_matrix( q(:nsvd,:nsvd) )
!
        b(:nsvd,:nsvd) = abs( q(:nsvd,:nsvd) - matmul( transpose(leftvec), leftvec ) )
!
        err2 = maxval( b(:nsvd,:nsvd) )/real(m,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v(:m,:nsvd)**(t)*v(:m,:nsvd).
!
        b(:nsvd,:nsvd) = abs( q(:nsvd,:nsvd) - matmul( transpose(rightvec), rightvec ) )
!
        err3 = maxval( b(:nsvd,:nsvd) )/real(n,stnd)
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    deallocate( a, q, b, v, leftvec, rightvec, beta, s, s0 )
#else
    deallocate( a, q, b, bt, leftvec, rightvec, beta, s, s0 )
#endif
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<eps .and. .not.failure_svd ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( failure_qb ) then
!
        write (prtunit,*) 
        write (prtunit,*) 'Fail to converge within ', maxiter,   &
                          ' iterations! ||A-rSVD||_F / ||A||_F = ', relerr, ' >= ', relerr0
        write (prtunit,*) 
!
    else
!
        write (prtunit,*) 
        write (prtunit,*) 'Converge with ', i ,' iterations! ||A-rSVD||_F / ||A||_F = ', &
                          relerr, ' < ', relerr0
        write (prtunit,*) 
!
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing approximate ', nsvd, ' singular values and vectors of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_random_svd_fixed_precision_with_blas
! =======================================================
!
end program ex1_random_svd_fixed_precision_with_blas
