program ex1_rqb_cmp
!
!
! Purpose
! =======
!
!   This program illustrates how to compute a randomized partial QB or QR factorization of a matrix
!   using subroutine RQB_CMP in module Random.
!                                                                              
!                                                                              
! LATEST REVISION : 05/02/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, seven, ten, c30, merror, allocate_error,  &
                         norm, unit_matrix, random_seed_, singval_sort, rqb_cmp, gen_random_mat
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE GENERATED MATRIX,
! nsvd0 IS THE NUMERICAL RANK OF THE GENERATED MATRIX,
! nqb IS THE TARGET RANK OF THE RANDOMIZED PARTIAL QB OR QR FACTORIZATION, WHICH IS SOUGHT.
!
    integer(i4b), parameter :: prtunit=6, m=10000, n=10000, nsvd0=1000, nqb=5
!
    character(len=*), parameter :: name_proc='Example 1 of rqb_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, eps, elapsed_time, norma, normb, &
                                               tmp, relerr, relerr2
    real(stnd), dimension(:,:), allocatable :: a, q, b, res, id
    real(stnd), dimension(:),   allocatable :: singval0
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: niter, i, mat_type
    integer(i4b), allocatable, dimension(:) :: ip
!
    logical(lgl) :: ortho, comp_qr, pivoting, do_test
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PARTIAL QB OR QR FACTORIZATION OF A m-BY-n REAL MATRIX USING A RANDOMIZED
!               ALGORITHM AS
!
!                              a(:m,:n) ≈ q(:m,:nqb)*b(:nqb,:n)
!
!               WHERE q IS A m-BY-nqb MATRIX WITH ORTHONORMAL COLUMNS, b IS A nqb-BY-n MATRIX AND
!               THE PRODUCT q*b IS A GOOD APPROXIMATION OF a ACCORDING TO THE SPECTRAL OR FROBENIUS
!               NORM.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type > 3  -> VERY SLOW DECAY OF SINGULAR VALUES
!
    mat_type = 2_i4b
!
!   SET TOLERANCE FOR THE ERROR OF THE RELATIVE ERROR IN FROBENIUS NORM
!   OF THE COMPUTED PARTIAL QB OR QR FACTORIZATION.
!
    eps = 0.05_stnd
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ORTHOGONALITY OF THE q MATRIX.
!
    do_test = true
!
!   CHOOSE TUNING PARAMETERS FOR THE RANDOMIZED QB OR QR ALGORITHM.
!
!   DETERMINE THE NUMBER OF SUBSPACE ITERATIONS niter TO BE PERFORMED.
!
    niter = 2_i4b
!
!   DETERMINE IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE (POWER) SUBSPACE ITERATIONS, TO AVOID LOSS
!   OF ACCURACY DUE TO ROUNDING ERRORS.
!
    ortho = true
!
!   DETERMINE IF A QB OR QR FACTORIZATION IS COMPUTED.
!
    comp_qr = true
!
!   DETERMINE IF A QR FACTORIZATION WITH COLUMN PIVOTING IS COMPUTED.
!
    pivoting = true
!
!   ALLOCATE WORK ARRAYS.
!
    if ( pivoting ) then
!
        allocate( a(m,n), q(m,nqb), b(nqb,n), ip(n), singval0(nsvd0), stat=iok )
!
    else
!
        allocate( a(m,n), q(m,nqb), b(nqb,n), singval0(nsvd0), stat=iok )
!
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A SPECIFIED DISTRIBUTION
!   OF SINGULAR VALUES.
!
!   GENERATE SINGULAR VALUES.
!
    select case( mat_type )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = 0.0001_stnd + one/( one + exp( tmp - c30 ) )
!        
            end do
!
        case default
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            norma = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i - 1_i4b, stnd )
!
                singval0(i) = exp( -tmp/norma )
!        
            end do
!
    end select
!
!   SORT THE SINGULAR VALUES.
!
    call singval_sort( sort, singval0(:nsvd0) )
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE A m-BY-n MATRIX a OF RANK nsvd0 .
!
    call gen_random_mat( singval0(:nsvd0), a )
!
!   COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
!    norma = norm( a(:m,:n) )
    norma = sqrt(sum( singval0(:nsvd0)**2 ) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   rqb_cmp COMPUTES AN APPROXIMATE PARTIAL QB OR QR DECOMPOSITION OF A REAL m-BY-n MATRIX a
!   USING A RANDOMIZED ALGORITHM AND SUBSPACE ITERATIONS. THE PARTIAL QB OR QR DECOMPOSITION
!   IS WRITTEN
!
!                  a(:m,:n) ≈ q(:m,:nqb)*b(:nqb,:n)
!
!   WHERE q IS A m-BY-nqb MATRIX WITH ORTHONORMAL COLUMNS, b IS A nqb-BY-n (FULL OR TRAPEZOIDAL)
!   MATRIX AND THE PRODUCT q*b IS A GOOD APPROXIMATION OF a ACCORDING TO THE SPECTRAL OR FROBENIUS
!   NORM.
!
    if ( pivoting ) then
!
        call rqb_cmp( a(:m,:n), q(:m,:nqb), b(:nqb,:n), niter=niter, ortho=ortho, ip=ip(:n) )
!
    else
!
        call rqb_cmp( a(:m,:n), q(:m,:nqb), b(:nqb,:n), niter=niter, ortho=ortho, comp_qr=comp_qr )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   COMPUTE THE ESTIMATED RELATIVE ERROR OF THE QB OR QR DECOMPOSITION.
!
    normb = norm( b(:nqb,:n) )
!
    tmp = one - (normb/norma)**2
    relerr = sqrt( max( tmp, zero ) )
!
!   COMPUTE THE BEST RELATIVE ERROR FROM THE SVD.
!
    relerr2 = sqrt(sum( (singval0(nqb+1_i4b:nsvd0)/norma)**2 ))
!
!   COMPUTE ERROR BETWEEN THE BEST AND ESTIMATED RELATIVE ERRORS.
!
    err = abs( relerr - relerr2 )
!
!   TEST ACCURACY OF THE ORTHONORMAL MATRIX Q IF REQUIRED.
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( res(nqb,nqb), id(nqb,nqb), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q(:m,:nqb)**(t)*q(:m,:nqb).
!
        call unit_matrix( id(:nqb,:nqb) )
!
        res(:nqb,:nqb) = abs( id(:nqb,:nqb) - matmul( transpose(q(:m,:nqb)), q(:m,:nqb) ) )
!
        err1 = maxval( res(:nqb,:nqb) )/real(m,stnd)
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( res, id )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( pivoting ) then
        deallocate( a, singval0, q, b, ip )
    else
        deallocate( a, singval0, q, b )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = 1  -> slow decay of singular values'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of singular values'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of singular values'
    write (prtunit,*) ' Matrix type > 3  -> very slow decay of singular values'
!
    write (prtunit,*) 
    write (prtunit,*) 'Rank of the approximate partial QB or QR decomposition              = ', &
                      nqb
    write (prtunit,*) 'Relative error in Frobenius norm     : ||A-Q*B||_F / ||A||_F        = ', &
                      relerr
    write (prtunit,*) 'Best relative error in Frobenius norm: min( ||A-Q*B||_F / ||A||_F ) = ', &
                      relerr2
    write (prtunit,*) 
!
    if ( do_test ) then
        write (prtunit,*)
        write (prtunit,*) 'Orthogonality of the computed orthonormal matrix Q = ', err1
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing a randomized partial QB or QR factorization of rank ', nqb, ' of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_rqb_cmp
! ==========================
!
end program ex1_rqb_cmp
