program ex1_rqb_cmp_fixed_precision
!
!
! Purpose
! =======
!
!   This program illustrates how to compute a randomized partial QB or QR factorization of a matrix,
!   which fullfills a given relative error in Frobenius norm using subroutine RQB_CMP_FIXED_PRECISION
!   in module Random.
!                                                                              
!                                                                              
! LATEST REVISION : 18/03/2022
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, two, three, seven, c30, allocate_error, &
                         merror, norm, unit_matrix, random_seed_, singval_sort, gen_random_mat,          &
                         rqb_cmp_fixed_precision
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE GENERATED MATRIX,
! nsvd0 IS THE NUMERICAL RANK OF THE GENERATED MATRIX,
! relerr0 IS THE REQUESTED TOLERANCE FOR THE RELATIVE ERROR OF THE PARTIAL QB OR QR FACTORIZATION IN FROBENIUS NORM.
!
    integer(i4b), parameter :: prtunit=6, m=10000, n=5000, nsvd0=1000
!   
    real(stnd), parameter  :: relerr0=0.5_stnd
!
    character(len=*), parameter :: name_proc='Example 1 of rqb_cmp_fixed_precision'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, eps, elapsed_time, norma, &
                                               tmp, relerr, relerr2
    real(stnd), dimension(:,:), allocatable :: a, res, id
    real(stnd), dimension(:),   allocatable :: singval0
!
    real(stnd), dimension(:,:), pointer :: q, b
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: blk_size, niter, niter_qb, maxiter_qb, nqb, i, mat_type
    integer(i4b), allocatable, dimension(:) :: ip
!
    logical(lgl) :: ortho, reortho, comp_qr, pivoting, failure_relerr, do_test
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PARTIAL QB OR QR FACTORIZATION OF RANK nqb OF A m-BY-n REAL MATRIX USING A RANDOMIZED
!               ALGORITHM AS
!
!                              a(:m,:n) ≈ q(:m,:nqb)*b(:nqb,:n)
!
!               WHERE q IS A m-BY-nqb MATRIX WITH ORTHONORMAL COLUMNS, b IS A nqb-BY-n MATRIX AND
!               THE PRODUCT q*b IS A GOOD APPROXIMATION OF a ACCORDING TO THE SPECTRAL OR FROBENIUS
!               NORM. The RANK nqb IS DETERMINED SUCH THAT THE ASSOCIATED QB OR QR APPROXIMATION FULLFILLS
!               A PRESCRIBED RELATIVE ERROR IN THE FROBENIUS NORM AND IS NOT KNOWN IN ADVANCE.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type > 3  -> VERY SLOW DECAY OF SINGULAR VALUES
!
    mat_type = 2_i4b
!
!   SET TOLERANCE FOR THE ERROR OF THE RELATIVE ERROR IN FROBENIUS NORM
!   OF THE COMPUTED PARTIAL QB OR QR FACTORIZATION.
!
    eps = 0.05_stnd
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ORTHOGONALITY OF THE q MATRIX.
!
    do_test = true
!
!   CHOOSE TUNING PARAMETERS FOR THE RANDOMIZED QB OR QR ALGORITHM.
!
!   blk_size*maxiter_qb IS THE MAXIMUM ALLOWABLE RANK OF THE
!   PARTIAL QB OR QR FACTORIZATION, WHICH IS SOUGHT.
!
    blk_size   = 10_i4b
    maxiter_qb = 20_i4b
!
!   DETERMINE THE NUMBER OF POWER OR SUBSPACE ITERATIONS niter TO BE PERFORMED
!   IN THE FIRST STEP OF THE QB OR QR FACTORIZATION.
!
    niter = 1_i4b
!
!   SPECIFY IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE (POWER) SUBSPACE ITERATIONS, TO AVOID LOSS OF ACCURACY DUE
!   TO ROUNDING ERRORS.
!
    ortho = true
!
!   SPECIFY IF REORTHONORMALIZATION IS CARRIED OUT TO AVOID LOSS OF ORTHOGONALITY
!   IN THE BLOCK GRAM-SCHMIDT ORTHOGONALISATION SCHEME USED TO BUILD THE ORTHONORMAL
!   MATRIX OF THE QB OR QR DECOMPOSITION OF THE INPUT MATRIX.
!
    reortho = true
!
!   DETERMINE THE NUMBER OF SUBSPACE ITERATIONS niter_qb TO BE PERFORMED
!   IN THE LAST STEP OF THE QB FACTORIZATION.
!
    niter_qb = 1_i4b
!
!   DETERMINE IF A QB OR QR FACTORIZATION IS COMPUTED.
!
    comp_qr = true
!
!   DETERMINE IF A QR FACTORIZATION WITH COLUMN PIVOTING IS COMPUTED.
!
    pivoting = true
!
!   ALLOCATE WORK ARRAYS.
!
    if ( pivoting ) then
       allocate( a(m,n), ip(n), singval0(nsvd0), stat=iok )
    else
       allocate( a(m,n), singval0(nsvd0), stat=iok )
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A SPECIFIED DISTRIBUTION
!   OF SINGULAR VALUES.
!
!   GENERATE SINGULAR VALUES.
!
    select case( mat_type )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
            end do
!
        case default
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            norma = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i - 1_i4b, stnd )
!
                singval0(i) = exp( -tmp/norma )
!        
            end do
!
    end select
!
!   SORT THE SINGULAR VALUES.
!
    call singval_sort( sort, singval0(:nsvd0) )
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE A m-BY-n MATRIX a OF RANK nsvd0 .
!
    call gen_random_mat( singval0(:nsvd0), a )
!
!   COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
!    norma = norm( a(:m,:n) )
    norma = norm( singval0(:nsvd0) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   rqb_cmp_fixed_precision COMPUTES AN APPROXIMATE PARTIAL QB OR QR DECOMPOSITION OF A REAL m-BY-n MATRIX a,
!   WITH A RANK AS SMALL AS POSSIBLE, BUT WHICH FULFILLS A PRESET TOLERANCE FOR ITS RELATIVE ERROR
!   IN THE FROBENIUS NORM:
!
!                                || A - Q*B ||_F <= ||A||_F * relerr
!
!   , WHERE Q*B IS THE COMPUTED PARTIAL QB OR QR APPROXIMATION, || ||_F IS THE FROBENIUS NORM AND
!   relerr IS A PRESCRIBED ACCURACY TOLERANCE FOR THE RELATIVE ERROR OF THE COMPUTED PARTIAL QB OR QR
!   APPROXIMATION, SPECIFIED IN THE INPUT ARGUMENT relerr.
!
!   HERE THE RANK OF THE PARTIAL QB OR QR DECOMPOSITION IS NOT KNOWN IN ADVANCE AND THE ARGUMENTS q
!   AND b, WHICH WILL CONTAIN THE TWO FACTORS OF THE QB OR QR DECOMPOSITION MUST BE SPECIFIED AS REAL
!   ARRAY POINTERS.
!
!   On EXIT, nqb = size( q, 2 ) = size( b, 1 ) IS THE RANK OF THE COMPUTED PARTIAL QB OR QR APPROXIMATION.
!
!   FIRST SET THE TOLERANCE FOR THE RELATIVE ERROR OF THE PARTIAL QB OR QR FACTORIZATION IN FROBENIUS NORM.
!
    relerr = relerr0
!
!   NULLIFY THE POINTERS q AND b SO THAT THEIR STATUT CAN BE CHECKED INSIDE
!   rqb_cmp_fixed_precision SUBROUTINE.
!
    nullify( q, b )
!
    if ( pivoting ) then
!
        call rqb_cmp_fixed_precision( a(:m,:n), relerr, q, b, failure_relerr=failure_relerr,     &
                                      niter=niter, blk_size=blk_size, maxiter_qb=maxiter_qb,     &
                                      ortho=ortho, reortho=reortho, niter_qb=niter_qb, ip=ip(:n) )
!
    else
!
        call rqb_cmp_fixed_precision( a(:m,:n), relerr, q, b, failure_relerr=failure_relerr, &
                                      niter=niter, blk_size=blk_size, maxiter_qb=maxiter_qb, &
                                      ortho=ortho, reortho=reortho, niter_qb=niter_qb,       &
                                      comp_qr=comp_qr )
!
    end if
!
!   THE ROUTINE RETURNS THE TWO FACTORS OF A PARTIAL QB OR QR DECOMPOSITION, WHICH FULFILLS
!   THE PRESET TOLERANCE SPECIFIED IN ARGUMENT relerr.
!
!   ON EXIT OF rqb_cmp_fixed_precision :
!
!       relerr CONTAINS THE RELATIVE ERROR IN FROBENIUS NORM OF THE COMPUTED PARTIAL QB OR QR DECOMPOSITION.
!
!       POINTER q CONTAINS THE OTHONORMAL MATRIX Q OF THE COMPUTED PARTIAL QB OR QR DECOMPOSITION.
!   
!       POINTER b CONTAINS THE ASSOCIATED MATRIX B OF THE COMPUTED PARTIAL QB OR QR DECOMPOSITION.
!
!       failure_relerr = false : INDICATES SUCCESSFUL EXIT AND THE COMPUTED PARTIAL QB OR QR
!       APPROXIMATION FULFILLS THE REQUESTED RELATIVE ERROR SPECIFIED ON ENTRY IN THE ARGUMENT relerr.
!
!       failure_relerr = true  : INDICATES THAT THE COMPUTED PARTIAL QB OR QR FACTORIZATION HAS A RELATIVE ERROR
!       LARGER THAN THE REQUESTED RELATIVE ERROR. THIS MEANS THAT THE REQUESTED ACCURACY TOLERANCE
!       FOR THE RELATIVE ERROR IS TOO SMALL (I.E., relerr < 2 * sqrt( epsilon( relerr )/relerr )
!       OR THAT THE INPUT PARAMETERS blk_size AND/OR maxiter_qb HAVE A TOO SMALL VALUE (E.G., THE
!       PRODUCT blk_size*maxiter_qb SETS THE MAXIMUM ALLOWABLE RANK FOR THE PARTIAL QB OR QR APPROXIMATION,
!       WHICH IS SOUGHT), GIVEN THE DISTRIBUTION OF THE SINGULAR VALUES OF mat, AND MUST BE INCREASED
!       TO FULLFILL THE PRESET ACCURACY TOLERANCE FOR THE RELATIVE ERROR OF THE PARTIAL QB OR QR APPROXIMATION.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   GET THE RANK OF THE COMPUTED PARTIAL QB OR QR FACTORIZATION.
!
    nqb = size( q, 2 )
!
!   COMPUTE THE BEST RELATIVE ERROR FROM THE SVD.
!
    relerr2 = norm( singval0(nqb+1_i4b:nsvd0)/norma )
!
!   COMPUTE ERROR BETWEEN THE BEST AND ESTIMATED RELATIVE ERRORS.
!
    err = abs( relerr - relerr2 )
!
!   TEST ACCURACY OF THE ORTHONORMAL MATRIX Q IF REQUIRED.
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( res(nqb,nqb), id(nqb,nqb), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q(:m,:nqb)**(t)*q(:m,:nqb).
!
        call unit_matrix( id(:nqb,:nqb) )
!
        res(:nqb,:nqb) = abs( id(:nqb,:nqb) - matmul( transpose(q(:m,:nqb)), q(:m,:nqb) ) )
!
        err1 = maxval( res(:nqb,:nqb) )/real(m,stnd)
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( res, id )
!
    end if
!
!   DEALLOCATE WORK ARRAYS AND POINTERS.
!
    if ( pivoting ) then
        deallocate( a, singval0, q, b, ip )
    else
        deallocate( a, singval0, q, b )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure_relerr ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = 1  -> slow decay of singular values'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of singular values'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of singular values'
    write (prtunit,*) ' Matrix type > 3  -> very slow decay of singular values'
!
    write (prtunit,*) 
    write (prtunit,*) 'Rank of the approximate partial QB or QR decomposition              = ', &
                      nqb
    write (prtunit,*) 'Relative error in Frobenius norm     : ||A-Q*B||_F / ||A||_F        = ', &
                      relerr
    write (prtunit,*) 'Best relative error in Frobenius norm: min( ||A-Q*B||_F / ||A||_F ) = ', &
                      relerr2
!
    if ( failure_relerr ) then
!
        write (prtunit,*) 
        write (prtunit,*) 'Fail to converge within ', maxiter_qb,   &
                          ' iterations! ||A-Q*B||_F / ||A||_F = ', relerr, ' >= ', relerr0
        write (prtunit,*) 
!
    else
!
        write (prtunit,*) 
        write (prtunit,*) 'Converge within less than ', maxiter_qb ,' iterations! ||A-Q*B||_F / ||A||_F = ', &
                          relerr, ' < ', relerr0
        write (prtunit,*) 
!
    end if
!
    if ( do_test ) then
        write (prtunit,*)
        write (prtunit,*) 'Orthogonality of the computed orthonormal matrix Q = ', err1
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing a randomized partial QB or QR factorization of rank ', nqb, ' of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_rqb_cmp_fixed_precision
! ==========================================
!
end program ex1_rqb_cmp_fixed_precision
