program ex1_rqlp_svd_cmp_fixed_precision
!
!
! Purpose
! =======
!
!   This program illustrates how to compute an approximate partial SVD with a (randomized)
!   QLP-SVD algorithm, which fullfills a given relative error in Frobenius norm using
!   subroutine RQLP_SVD_CMP_FIXED_PRECISION in module SVD_Procedures.
!                                                                              
!                                                                              
! LATEST REVISION : 24/03/2022
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, seven, c30, c1_e6, merror, &
                         norm, unit_matrix, random_seed_, random_number_, gen_random_mat,   &
                         rqlp_svd_cmp_fixed_precision, merror, allocate_error
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE GENERATED MATRIX,
! nsvd0 IS THE NUMERICAL RANK OF THE GENERATED MATRIX (WHICH MUST BE LESS OR EQUAL TO mn),
!
    integer(i4b), parameter :: prtunit=6, m=3000, n=3000, mn=min(m,n), nsvd0=3000
!   
! relerr0 IS THE REQUESTED RELATIVE ERROR OF THE RESTRICTED SVD IN FROBENIUS NORM,
! conda IS THE CONDITION NUMBER OF THE GENERATED MATRIX IN CASES 5, 6 and 7.
!   
    real(stnd), parameter  :: relerr0=0.5_stnd, conda=c1_e6
!
    character(len=*), parameter :: name_proc='Example 1 of rqlp_svd_cmp_fixed_precision'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: err, err1, err2, err3, eps, elapsed_time, anorm, relerr, relerr2, tmp, tmp2
    real(stnd), dimension(:,:), allocatable :: a, res, id
    real(stnd), dimension(:),   allocatable :: singval0
!
    real(stnd), dimension(:,:), pointer :: leftvec, rightvec
    real(stnd), dimension(:),   pointer :: singval
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: niter, blk_size, nover, i, nsvd, mat_type
!
    logical(lgl) :: random_qr, failure, do_test
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : APPROXIMATE PARTIAL SVD OF A m-BY-n REAL MATRIX USING A (RANDOMIZED)
!               QLP-SVD ALGORITHM, WHICH FULLFILLS A PRESCRIBED TOLERANCE
!               FOR ITS RELATIVE ERROR IN THE FROBENIUS NORM.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type = 4  -> VERY SLOW DECAY OF SINGULAR VALUES
!   mat_type = 5  -> STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda
!                    SPECTRUM WITH FEW DEFLATIONS FOR LAPACK GESDD ROUTINE
!   mat_type = 6  -> GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH MODERATE DEFLATION FOR LAPACK GESDD ROUTINE
!   mat_type = 7  -> ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!                    SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK GESDD ROUTINE
!   mat_type > 7  -> UNIFORM DISTRIBUTION OF SINGULAR VALUES
!
    mat_type = 7_i4b
!
!   SET TOLERANCE FOR THE ERROR OF THE RELATIVE ERROR IN FROBENIUS NORM
!   OF THE COMPUTED PARTIAL SVD.
!
    eps = 0.01_stnd
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE SINGULAR TRIPLETS.
!
    do_test = true
!
!   CHOOSE TUNING PARAMETERS FOR THE RANDOMIZED QLP-SVD ALGORITHM.
!
!   DETERMINE IF A RANDOMIZED PARTIAL QR ALGORITHM IS USED IN THE FIRST QR PHASE OF THE QLP-SVD ALGORITHM.
!
    random_qr = true
!
!   DETERMINE THE BLOCK SIZE blk_size FOR THE RANDOMIZED PARTIAL QR PHASE OF THE QLP-SVD ALGORITHM.
!
    blk_size = 30_i4b
!
!   DETERMINE THE OVERSAMPLING SIZE nover FOR THE RANDOMIZED PARTIAL QR PHASE OF THE QLP-SVD ALGORITHM.
!
    nover = 10_i4b
!
!   DETERMINE THE NUMBER niter OF QR-QL (E.G. SUBSPACE) ITERATIONS PERFORMED AFTER THE INITIAL QLP
!   FACTORIZATION IN THE QLP-SVD ALGORITHM. BY DEFAULT, NO SUBSPACES ITERATIONS ARE PERFORMED.
!
    niter = 2_i4b
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), singval0(nsvd0), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A SPECIFIED DISTRIBUTION
!   OF SINGULAR VALUES AND A RANK EQUAL TO nsvd0.
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE SINGULAR VALUES.
!
    select case( mat_type )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
            end do
!
        case( 4_i4b )
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            tmp = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                singval0(i) = exp( -tmp2/tmp )
!        
            end do
!
        case( 5_i4b )
!
!           STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH FEW DEFLATIONS FOR LAPACK GESDD ROUTINE.
!
            singval0(:nsvd0-1_i4b) = one
            singval0(nsvd0) = one/conda
!
        case( 6_i4b )
!
!           GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH MODERATE DEFLATION FOR LAPACK GESDD ROUTINE.
!
            tmp  = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp2 = real( i - 1_i4b, stnd )
!
                singval0(i) = conda**( -tmp2/tmp )
!        
            end do
!
        case( 7_i4b )
!
!           ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!           THIS IS A SPECTRUM WITH COMPLETE DEFLATION FOR LAPACK GESDD ROUTINE.
!
            tmp  = real( nsvd0 - 1_i4b, stnd )
            tmp2 = one - one/conda
!
            do i = 1_i4b, nsvd0
!
                singval0(i) = one - (real( i - 1_i4b, stnd )/tmp)*tmp2
!        
            end do
!
        case default
!
!           UNIFORM DISTRIBUTION OF SINGULAR VALUES.
!
            call random_number_( singval0(:nsvd0) )
!
    end select
!
#ifdef _F2003
    if ( ieee_support_datatype( singval0 ) ) then
!
        if ( .not.all( ieee_is_normal( singval0(:nsvd0) ) ) ) then
!
            call merror( name_proc//' : Exceptions occurred when generating the input matrix !'  )                    
!
        end if
!
    end if
#endif
!
!   GENERATE A m-BY-n MATRIX a OF RANK nsvd0 .
!
    call gen_random_mat( singval0(:nsvd0), a )
!
!   COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
    anorm = norm( singval0(:nsvd0) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   rqlp_svd_cmp_fixed_precision COMPUTES AN APPROXIMATE  PARTIAL SINGULAR VALUE DECOMPOSITION (SVD)
!   OF A REAL m-BY-n MATRIX a, WITH A RANK AS SMALL AS POSSIBLE, BUT WHICH FULFILLS A PRESET
!   TOLERANCE FOR ITS RELATIVE ERROR IN THE FROBENIUS NORM:
!
!            ||a-rSVD||_F <= ||a||_F * relerr
!
!   , WHERE rSVD IS THE COMPUTED PARTIAL SVD APPROXIMATION, || ||_F IS THE FROBENIUS NORM AND
!   relerr IS A PRESCRIBED ACCURACY TOLERANCE FOR THE RELATIVE ERROR OF THE COMPUTED PARTIAL SVD
!   APPROXIMATION, SPECIFIED IN THE INPUT ARGUMENT relerr.
!
!   THE PARTIAL SVD IS WRITTEN
!
!                       U * S * V'
!
!   WHERE S IS AN nsvd-BY-nsvd MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   DIAGONAL ELEMENTS, U IS AN m-BY-nsvd ORTHONORMAL MATRIX, AND
!   V IS AN n-BY-nsvd ORTHONORMAL MATRIX.  THE DIAGONAL ELEMENTS OF S
!   ARE THE nsvd LARGEST SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!   THE COLUMNS OF U AND V ARE THE ASSOCIATED LEFT AND RIGHT SINGULAR VECTORS OF a.
!
!   HERE THE RANK OF THE PARTIAL SVD IS NOT KNOWN IN ADVANCE AND THE ARGUMENTS singval,
!   leftvec AND rightvec, WHICH WILL CONTAIN THE SINGULAR VALUES AND THE ASSOCIATED SINGULAR
!   VECTORS IN EXIT OF THE SUBROUTINE MUST BE SPECIFIED AS REAL ARRAY POINTERS.
!   On EXIT, nsvd = size( singval ) IS THE RANK OF THE COMPUTED PARTIAL SVD.
!
!   SET THE TOLERANCE FOR THE RELATIVE ERROR IN FROBENIUS NORM.
!
    relerr = relerr0
!
!   NULLIFY THE POINTERS singval, leftvec AND rightvec SO THAT THEIR STATUT CAN BE CHECKED INSIDE
!   rqlp_svd_cmp_fixed_precision SUBROUTINE.
!
    nullify( singval, leftvec, rightvec )
!
    call rqlp_svd_cmp_fixed_precision( a(:m,:n), relerr, singval, leftvec, rightvec, failure=failure,   &
                                       niter=niter, random_qr=random_qr, blk_size=blk_size, nover=nover )
!
!   THE ROUTINE RETURNS THE nsvd LARGEST SINGULAR VALUES AND THE ASSOCIATED LEFT
!   AND RIGHT SINGULAR VECTORS, WHICH FULFILLS THE PRESET TOLERANCE SPECIFIED IN ARGUMENT
!   relerr.
!
!   ON EXIT OF rqlp_svd_cmp_fixed_precision :
!
!       relerr CONTAINS THE RELATIVE ERROR IN FROBENIUS NORM OF THE COMPUTED PARTIAL SVD.
!
!       POINTER singval CONTAINS THE nsvd = size(singval) LARGEST SINGULAR VALUES OF a
!       IN DECREASING ORDER.
!
!       POINTER leftvec CONTAINS THE ASSOCIATED nsvd LEFT SINGULAR VECTORS
!       STORED COLUMNWISE.
!   
!       POINTER rightvec CONTAINS THE ASSOCIATED nsvd RIGHT SINGULAR VECTORS
!       STORED COLUMNWISE.
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND THAT
!                         FULL ACCURACY WAS NOT ATTAINED IN THE (RANDOMIZED)
!                         PARTIAL QLP-SVD ALGORITHM WITH niter QR-QL ITERATIONS
!                         FOR THE REQUESTED ACCURACY.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   GET THE RANK OF THE COMPUTED PARTIAL SVD.
!
    nsvd = size( singval )
!
!   COMPUTE THE TRUE RELATIVE ERROR.
!
    if ( nsvd0>nsvd ) then
        relerr2 = norm( singval0(nsvd+1_i4b:nsvd0)/anorm )
    else
        relerr2 = zero
    end if
!
!   COMPUTE ERROR BETWEEN THE TRUE AND ESTIMATED RELATIVE ERRORS.
!
    err = abs( relerr - relerr2 )
!
!   TEST ACCURACY OF THE COMPUTED SINGULAR TRIPLETS IF REQUIRED.
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( res(m,nsvd), id(nsvd,nsvd), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v(:n,:nsvd) - u(:m,:nsvd)*diag(singval(:nsvd)),
!       WHERE u AND v ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
        res(:m,:nsvd) = matmul(a,rightvec) - leftvec(:m,:nsvd)*spread(singval(:nsvd),dim=1,ncopies=m)
        id(:nsvd,1_i4b) = norm( res(:m,:nsvd), dim=2_i4b )
!
        if ( anorm==zero ) then
            anorm = one
        end if
!
        err1 = maxval( id(:nsvd,1_i4b) )/( anorm*real(mn,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u(:m,:nsvd)**(t)*u(:m,:nsvd).
!
        call unit_matrix( id(:nsvd,:nsvd) )
!
        res(:nsvd,:nsvd) = abs( id(:nsvd,:nsvd) - matmul( transpose(leftvec(:m,:nsvd)), leftvec(:m,:nsvd) ) )
!
        err2 = maxval( res(:nsvd,:nsvd) )/real(m,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v(:n,:nsvd)**(t)*v(:n,:nsvd).
!
        res(:nsvd,:nsvd) = abs( id(:nsvd,:nsvd) - matmul( transpose(rightvec), rightvec ) )
!
        err3 = maxval( res(:nsvd,:nsvd) )/real(n,stnd)
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( res, id )
!
    end if
!
!   DEALLOCATE WORK ARRAYS AND POINTERS.
!
    deallocate( a, singval0, singval, leftvec, rightvec )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = ', mat_type
!
    write (prtunit,*) 
    write (prtunit,*) ' Matrix type = 1  -> slow decay of singular values'
    write (prtunit,*) ' Matrix type = 2  -> fast decay of singular values'
    write (prtunit,*) ' Matrix type = 3  -> s-shaped decay of singular values'
    write (prtunit,*) ' Matrix type = 4  -> very slow decay of singular values'
    write (prtunit,*) ' Matrix type = 5  -> strongly clustered singular values at 1'
    write (prtunit,*) '                     spectrum with few deflations (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type = 6  -> geometric distribution of singular values'
    write (prtunit,*) '                     spectrum with moderate deflation (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type = 7  -> arithmetic distribution of singular values'
    write (prtunit,*) '                     spectrum with complete deflation (for LAPACK GESDD routine)'
    write (prtunit,*) ' Matrix type > 7  -> uniform distribution of singular values'
!
    write (prtunit,*) 
    write (prtunit,*) 'Rank of the approximate partial SVD                                  = ', &
                      nsvd
    write (prtunit,*) 'Relative error in Frobenius norm     : ||A-rSVD||_F / ||A||_F        = ', &
                      relerr, ' < ', relerr0
    write (prtunit,*) 'Best relative error in Frobenius norm: min( ||A-rSVD||_F / ||A||_F ) = ', &
                      relerr2
    write (prtunit,*) 
!
    write (prtunit,*) 
    write (prtunit,*) 'FAILURE ( from rqlp_svd_cmp_fixed_precision() ) = ', failure
!
    if ( do_test ) then
        write (prtunit,*)
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing approximate ', nsvd, ' singular values and vectors of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_rqlp_svd_cmp_fixed_precision
! ===============================================
!
end program ex1_rqlp_svd_cmp_fixed_precision
