program ex1_symtrid_ratqri2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SYMTRID_RATQRI2
!   in module Eig_Procedures.
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutine TRID_INVITER in module Eig_Procedures.
!
! LATEST REVISION : 27/01/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, two, ten, c50,  &
                         allocate_error, merror, trid_inviter, symtrid_ratqri2,  &
                         norm, unit_matrix 
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Utilities, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT.
! n IS THE DIMENSION OF THE SYMMETRIC TRIDIAGONAL MATRIX.
!
    integer(i4b), parameter :: prtunit=6, n=1000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of symtrid_ratqri2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, elapsed_time
    real(stnd), allocatable, dimension(:)   :: d, e, e2, eigval
    real(stnd), allocatable, dimension(:,:) :: eigvec, a, a2, resid
!
    integer(i4b) :: maxiter=2, l, neig
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: failure, do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : COMPUTE THE neig LARGEST OR SMALLEST EIGENVALUES OF
!               A SYMMETRIC TRIDIAGONAL MATRIX WHOSE SUM OF ABSOLUTE VALUES
!               EXCEEDS A PRESCRIBED THRESHOLD USING A RATIONAL QR METHOD
!               AND SELECTED EIGENVECTORS BY INVERSE ITERATION.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( eps )
!
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( d(n), e(n), e2(n), eigval(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A (1-2-1) TRIDIAGONAL MATRIX.
!   THE DIAGONAL ELEMENTS ARE STORED IN d .
!   THE OFF-DIAGONAL ELEMENTS ARE STORED IN e .
!
    d(:n) = two
    e(:n) = one
!
!   SAVE THE TRIDIAGONAL FORM FOR LATER USE.
!
    eigval(:n) = d(:n)
    e2(:n)     = e(:n)
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE neig LARGEST EIGENVALUES OF THE SYMMETRIC TRIDIAGONAL MATRIX,
!   WHOSE SUM OF ABSOLUTE VALUES EXCEEDS 10.
!
    call symtrid_ratqri2( eigval(:n), e2(:n), ten, failure, neig )
!
!   ON EXIT, THE COMPUTED EIGENVALUES ARE STORED IN eigval(:neig)
!   AND THE SYMMETRIC TRIDIAGONAL MATRIX IS OVERWRITTEN.
!
!   ON EXIT OF symtrid_ratqri2 :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE RATIONAL
!                         QR ALGORITHM.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( .not.failure .and. neig>0_i4b ) then
!
        allocate( eigvec(n,neig), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE THE FIRST neig EIGENVECTORS OF THE (1-2-1) TRIDIAGONAL MATRIX BY maxiter INVERSE ITERATIONS.
!
        call trid_inviter( d(:n), e(:n), eigval(:neig), eigvec(:n,:neig), failure, maxiter=maxiter )
!
        if ( do_test ) then
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( a(n,n), a2(neig,neig), resid(n,neig), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           FORM THE TRIDIAGONAL MATRIX.
!
            a(:n,:n) = zero
!
            do l = 1_i4b, n-1_i4b
!
                a(l,l)       = d(l)
                a(l+1_i4b,l) = e(l)
                a(l,l+1_i4b) = e(l)
!
            end do
!
            a(n,n) = d(n)
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*s
!           WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX a.
!
            resid = matmul( a, eigvec ) - eigvec*spread( eigval(:neig), 1, n )
!
            err1 = norm(resid)/( norm(a)*real(n,stnd) )
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!           WHERE u are THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX a.
!
            call unit_matrix( a2 )
!
            resid(:neig,:neig) = a2 - matmul( transpose( eigvec ), eigvec )
!
            err2 = norm(resid(:neig,:neig))/real(n,stnd)
!
            err = max( err1, err2 )
!
!           ALLOCATE WORK ARRAYS.
!
            deallocate( a, a2, resid )
!
        end if
!
!       ALLOCATE WORK ARRAY.
!
        deallocate( eigvec )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( d, e, e2, eigval )
!
!   PRINT THE RESULTS OF THE TESTS.
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', neig, ' eigenvalues of a ', &
       n, ' by ', n,' real symmetric tridiagonal matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_symtrid_ratqri2
! ==================================
!
end program ex1_symtrid_ratqri2
