program ex2_llsq_svd_solve
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine LLSQ_SVD_SOLVE
!   in module LLSQ_Procedures. 
!                                                                              
! LATEST REVISION : 20/01/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c500, lamch, norm,        &
                         print_array, llsq_svd_solve, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT.
! m IS THE NUMBER OF ROWS OF THE RANDOM MATRIX.
! n IS THE NUMBER OF COLUMNS OF THE RANDOM MATRIX.
! nrhs IS THE NUMBER OF COLUMNS OF THE RIGHT HAND SIDE MATRIX.
!
    integer(i4b), parameter :: prtunit=6, m=4000, n=2000, mn=min(m,n), nrhs=10
!   
    real(stnd), parameter  :: fudge=c500
!
    character(len=*), parameter :: name_proc='Example 2 of llsq_svd_solve'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: err, err1, err2, ulp, eps, tol, anorm, cond, sfmin, elapsed_time
    real(stnd), allocatable, dimension(:)   :: sing_values, rnorm, bnorm
    real(stnd), allocatable, dimension(:,:) :: a, a2, b, b2, res, res2, x
!
    integer(i4b) :: krank, j
!
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: failure, do_test, do_print
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : SOLVE A LINEAR LEAST SQUARES SYSTEM WITH ONE RIGHT HAND-SIDE
!               BY THE SINGULAR VALUE DECOMPOSITION OF THE COEFFICIENT MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( ulp )
    eps = fudge*ulp
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE SUBROUTINE
!   AND IF DETAILED RESULTS MUST BE PRINTED.
!
    do_test  = true
    do_print = false
!
!   SET TOLERANCE FOR DETERMINING THE RANK OF THE MATRIX IN THE SUBROUTINE.
!
!    tol = 0.0000001_stnd
    tol = sqrt( ulp )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), b(m,nrhs), x(n,nrhs), sing_values(mn), &
              bnorm(nrhs), rnorm(nrhs), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-by-n REAL RANDOM COEFFICIENT MATRIX a .
!
    call random_number( a )
!
!   GENERATE A DEFICIENT COEFFICIENT MATRIX a(:m,:n) WITH RANK min(mn-1,n) .
!
    j = min( n, 5_i4b ) 
    a(:m,j) = a(:m,1_i4b) + a(:m,2_i4b)
!
!   GENERATE A RANDOM RIGHT HAND-SIDE MATRIX b .
!
    call random_number( b )
!
!   COMPUTE THE NORMS OF THE nrhs DEPENDENT VARIABLES b .
!
    bnorm(:nrhs) = norm( b(:m,:nrhs), dim=2_i4b )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(m,n), b2(m,nrhs), res(m,nrhs), res2(nrhs,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE FROBENIUS NORM OF THE MATRIX.
!
        anorm = norm( a(:m,:n) )
!
!       SAVE DATA MATRIX .
!
        a2(:m,:n) = a(:m,:n)
!
!       SAVE RIGHT HAND SIDE MATRIX .
!
        b2(:m,:nrhs) = b(:m,:nrhs)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   llsq_svd_solve COMPUTES THE MINIMUM NORM SOLUTION TO A REAL LINEAR LEAST
!   SQUARES PROBLEM :
!
!                       Minimize || b - a*x ||_2
!
!   USING THE SINGULAR VALUE DECOMPOSITION (SVD) OF a. A IS AN m-BY-n MATRIX
!   WHICH MAY BE RANK-DEFICIENT. b AND x CAN BE VECTORS OF MATRICES, BUT THEIR
!   SHAPES MUST BE CONFORMABLE WITH THE SHAPE OF a.
!
!   IN OTHER WORDS, IF b AND x ARE MATRICES, SEVERAL RIGHT HAND SIDE VECTORS b
!   AND SOLUTION VECTORS x CAN BE HANDLED IN A SINGLE CALL; THEY ARE STORED AS
!   THE COLUMNS OF THE m-BY-nrhs RIGHT HAND SIDE MATRIX b AND THE n-BY-nrhs SOLUTION
!   MATRIX b, RESPECTIVELY.
!
!   THE EFFECTIVE RANK OF a, krank,IS DETERMINED BY TREATING AS ZERO THOSE
!   SINGULAR VALUES WHICH ARE LESS THAN tol TIMES THE LARGEST SINGULAR VALUE.
!
    call llsq_svd_solve( a, b, failure, x,                                            &
                         singvalues=sing_values, krank=krank, rnorm=rnorm, tol=tol )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK THAT THE RESIDUAL MATRIX IS ORTHOGONAL TO THE RANGE OF a .
!
        res(:m,:nrhs)  = b2(:m,:nrhs) - matmul( a2(:m,:n), x(:n,:nrhs) )
        res2(:nrhs,:n) = matmul( transpose(res(:m,:nrhs)), a2(:m,:n) )
!
        err1 = maxval( abs(res2(:nrhs,:n)) )/anorm
!
!       CHECK THE NORMS OF THE COLUMNS OF RESIDUAL MATRIX.
!
        err2 = maxval( abs( norm( res(:m,:nrhs), dim=2_i4b ) - rnorm(:nrhs) ) )
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, b2, res, res2 )
!
    end if

    write (prtunit,*) err1, err2, eps, failure

!
!   PRINT THE RESULT OF THE TESTS.
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
!   GET MACHINE CONSTANT sfmin, SUCH THAT 1/sfmin DOES NOT OVERFLOW.
!
    sfmin = lamch( 's' )
!
!   COMPUTE THE CONDITION NUMBER OF a(:m,:n) IN THE 2-NORM
!
!            singvalues(1)/singvalues(min(m,n)) .
!
    if ( sing_values(mn)/sing_values(1_i4b)<=sfmin ) then
        cond = huge( cond )
    else
        cond = sing_values(1_i4b)/sing_values(mn)
    end if
!
!   PRINT RESULTS .
!
    write (prtunit,*)
    write (prtunit,*)
    write (prtunit,*) 'Least squares solution via Singular Value Decomposition'
    write (prtunit,*)
    write (prtunit,*) '    min of ||a(:,:)*x(:,:)-b(:,:)||**2 for matrix x(:,:) '
    write (prtunit,*)
    write (prtunit,*) 'Tolerance for zero singular values (tol*sing_values(1)):',tol*sing_values(1)
    write (prtunit,*)
    write (prtunit,*) 'Condition number (in the 2-norm) of a :',cond
    write (prtunit,*) 'Rank of a                             :',krank
    write (prtunit,*)
    write (prtunit,*) 'Residual sum of squares     ||a*x(:,i)-b(:,i)||**2               :',rnorm(:nrhs)**2
    write (prtunit,*) 'Residual sum of squares (%) ||a*x(:,i)-b(:,i)||**2/||b(:,i)||**2 :',(rnorm(:nrhs)/bnorm(:nrhs))**2
    write (prtunit,*)
!
    if ( do_print ) then
!
!       PRINT DETAILED RESULTS.
!
        call print_array( sing_values, title=' Singular values of a ' )
!
        write (prtunit,*)
!
        call print_array( x, title=' Least squares solution matrix x ' )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, b, x, sing_values, rnorm, bnorm )
!
    write (prtunit,*)
    write (*,'(a,i5,a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for solving a linear least squares problem with a ', &
      m, ' by ', n,' real coefficient matrix and a ', m, ' by ', nrhs,       &
      ' right hand side matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_llsq_svd_solve
! =================================
!
end program ex2_llsq_svd_solve
