program ex2_qr_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines QR_CMP and QR_SOLVE
!   in modules QR_Procedures and LLSQ_Procedures.
!    
! LATEST REVISION : 21/01/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, lgl, stnd, zero, true, false, c100, allocate_error, merror, &
                         qr_cmp, qr_solve, norm
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT.
! m IS THE NUMBER OF ROWS OF THE RANDOM MATRIX.
! n IS THE NUMBER OF COLUMNS OF THE RANDOM MATRIX.
!
    integer(i4b), parameter :: prtunit=6, m=5000, n=1000, mn=min( m, n )
!   
    real(stnd), parameter  :: fudge=c100
!
    character(len=*), parameter :: name_proc='Example 2 of qr_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: ulp, eps, anorm, rnorm, err1, err2, err, elapsed_time
    real(stnd), allocatable, dimension(:)   :: b, x, res, diagr, beta
    real(stnd), allocatable, dimension(:,:) :: a, a2
!
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: comp_resid, do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : SOLVING A LINEAR LEAST SQUARES PROBLEM WITH A m-BY-n REAL COEFFICIENT
!               MATRIX USING A QR DECOMPOSITION OF THE COEFFICIENT MATRIX. THE COEFFICIENT
!               MATRIX IS ASSUMED OF FULL RANK, BUT BOTH m>=n OR m<n ARE PERMITTED.
!
!               COMPUTE SOLUTION VECTOR FOR LINEAR LEAST SQUARES SYSTEM
!
!                              a(:m,:n)*x(:n) ≈ b(:m) .
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = fudge*ulp
!
    err = zero
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE SUBROUTINE.
!
    do_test = true
!
!   DECIDE IF RESIDUAL VECTOR MUST BE COMPUTED.
!
    comp_resid = do_test
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), b(m), x(n), diagr(mn), beta(mn), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM COEFFICIENT MATRIX a(:m,:n) .
!
    call random_number( a(:m,:n) )
!
!   GENERATE A RANDOM RIGHT HAND-SIDE VECTOR b(:m) .
!
    call random_number( b(:m) )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS FOR LATER USE.
!
        allocate( a2(m,n), res(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE FROBENIUS NORM OF THE DATA MATRIX.
!
        anorm = norm( a(:m,:n) )
!
!       MAKE A COPY OF THE DATA MATRIX FOR LATER USE.
!
        a2(:m,:n) = a(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE A QR FACTORIZATION OF RANDOM DATA MATRIX WITH SUBROUTINE qr_cmp.
!
    call qr_cmp( a(:m,:n), diagr(:mn), beta(:mn) )
!
!   qr_cmp COMPUTES AN ORTHOGONAL FACTORIZATION OF A REAL m-BY-n MATRIX
!   a . a IS ASSUMED OF FULL RANK. THE ROUTINE COMPUTES A QR FACTORIZATION
!   OF a AS:
!
!                     a = Q * R
!
!   ON EXIT, a HAS BEEN OVERWRITTEN BY DETAILS OF ITS ORTHOGONAL
!   FACTORIZATION. 
!
!   THE MATRIX Q IS REPRESENTED AS A PRODUCT OF ELEMENTARY REFLECTORS
!
!            Q = h(1)*h(2)* ... *h(k), WHERE k = min( size(a,1) , size(a,2) )
!
!   EACH h(i) HAS THE FORM
!
!            h(i) = I + BETA * ( V * V' ) ,
!                      
!   WHERE BETA IS A REAL SCALAR AND V IS A REAL m-ELEMENTS VECTOR WITH V(1:i-1) = 0.
!   V(i:m) IS STORED ON EXIT IN a(i:m,i) AND BETA IN beta(i).
!                      
!   THE ELEMENTS ABOVE THE DIAGONAL OF THE ARRAY a CONTAIN THE
!   CORRESPONDING ELEMENTS OF THE TRIANGULAR MATRIX R. THE ELEMENTS
!   OF THE DIAGONAL OF R ARE STORED IN THE ARRAY diagr. 
!
!   NOW, COMPUTE SOLUTION AND RESIDUAL VECTORS FOR LINEAR LEAST SQUARES SYSTEM
!   WITH SUBROUTINE qr_solve.
!
    call qr_solve( a(:m,:n), diagr(:mn), beta(:mn), b(:m), x(:n),  &
                   rnorm=rnorm, comp_resid=comp_resid )
!
!   qr_solve SOLVES OVERDETERMINED OR UNDERDETERMINED REAL LINEAR SYSTEMS
!
!                a(:m,:n)*x(:n) ≈ b(:m) .
!
!   WITH AN m-BY-n COEFFICIENT MATRIX a, USING AN ORTHOGONAL FACTORIZATION OF a, AS 
!   COMPUTED BY qr_cmp. m>=n OR n>m IS PERMITTED, BUT a IS ASSUMED OF FULL RANK.
!
!   b IS A m-ELEMENTS RIGHT HAND SIDE VECTOR AND x IS A n-ELEMENTS SOLUTION VECTOR. SEVERAL
!   RIGHT HAND SIDES AND SOLUTION VECTORS MAY BE HANDLE IN A SINGLE CALL. IN THIS CASE,
!   b IS AN m-BY-nrhs MATRIX AND x is AN n-BY-nrhs MATRIX. b IS OVERWRITTEN BY qr_solve.
!   
!   IT IS ASSUMED THAT qr_cmp  HAS BEEN USED TO COMPUTE THE ORTHOGONAL 
!   FACTORIZATION OF a BEFORE CALLING qr_solve.
!
!   ON EXIT, IF THE OPTIONAL LOGICAL PARAMETER comp_resid IS PRESENT AND IS SET TO true,
!   THE RESIDUAL VECTOR OR MATRIX b - a*x is COMPUTED AND OVERWRITES b.
!
!   THE 2-NORM OF THE RESIDUAL VECTOR FOR THE SOLUTION VECTOR x, CAN ALSO BE COMPUTED
!   DIRECTLY BY SPECIFYING THE OPTIONAL REAL PARAMETER rnorm IN THE CALL OF qr_solve .
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK THAT THE RESIDUALS ARE ORTHOGONAL TO THE RANGE OF THE COEFFICIENT MATRIX a .
!
        res(:n) = matmul( b(:m), a2(:m,:n) )
!
        err1 = maxval( abs(res(:n)) )/anorm
!
!       CHECK THE NORM OF THE RESIDUAL VECTOR.
!
        err2 = abs( norm( b(:m) ) - rnorm )
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, res )
!
    end if
!
!   PRINT THE RESULTS OF THE TESTS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) '2-norm of residual vector ||a*x-b|| = ', rnorm
!
    write (prtunit,*)
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for solving a linear least squares problem with a ', &
       m, ' by ', n,' real coefficient matrix is ', elapsed_time, ' seconds'
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, b, x, diagr, beta )
!
!
! END OF PROGRAM ex2_qr_cmp
! =========================
!
end program ex2_qr_cmp
