program ex3_partial_qr_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines PARTIAL_QR_CMP in module
!   QR_Procedures and QR_SOLVE2 in module LLSQ_Procedures.
!                                                                              
! LATEST REVISION : 21/01/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, lgl, stnd, zero, true, false, c50, allocate_error, merror, &
                         partial_qr_cmp, qr_solve2, norm
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE GENERATED MATRIX,
!
    integer(i4b), parameter :: prtunit=6, m=8000, n=4000, mn=min( m, n )
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 3 of partial_qr_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: ulp, eps, tol, anorm, rnorm, err1, err2, err, elapsed_time
    real(stnd), allocatable, dimension(:)   :: b, x, res, diagr, beta, tau
    real(stnd), allocatable, dimension(:,:) :: a, a2
!
    integer(i4b)                            :: krank, l, j, idep
    integer(i4b), allocatable, dimension(:) :: ip
    integer                                 :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: comp_resid, do_test, test_lin
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 3 : SOLVING A LINEAR LEAST SQUARES PROBLEM WITH A m-BY-n REAL COEFFICIENT
!               MATRIX USING A QR DECOMPOSITION WITH COLUMN PIVOTING OR COMPLETE
!               ORTHOGONAL DECOMPOSITION OF THE COEFFICIENT MATRIX AND ONE RIGHT HAND SIDE.
!
!               COMPUTE SOLUTION VECTOR x FOR LINEAR LEAST SQUARES SYSTEM:
!
!                              a(:m,:n)*x(:n) ≈ b(:m) .
!
!
!   SET THE TOLERANCE AND REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( ulp )
    eps = sqrt( fudge*ulp )
!
    err      = zero
    test_lin = true
!
!   SET TOLERANCE FOR DETERMINING THE RANK OF THE COMPUTED QR APPROXIMATION IN THE SUBROUTINE.
!
!    tol = 0.0000001_stnd
    tol = sqrt( ulp )
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE SUBROUTINE.
!
    do_test = true
!
!   DECIDE IF RESIDUAL VECTOR MUST BE COMPUTED.
!
    comp_resid = do_test
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), b(m), x(n), diagr(mn), beta(mn), tau(mn), ip(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM COEFFICIENT MATRIX a(:m,:n) .
!
    call random_number( a )
!
!   GENERATE A DEFICIENT COEFFICIENT MATRIX a(:m,:n) WITH RANK min(mn-1,n) .
!
    idep = min( n, 5_i4b ) 
    a(:m,idep) = a(:m,1_i4b) + a(:m,n)
!
!   GENERATE A RANDOM RIGHT HAND-SIDE VECTOR b(:m) .
!
    call random_number( b )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(m,n), res(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE FROBENIUS NORM OF THE MATRIX.
!
        anorm = norm( a(:m,:n) )
!
!       SAVE DATA MATRIX FOR LATER USE.
!
        a2(:m,:n) = a(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE A QR DECOMPOSITION WITH COLUMN PIVOTING OR COMPLETE ORTHOGONAL
!   DECOMPOSITION OF RANDOM DATA MATRIX WITH SUBROUTINE partial_qr_cmp.
!
    call partial_qr_cmp( a(:m,:n), diagr(:mn), beta(:mn), ip(:n), krank, tol=tol, tau=tau )
!
!   partial_qr_cmp COMPUTES A (PARTIAL OR COMPLETE) ORTHOGONAL FACTORIZATION
!   OF A REAL m-BY-n MATRIX. THE INPUT MATRIX MAY BE RANK-DEFICIENT.
!
!   HERE THE ROUTINE FIRST COMPUTES A QR FACTORIZATION WITH COLUMN PIVOTING OF a AS:
!
!                     a * P = Q * R = Q * [ R11 R12 ]
!                                         [  0  R22 ]
!
!   P IS A n-BY-n PERMUTATION MATRIX, Q IS A m-BY-m ORTHOGONAL MATRIX AND
!   R IS A m-BY-n UPPER TRIANGULAR OR TRAPEZOIDAL MATRIX.
!   
!   R11 IS DEFINED AS THE LARGEST LEADING SUBMATRIX OF R WHOSE ESTIMATED CONDITION
!   NUMBER  IS LESS THAN 1/tol IF tol IS IN ]0,1[ OR SUCH THAT ABS(R11[j,j])>0 IF
!   tol IS EQUAL TO 0 (SEE BELOW FOR DETAILS).
!   
!   THE ORDER OF R11 IS AN ESTIMATE OF THE RANK OF a AND IS STORED
!   IN THE ARGUMENT krank ON EXIT OF partial_qr_cmp.
!
!   IN OTHER WORDS, krank CONTAINS THE EFFECTIVE RANK OF a. I.E., THE NUMBER
!   OF INDEPENDENT COLUMNS IN MATRIX a.
!
!   ON EXIT, a HAS BEEN OVERWRITTEN BY DETAILS OF ITS QR FACTORIZATION:
!                      
!   - THE ELEMENTS BELOW AND ON THE DIAGONAL OF THE ARRAY a(:m,:mn) AND THE ARRAY
!     beta(:mn) STORED Q IN FACTORED FORM.
!                      
!   - THE ELEMENTS ABOVE THE DIAGONAL OF THE ARRAY a CONTAIN THE
!     CORRESPONDING ELEMENTS OF THE TRIANGULAR OR TRAPEZOIDAL MATRIX R.
!     THE ELEMENTS OF THE DIAGONAL OF R ARE STORED IN THE ARRAY diagr(:mn). 
!
!   - ip STORES THE PERMUTATION MATRIX P IN THE QR DECOMPOSITION OF a.
!     IF ip(j)=k, THEN THE jTH COLUMN OF a*P WAS THE kTH COLUMN OF a.
!     THE MATRIX P IS REPRESENTED IN THE ARRAY ip AS FOLLOWS:
!     IF ip(j) = i THEN THE jTH COLUMN OF P IS THE iTH CANONICAL UNIT VECTOR.
!                      
!   - krank CONTAINS THE EFFECTIVE RANK OF a (E.G. THE RANK OF R11),
!                      
!   IF tol IS PRESENT AND IS IN ]0,1[, THEN :
!       CALCULATIONS TO DETERMINE THE CONDITION NUMBER OF THE SUBMATRIX R11
!       ARE PERFORMED. tol IS USED TO DETERMINE THE EFFECTIVE RANK
!       OF R (and a), WHICH IS DEFINED AS THE ORDER OF THE LARGEST LEADING
!       TRIANGULAR SUBMATRIX IN THE QR FACTORIZATION WITH PIVOTING OF a,
!       WHOSE ESTIMATED CONDITION NUMBER (IN THE 1-NORM) IS LESS THAN 1/tol.
!       ON EXIT, THE RECIPROCAL OF THE CONDITION NUMBER IS RETURNED IN tol.
!
!   IF tol IS PRESENT AND IS EQUAL TO 0, THEN :
!       THE NUMERICAL RANK OF R (and a) IS DETERMINED, E.G., CRUDE TESTS ON R(j,j) ARE
!       DONE TO DETERMINE THE RANK OF R AND tol IS NOT CHANGED ON EXIT.
!
!   IF tol IS NOT SPECIFIED OR IS OUTSIDE [0,1[ :
!       THE CALCULATIONS TO DETERMINE THE CONDITION NUMBER OF R11 ARE NOT
!       PERFORMED AND THE RANK OF R IS ASSUMED TO BE EQUAL TO mn = min(m,n).
!
!   IF IT IS POSSIBLE THAT a MAY NOT BE FULL RANK (I.E., CERTAIN COLUMNS OF a ARE
!   LINEAR COMBINATIONS OF OTHER COLUMNS), THEN THE LINEARLY DEPENDENT COLUMNS CAN
!   USUALLY BE DETERMINED BY USING tol=RELATIVE PRECISION OF THE ELEMENTS IN a.
!   IF EACH ELEMENT IS CORRECT TO, SAY, 5 DIGITS THEN tol=0.00001 SHOULD BE USED.
!                      
!   IF THE OPTIONAL PARAMETER tau IS PRESENT, partial_qr_cmp COMPUTES A (PARTIAL OR COMPLETE)
!   ORTHOGONAL FACTORIZATION OF a FROM THE QR FACTORIZATION WITH COLUMN PIVOTING OF a .
!
!   THIS COMPLETE ORTHOGONAL FACTORIZATION CAN THEN BE USED TO COMPUTE THE MINIMAL 2-NORM
!   SOLUTIONS FOR RANK DEFICIENT LINEAR LEAST SQUARES PROBLEMS WITH SUBROUTINE qr_solve2.
!
!   NEXT, COMPUTE THE SOLUTION VECTOR FOR LINEAR LEAST SQUARE SYSTEM
!
!                   a(:m,:n)*x(:n) ≈ b(:m) .
!
!   WITH SUBROUTINE qr_solve2.
!
    call qr_solve2( a(:m,:n), diagr(:mn), beta(:mn), ip(:n), krank, b(:m), x(:n),  &
                    tau=tau, rnorm=rnorm, comp_resid=comp_resid )
!
!   qr_solve2 COMPUTES SOLUTION(S) TO REAL LINEAR LEAST SQUARES PROBLEMS 
!   OF THE FORM:
!
!                       Minimize || b - a*x ||_2
!
!   USING AN ORTHOGONAL FACTORIZATION WITH COLUMNS PIVOTING OR A COMPLETE
!   ORTHOGONAL FACTORIZATION OF a AS COMPUTED BY qr_cmp2, partial_qr_cmp,
!   partial_rqr_cmp AND partial_rqr_cmp2.
!
!   a IS A m-BY-n MATRIX WHICH MAY BE RANK-DEFICIENT. FURTHERMORE m>=n
!   OR n>m IS PERMITTED.
!
!   SEVERAL RIGHT HAND SIDE VECTORS b AND SOLUTION VECTORS x CAN BE
!   HANDLED IN A SINGLE CALL; THEY ARE STORED AS THE COLUMNS OF THE
!   m-BY-nrhs RIGHT HAND SIDE MATRIX b AND THE n-BY-nrhs SOLUTION
!   MATRIX x, RESPECTIVELY. IN OTHER WORDS, THE ARGUMENTS b AND x MAY
!   BE VECTORS OR MATRICES. b IS OVERWRITTEN BY qr_solve2.
!   
!   ON EXIT, IF THE OPTIONAL LOGICAL PARAMETER comp_resid IS PRESENT AND IS SET TO true,
!   THE RESIDUAL VECTOR OR MATRIX b - a*x is COMPUTED AND OVERWRITES b.
!
!   THE 2-NORM OF THE RESIDUAL VECTOR(S) FOR THE SOLUTION VECTOR(S) x, CAN ALSO BE COMPUTED
!   DIRECTLY BY SPECIFYING THE OPTIONAL PARAMETER rnorm IN THE CALL OF qr_solve2.
!
!   THE MINIMUN 2-NORM SOLUTION(S) ARE COMPUTED IF THE OPTIONAL ARRAY PARAMETER tau IS
!   PRESENT IN BOTH THE CALLS OF partial_qr_cmp AND qr_solve2. OTHERWISE, SOLUTION(S) ARE
!   COMPUTED SUCH THAT IF THE jTH COLUMN OF a IS OMITTED FROM THE BASIS, x[j] O
!   x[j,:nrhs] IS SET TO ZERO.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   CHECK IF THE DEPENDENT COLUMNS IN MATRIX a HAVE BEEN IDENTIFIED.
!
    if ( m>=n .and. idep<n ) then
!
        do l = krank+1_i4b, n
!
            j = ip(l)
!
            if ( j==idep .or. j==1_i4b .or. j==n ) exit
!
        end do
!
        test_lin = l/=n+1_i4b
!
    end if
!
    if ( do_test ) then
!
!       CHECK THAT THE RESIDUALS ARE ORTHOGONAL TO THE COEFFICIENT MATRIX a .
!
        res(:n) = matmul( b(:m), a2(:m,:n) )
!
        err1 = maxval( abs(res(:n)) )/anorm
!
!       CHECK THE NORM OF THE RESIDUAL VECTOR.
!
        err2 = abs( norm( b(:m) ) - rnorm )
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, res )
!
    end if
!
!   PRINT THE RESULT OF THE TESTS.
!
    if ( err<=eps .and. test_lin ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*)
    write (prtunit,*) 'Estimated rank of the coefficient matrix = ', krank
!
    if ( krank/=mn ) then
        write (prtunit,*) 'Indices of linearly dependent columns    = ', ip(krank+1:n)
    end if
    write (prtunit,*) '2-norm of residual vector ||a*x-b||_2    = ', rnorm
!
    write (prtunit,*)
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for solving a linear least squares problem with a ', &
       m, ' by ', n,' real coefficient matrix is ', elapsed_time, ' seconds'
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, b, x, diagr, beta, tau, ip )
!
!
! END OF PROGRAM ex3_partial_qr_cmp
! =================================
!
end program ex3_partial_qr_cmp
