program test_bd_deflate2
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines SELECT_SINGVAL_CMP3, BD_DEFLATE and BD_DEFLATE2
!   in module SVD_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for checking the accuracy
!    of subroutines SELECT_SINGVAL_CMP3, BD_DEFLATE and BD_DEFLATE2.
!
! LATEST REVISION : 14/02/2022
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,true, false, safmin,  &
                         zero, half, one, two, five, seven, ten, c30, c100, c4000,    &
                         qr_cmp, ortho_gen_qr, eigval_sort, select_singval_cmp3,      &
                         bd_deflate2, norm, arth, ifirstloc
!   
#ifdef _MATMUL
    use Utilities,       only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Utilities,       only : transpose=>transpose2
#endif
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, max_qr_steps IS NUMBER OF QR SWEEPS FOR COMPUTING SINGULAR VECTORS.
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, nsing_def=5, max_qr_steps=8
!   
! def_conda IS THE CONDITION NUMBER OF THE GENERATED MATRIX IN SOME CASES.
!   
    real(stnd), parameter  :: fudge=c4000, def_conda=50000._stnd
!
    character(len=*), parameter :: name_proc='bd_deflate2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err, eps, ulp, abstol,    &
                                               shift, thr, tol, tmp, norma, conda
    real(stnd), allocatable, dimension(:)   :: eigval, singval, singval2, d, e, diagr, beta, tauo
    real(stnd), allocatable, dimension(:,:) :: a, a2, leftvec, rightvec, leftvec2, rightvec2, res, ra, p
!
    integer      :: iok
    integer(i4b) :: i, j, k, l, n, nr, m, o, q, nsing, nsing_old, success, notdone
!
    logical(lgl) :: gen_p, gen_o, two_stage, ortho, failure, failure2, failure_bd, details
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
!
!   SPECIFY A TOLERANCE FOR SELECTING THE SINGULAR TRIPLETS IN THE TESTS OF ACCURACY.
!
    tol = 0.001_stnd
!
!   ESTIMATE SINGULAR VAlUES TO MAXIMUM ACCURACY.
!
    abstol = sqrt( safmin )
!
!   SPECIFY A POSITIVE SHIFT FOR SINGULAR VALUES, EVENTUALLY.
!
    shift = zero
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    gen_o = false
    gen_p = true
!
    do n = n1, n2, n3
!
        gen_o = .not.gen_o
        gen_p = .not.gen_p
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( ra(n,n), eigval(n), singval(n), singval2(n), d(n), e(n),    &
                  diagr(n), beta(n), tauo(n), rightvec(n,n), rightvec2(n,n),  &
                  p(n,n), stat=iok  )
!
        if ( iok/=0 ) then
!
            if ( allocated(ra) )        deallocate( ra )
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(singval2) )  deallocate( singval2 )
            if ( allocated(d) )         deallocate( d )
            if ( allocated(e) )         deallocate( e )
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
            if ( allocated(tauo) )      deallocate( tauo )
            if ( allocated(rightvec) )  deallocate( rightvec )
            if ( allocated(rightvec2) ) deallocate( rightvec2 )
            if ( allocated(p) )         deallocate( p )
!
            cycle
!
        end if
!
        do k = 1_i4b, 3_i4b
!
            select case( k )
!
                case( 1_i4b )
!
!                   TEST 1: SVD OF a(:nr,:n) WITH nr=n .
!
                    nr = n
!
                case( 2_i4b )
!
!                   TEST 2: SVD OF a(:nr,:n) WITH nr>=n AND nr NOT MUCH LARGER THAN n .
!
                    nr = n + 1_i4b
!
                case( 3_i4b )
!
!                   TEST 3: SVD OF a(:nr,:n) WITH nr>=n AND nr MUCH LARGER THAN n .
!
                    nr = n + n
!
            end select
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( a(nr,n), a2(nr,n), res(nr,n),             &
                      leftvec(nr,n), leftvec2(nr,n),  stat=iok  )
!
            if ( iok/=0 ) then
!
                if ( allocated(a) )         deallocate( a )
                if ( allocated(a2) )        deallocate( a2 )
                if ( allocated(res) )       deallocate( res )
                if ( allocated(leftvec) )   deallocate( leftvec )
                if ( allocated(leftvec2) )  deallocate( leftvec2 )
!
                cycle
!
            end if
!
!           GENERATE A RANDOM COEFFICIENT MATRIX leftvec(:nr,:n) .
!
            call random_number( leftvec(:nr,:n) )
!
            leftvec(:nr,:n) = leftvec(:nr,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                leftvec(:nr,:n) = leftvec(:nr,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( leftvec(:nr,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( leftvec(:nr,:n), beta(:n) )
!
!           GENERATE A RANDOM COEFFICIENT MATRIX rightvec(:n,:n) .
!
            call random_number( rightvec(:n,:n) )
!
            rightvec(:n,:n) = rightvec(:n,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                rightvec(:n,:n) = rightvec(:n,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( rightvec(:n,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( rightvec(:n,:n), beta(:n) )
!
            do j = 1_i4b, 25_i4b
!
                select case( j )
!
                    case( 1_i4b )
!
!                       TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b)   = one
                        eigval(2_i4b:n) = ulp
!
                    case( 2_i4b )
!
!                       TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = one
                        eigval(n)             = ulp
!
                    case( 3_i4b )
!
!                       TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                        eigval(n)             = one
!
                    case( 4_i4b )
!
!                       TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                        eigval(1_i4b)         = ulp
                        eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                        eigval(n)             = two
!
                    case( 5_i4b )
!
!                       TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
                    case( 6_i4b )
!
!                       TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                        err = real( n-1_i4b, stnd )
!
                        do l = 1_i4b, n
                            eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                        end do
!
                    case( 7_i4b )
!
!                       TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                        eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
                    case( 8_i4b )
!
!                       TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                        eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
                    case( 9_i4b )
!
!                       TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
                    case( 10_i4b )
!
!                       TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
                    case( 11_i4b )
!
!                       TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                        call random_number( eigval(:n) )
!
                        eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
                    case( 12_i4b )
!
!                       TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
                    case( 13_i4b )
!
!                       TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 14_i4b )
!
!                       TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 15_i4b )
!
!                       TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 16_i4b )
!
!                       TEST 16: CLUSTER OF EIGENVALUES AT 1 WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        eigval(1_i4b)   = one/conda
                        eigval(2_i4b:n) = one
!
                    case( 17_i4b )
!
!                       TEST 17: CLUSTER OF EIGENVALUES AT 1/conda WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        eigval(1_i4b)   = one
                        eigval(2_i4b:n) = one/conda
!
                    case( 18_i4b )
!
!                       TEST 18: ARITHMETIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        tmp = real( n-1_i4b, stnd )
                        err1 = one - one/conda
!
                        do l = 1_i4b, n
                            eigval(l) = one - ( real(l-1_i4b,stnd)/tmp )*err1
                        end do
!
                    case( 19_i4b )
!
!                       TEST 19: GEOMETRIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        tmp = real( n-1_i4b, stnd )
!
                        do l = 1_i4b, n
                            eigval(l) = conda**( -real(l-1_i4b,stnd)/tmp )
                        end do
!
                    case( 20_i4b )
!
!                       TEST 20: RANDOM DISTRIBUTION OF EIGENVALUES IN THE INTERVAL 1/conda TO 1
!                                WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        call random_number( beta(:n) )
!
                        tmp = log( one/conda )

                        beta(:n) = tmp + beta(:n)*abs( tmp )
                        eigval(:n)    = exp( beta(:n) )
!
                    case( 21_i4b )
!
!                       TEST 21: VERY SLOW DECAY OF EIGENVALUES.
!
                        conda = real( n - 1_i4b, stnd )
!
                        do l = 1_i4b, n
!
                            tmp = real( l-1_i4b, stnd )
!
                            eigval(l) = exp( -tmp/conda )
!        
                        end do
!
                    case( 22_i4b )
!
!                       TEST 22: SLOW DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = one/( tmp*tmp )
!        
                        end do
!
                    case( 23_i4b )
!
!                       TEST 23: FAST DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = exp( -tmp/seven )
!        
                        end do
!
                    case( 24_i4b )
!
!                       TEST 24: S-SHAPE DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
                        end do
!
                    case default
!
!                       DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                        call random_number( eigval(:n) )
!
                        call random_number( tmp )
!
                        if ( tmp>safmin ) then
                             eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                        end if
!
                end select
!
#ifdef _F2003
                if ( ieee_support_datatype( eigval ) ) then
                    if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
                end if
#endif
!
!               COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!
                singval(:n) = sqrt( eigval(:n) )
!
!               SORT THE SINGULAR VALUES.
!
                call eigval_sort( sort, singval(:n) )
!
!               APPLY POSITIVE SHIFT FOR SINGULAR VALUES, EVENTUALLY.
!
                singval(:n) = singval(:n) + shift
!
                where ( singval(:n)==zero )
!
                    eigval(:n) = one
!
                elsewhere
!
                    eigval(:n) = singval(:n)
!
                end where
!
!               FORM A MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
                p(:n,:n) = transpose( rightvec(:n,:n) )
!
                do i = 1_i4b, n
!
                    p(:n,i) = singval(:n)*p(:n,i)
!
                end do
!
                a(:nr,:n) = matmul( leftvec(:nr,:n), p(:n,:n) )
!
!               COMPUTE THE NORM OF THE MATRIX.
!
!                norma = norm( a(:nr,:n) )
                norma = norm( singval(:n) )
!
                do q = 1_i4b, 2_i4b
!
!                   SAVE THE ORIGINAL MATRIX.
!
                    a2(:nr,:n) = a(:nr,:n)
!
!                   DETERMINE IF A TWO-STAGE TRANSFORMATION IS USED FOR REDUCTION.
!
                    two_stage = q==2_i4b
!
!                   REDUCE THE MATRIX a TO BIDIAGONAL FORM AND
!                   COMPUTE ALL SINGULAR VALUES OF THE MATRIX a TO HIGH ACCURACY.
!
                    if ( two_stage ) then
!
                        if ( gen_o ) then
!
                            call select_singval_cmp3( a2, ra, nsing, singval2, failure=failure, sort=sort, &
                                                      d=d, e=e, abstol=abstol, p=p, gen_p=gen_p,           &
                                                      failure_bd=failure_bd )
!
                        else
!
                            call select_singval_cmp3( a2, ra, nsing, singval2, failure=failure, sort=sort,  &
                                                      d=d, e=e, abstol=abstol, tauo=tauo, p=p, gen_p=gen_p, &
                                                      failure_bd=failure_bd )
!
                        end if
!
                    else
!
                        call select_singval_cmp3( a2, nsing, singval2, failure=failure, sort=sort, &
                                                  d=d, e=e, abstol=abstol, p=p, gen_p=gen_p,       &
                                                  failure_bd=failure_bd )
!
                    end if
!
                    if ( failure ) cycle
!
!                   COMPUTE SINGULAR VECTORS OF THE MATRIX.
!
                    do o = 1_i4b, 2_i4b
!
                        if ( o==1_i4b ) then
                            nsing = min( n, nsing_def )
                        else
                            nsing = n
                        end if
!
                        ortho = true
!
                        do i = 1_i4b, 2_i4b
!
!                           COMPUTE SINGULAR VECTORS OF THE MATRIX BY A DEFLATION METHOD.
!
!                           ALWAYS USE ORTHOGONALIZATION FOR CASE 24, WHICH IS DIFFICULT FOR DEFLATION.
!
                            if ( j/=24_i4b ) then
                                ortho = .not.ortho
                            end if
!
!                           ALWAYS USE ORTHOGONALIZATION FOR CASES 6, 20 AND 24 WHEN n IS LARGE,
!                           WHICH ARE DIFFICULT FOR DEFLATION (IN SINGLE PRECISION).
!
!                            if ( (j/=24_i4b .and. j/=20_i4b .and. j/=6_i4b) .or. n<=100_i4b ) then
!                                ortho = .not.ortho
!                            end if
!
                            if ( two_stage ) then
!
                                if ( gen_o ) then
!
                                    call bd_deflate2( a2(:nr,:n), ra(:n,:n), p(:n,:n), d(:n), e(:n),   &
                                                      singval2(:nsing), leftvec2(:nr,:nsing),          &
                                                      rightvec2(:n,:nsing), failure=failure2,          &
                                                      max_qr_steps=max_qr_steps, ortho=ortho           )
!
                                else
!
                                    call bd_deflate2( a2(:nr,:n), ra(:n,:n), p(:n,:n), d(:n), e(:n),   &
                                                      singval2(:nsing), leftvec2(:nr,:nsing),          &
                                                      rightvec2(:n,:nsing), failure=failure2,          &
                                                      tauo=tauo(:n), max_qr_steps=max_qr_steps,        &
                                                      ortho=ortho  )
!
                                end if
!
                            else
!
                                call bd_deflate2( a2(:nr,:n), p(:n,:n), d(:n), e(:n), singval2(:nsing),     &
                                                  leftvec2(:nr,:nsing), rightvec2(:n,:nsing),               &
                                                  failure=failure2,  max_qr_steps=max_qr_steps, ortho=ortho )
!
                            end if
!
                            m = m + 1_i4b
!
!                           DETERMINE THE NUMBER OF SINGULAR VECTORS FOR CHECKING ACCURACY
!                           OF bd_deflate2 SUBROUTINE.
!
                            thr = max( tol*singval2(1_i4b), safmin )
!
                            nsing_old = nsing
!
                            nsing = ifirstloc( logical( singval2(:nsing_old)<=thr, lgl ) ) - 1_i4b
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                           WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                           v THE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                            res(:nr,:nsing) = matmul( a(:nr,:n), rightvec2(:n,:nsing) )         &
                                             - leftvec2(:nr,:nsing)*spread(singval2(:nsing),1,nr)
!
                            if ( norma/=zero ) then
                                err1 = maxval( norm(res(:nr,:nsing), dim=2_i4b) )/( norma*real(nr,stnd) )
                            else
                                err1 = maxval( norm(res(:nr,:nsing), dim=2_i4b) )/real(nr,stnd)
                            end if
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                           WHERE u ARE LEFT SINGULAR VECTORS OF THE MATRIX.
!
                            res(:nsing,:nsing) = matmul( transpose( leftvec2(:nr,:nsing) ), leftvec2(:nr,:nsing) )
!
                            do l = 1_i4b, nsing
                                res(l,l) = one - res(l,l)
                            end do
!
                            err2 = maxval( norm(res(:nsing,:nsing), dim=2_i4b) )/real(nr,stnd)
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!                           WHERE v ARE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                            res(:nsing_old,:nsing_old) = matmul( transpose( rightvec2(:n,:nsing_old) ), rightvec2(:n,:nsing_old) )
!
                            do l = 1_i4b, nsing_old
                                res(l,l) = one - res(l,l)
                            end do
!
                            err3 = maxval( norm(res(:nsing_old,:nsing_old), dim=2_i4b) )/real(n,stnd)
!
                            err = max( err1, err2, err3 )
!
                            if ( err<=eps ) then
!
                                success = success + 1_i4b
!
                                if ( details ) then
!
                                    write (prtunit,*) 'Test ',m,' is correct'
!
                                    write (prtunit,*) 'nr= ', nr,' , nc= ', n,' , case= ', j,          &
                                                      ' , nsing_old= ', nsing_old,' , nsing= ', nsing, &
                                                      ' , two_stage= ', two_stage,' , gen_o= ', gen_o, &
                                                      ' , gen_p= ', gen_p, ' , ortho= ', ortho
!
!                                   COMPUTE ABSOLUTE ERRORS OF SINGULAR VALUES.
!
                                    write (prtunit,*) 'Maximum absolute error of the computed singular values = ', &
                                                      maxval( abs( singval2(:nsing_old)-singval(:nsing_old) ) )
!
!                                   COMPUTE RELATIVE ERRORS OF SINGULAR VALUES.
!
                                    write (prtunit,*) 'Maximum relative error of the computed singular values = ', &
                                     maxval( abs( ( singval2(:nsing_old)-singval(:nsing_old) )/eigval(:nsing_old) ) )
!
!                                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                                   WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                                   v THE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                                    write (prtunit,*) 'Accuracy of the computed singular triplets             = ', &
                                                      err1/ulp, ' (must be less than :', fudge, ' )'
!
!                                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                                   WHERE u ARE LEFT SINGULAR VECTORS OF THE MATRIX.
!
                                    write (prtunit,*) 'Orthogonality of the computed left singular vectors    = ', &
                                                      err2/ulp, ' (must be less than :', fudge, ' )'
!
!                                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!                                   WHERE v ARE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                                    write (prtunit,*) 'Orthogonality of the computed right singular vectors   = ', &
                                                      err3/ulp, ' (must be less than :', fudge, ' )'
!
                                    write (prtunit,*)
!
                                end if
!
                            else if ( failure2 ) then
!
                                notdone = notdone + 1_i4b
!
                            else
!
                                write (prtunit,*) 'Test ',m,' is incorrect'
!
                                write (prtunit,*) '=> ',err,' > ',eps,' , nr= ', nr,' , nc= ', n,' , case= ', j,    &
                                                  ' , nsing_old= ', nsing_old,' , nsing= ', nsing,                  &
                                                  ' , two_stage= ', two_stage,' , gen_o= ', gen_o,                  &
                                                  ' , gen_p= ', gen_p, ' , ortho= ', ortho,                         &
                                                  ' , failure= ', failure2,' , failure_bd= ', failure_bd
!
!                               COMPUTE ABSOLUTE ERRORS OF SINGULAR VALUES.
!
                                write (prtunit,*) 'Maximum absolute error of the computed singular values = ', &
                                                  maxval( abs( singval2(:nsing_old)-singval(:nsing_old) ) )
!
!                               COMPUTE RELATIVE ERRORS OF SINGULAR VALUES.
!
                                write (prtunit,*) 'Maximum relative error of the computed singular values = ', &
                                 maxval( abs( ( singval2(:nsing_old)-singval(:nsing_old) )/eigval(:nsing_old) ) )
!
!                               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                               WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                               v THE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                                write (prtunit,*) 'Accuracy of the computed singular triplets             = ', &
                                                  err1/ulp, ' (must be less than :', fudge, ' )'
!
!                               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                               WHERE u ARE LEFT SINGULAR VECTORS OF THE MATRIX.
!
                                write (prtunit,*) 'Orthogonality of the computed left singular vectors    = ', &
                                                  err2/ulp, ' (must be less than :', fudge, ' )'
!
!                               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!                               WHERE v ARE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                                write (prtunit,*) 'Orthogonality of the computed right singular vectors   = ', &
                                                  err3/ulp, ' (must be less than :', fudge, ' )'
!
                                write (prtunit,*)
!
                            end if
!
                        end do
!
                    end do
!
                end do
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, a2, leftvec, leftvec2, res )
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( ra, eigval, singval, singval2, d, e, diagr, &
                    beta, tauo, rightvec, rightvec2, p          )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    i = success + notdone
!
    if ( m/=i ) then
        write (prtunit,*) m - i, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_bd_deflate2
! ===============================
!
end program test_bd_deflate2
