program test_comp_eof3
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine COMP_EOF3 in module Mul_Stat_Procedures.
!                                                                              
! LATEST REVISION : 01/02/2022
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, false, one, seven, c30, c1_e6, comp_eof3, random_seed_,  &
                         random_number_, gen_random_mat, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m IS THE NUMBER OF OBSERVATIONS AND VARIABLES
! AND neig THE NUMBER OF COMPUTED EIGENVECTORS.
!
    integer(i4b), parameter :: prtunit=6, m=20, neig=3
!   
! conda IS THE CONDITION NUMBER OF THE GENERATED MATRIX IN SOME CASES.
!   
    real(stnd), parameter  :: conda=c1_e6
!
    character(len=*), parameter :: name_proc='comp_eof3'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err_eig, eps, tmp, tmp2
    real(stnd), allocatable, dimension(:,:) :: x, sscp, eigvec, res, eigval
    real(stnd), allocatable, dimension(:)   :: singval0
!
    integer      :: iok
    integer(i4b) :: mat_type, i, l, success, maxiter=4
!
    logical(lgl) :: failure
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon(eps) )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( x(m,m), sscp(m,m), eigvec(m,neig), res(m,neig),  &
              eigval(m,2), singval0(m), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    success = 0_i4b
    l = 0_i4b
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
    do mat_type = 1_i4b, 8_i4b
!
!      GENERATE A m-BY-p RANDOM REAL MATRIX x WITH A SPECIFIED DISTRIBUTION
!      OF SINGULAR VALUES.
!
!      mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!      mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!      mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!      mat_type = 4  -> VERY SLOW DECAY OF SINGULAR VALUES
!      mat_type = 5  -> STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda
!      mat_type = 6  -> GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!      mat_type = 7  -> ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda
!      mat_type > 7  -> UNIFORM DISTRIBUTION OF SINGULAR VALUES
!
!      GENERATE SINGULAR VALUES.
!
       select case( mat_type )
!
           case( 1_i4b )
!
!              SLOW DECAY OF SINGULAR VALUES.
!
               do i = 1_i4b, m
!
                   tmp = real( i, stnd )
!
                   singval0(i) = one/( tmp*tmp )
!        
               end do
!
           case( 2_i4b )
!
!               FAST DECAY OF SINGULAR VALUES.
!
                do i = 1_i4b, m
!
                    tmp = real( i, stnd )
!
                    singval0(i) = exp( -tmp/seven )
!        
                end do
!
            case( 3_i4b )
!
!               S-SHAPED DECAY OF SINGULAR VALUES.
!
                do i = 1_i4b, m
!
                    tmp = real( i, stnd )
!
                    singval0(i) = 0.0001_stnd + one/( one + exp( tmp - c30 ) )
!        
                end do
!
            case( 4_i4b )
!
!               VERY SLOW DECAY OF SINGULAR VALUES.
!
                tmp = real( m - 1_i4b, stnd )
!
                do i = 1_i4b, m
!
                    tmp2 = real( i - 1_i4b, stnd )
!
                    singval0(i) = exp( -tmp2/tmp )
!        
                end do
!
            case( 5_i4b )
!
!               STRONGLY CLUSTERED SINGULAR VALUES AT 1 WITH CONDITION NUMBER conda.
!
                singval0(:m-1_i4b) = one
                singval0(m) = one/conda
!
            case( 6_i4b )
!
!               GEOMETRIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!
                tmp  = real( m - 1_i4b, stnd )
!
                do i = 1_i4b, m
!
                    tmp2 = real( i - 1_i4b, stnd )
!
                    singval0(i) = conda**( -tmp2/tmp )
!        
                end do
!
            case( 7_i4b )
!
!               ARITHMETIC DISTRIBUTION OF SINGULAR VALUES WITH CONDITION NUMBER conda.
!
                tmp  = real( m - 1_i4b, stnd )
                tmp2 = one - one/conda
!
                do i = 1_i4b, m
!
                    singval0(i) = one - (real( i - 1_i4b, stnd )/tmp)*tmp2
!        
                end do
!
            case default
!
!               UNIFORM DISTRIBUTION OF SINGULAR VALUES.
!
                call random_number_( singval0(:m) )
!
        end select
!
!       GENERATE A m-BY-p MATRIX x OF RANK mp .
!
        call gen_random_mat( singval0(:m), x )
!
        do i = 1_i4b, 2_i4b
!
!           COMPUTE EOFs.
!
            call comp_eof3( x(:m,:m), i, failure, xeigval=eigval, xeigvec=eigvec, maxiter=maxiter )
!
!           COMPUTE THE SSCP MATRIX  .
!
            if ( i==1_i4b ) then
                sscp = matmul( x, transpose( x ) )/m
            else
                sscp = matmul( transpose( x ), x  )/m
            end if
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION sscp*eigvec - eigvec*diag(eigvval(:neig,1))
!           WHERE eigvval ARE THE EIGENVALUES AND eigvec THE EIGENVECTORS OF sscp.
!
            res(:m,:neig) = matmul(sscp,eigvec) - eigvec(:m,:neig)*spread(eigval(:neig,1_i4b),1,m)
!
            err_eig = sum( abs( res(:m,:neig) ) )/sum( eigval(:neig,1_i4b) )
!
!           PRINT THE RESULTS OF THE TEST.
!
            l = l + 1_i4b
!
            if ( err_eig<=eps .and. .not.failure ) then
!
                success = success + 1_i4b
!
            else
!
                write (prtunit,*) 'Test ',l,' is incorrect'
                write (prtunit,*) ' mat_type = ', mat_type, ', i = ', i
!
                if ( failure ) then
                    write (prtunit,*) '=> failure  = ', failure
                end if
!
                if ( err_eig>eps ) then
                    write (prtunit,*) '=> err_eig  =  ', err_eig,' > ', eps
                end if
!
                write (prtunit,*)
!
            end if
!
        end do
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( x, sscp, eigvec, res, eigval, singval0 )
!
    if ( success/=l ) then
        write (prtunit,*) l - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_comp_eof3
! =============================
!
end program test_comp_eof3
