program test_comp_eof_miss3
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine COMP_EOF_MISS3 in module Mul_Stat_Procedures.
!
! LATEST REVISION : 29/09/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, zero, comp_eof_miss3
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m IS THE NUMBER OF OBSERVATIONS AND THE NUMBER OF VARIABLES,
! neig IS THE NUMBER OF EIGENVECTORS, WHICH WILL BE COMPUTED BY maxiter INVERSE ITERATIONS.
!
    integer(i4b), parameter :: prtunit=6, m=50, neig=3, maxiter=3
!   
! miss IS THE MISSING VALUE INDICATOR IN THE DATA.
!   
    real(stnd), parameter  :: miss=-999.99_stnd
!
    character(len=*), parameter :: name_proc='comp_eof_miss3'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                    :: err_eig, eps
    real(stnd), dimension(m,m)    :: sscp, x
    real(stnd), dimension(m,neig) :: eigvec, res
    real(stnd), dimension(m,2)    :: eigval
!
    integer(i4b) :: i, j, k, success
!
    logical(lgl)                 :: failure
    logical(lgl), dimension(m)   :: maskvec
    logical(lgl), dimension(m,m) :: xmask, mask
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon(eps) )
!
!   GENERATE A RANDOM OBSERVATION ARRAY x WITH MISSING VALUES.
!
    call random_number( x )
!
    xmask = x>0.05_stnd
    where ( .not.xmask ) x = miss
!
    success = 0_i4b
!
    do i = 1_i4b, 2_i4b
!
        call comp_eof_miss3( x(:m,:m), miss, i, failure,                    &
                             xeigval=eigval, xeigvec=eigvec, maxiter=maxiter )
!
        if ( i==2_i4b ) then
            sscp(:m,:m)  = transpose( x(:m,:m) )
            x(:m,:m)     = sscp(:m,:m)
            mask(:m,:m)  = transpose( xmask(:m,:m) )
            xmask(:m,:m) = mask(:m,:m)
        end if
!
!       COMPUTE THE SSCP MATRIX.
!
!BUG:   SOME COMPILERS (eg PGF90) DO NOT IMPLEMENT THIS SECTION OF CODE
!       CORRECTLY. THUS, IT IS SAFER TO USE ANOTHER SECTION OF CODE.
!
!        sscp(:m,:m) = zero
!
!        do k = 1_i4b, m
!
!            mask(k:m,:) = xmask(k:m,:) .and. spread(xmask(k,:),dim=1_i4b, ncopies=m-k+1_i4b )
!
!            sscp(k:m,k) = sum( x(k:m,:)*spread(x(k,:),dim=1_i4b, ncopies=m-k+1_i4b ),         &
!                          mask=mask(k:m,:), dim=2_i4b )
!            sscp(k:m,k) = sscp(k:m,k)/count( mask(k:m,:), dim=2_i4b )
!            sscp(k,k+1_i4b:m) = sscp(k+1_i4b:m,k)
!
!        end do
!
        sscp(:m,:m) = zero
!
        do k = 1_i4b, m
!
            do j = k, m
                maskvec(:m)  = xmask(j,:m) .and. xmask(k,:m)
                sscp(j,k)    = sum( x(j,:m)*x(k,:m), mask=maskvec(:m) )
                sscp(j,k)    = sscp(j,k)/count( maskvec(:m) )            
            end do
!
            sscp(k,k+1_i4b:m) = sscp(k+1_i4b:m,k)
!
        end do
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION sscp*eigvec - eigvec*diag(eigvval(:neig,1))
!       WHERE eigvval ARE THE EIGENVALUES AND eigvec THE EIGENVECTORS OF sscp.
!
        res(:m,:neig) = matmul(sscp,eigvec) - eigvec(:m,:neig)*spread(eigval(:neig,1_i4b),1,m)
!
        err_eig = sum( abs( res(:m,:neig) ) )/sum( eigval(:neig,1_i4b) )
!
!       PRINT THE RESULTS OF THE TESTS.
!
        if ( err_eig<=eps .and. .not.failure ) then
!
            success = success + 1_i4b
!
        else
!
            write (prtunit,*) 'Test ',i,' is incorrect'
            write (prtunit,*) ', neig = ', neig,', maxiter = ', maxiter
!
            if ( failure ) then
                write (prtunit,*) '=> failure  = ', failure
            end if
!
            if ( err_eig>eps ) then
                write (prtunit,*) '=> err_eig  =  ', err_eig,' > ', eps
            end if
!
            write (prtunit,*)
!
        end if
!
    end do
!
    if ( success/=2_i4b ) then
        write (prtunit,*) 2_i4b - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_comp_eof_miss3
! ==================================
!
end program test_comp_eof_miss3
