program test_id_cmp
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines ID_CMP in module Random and
!   ORTHO_GEN_QR in module QR_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for checking
!    the accuracy of subroutines ID_CMP and ORTHO_GEN_QR.
!
! LATEST REVISION : 03/02/2022
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, n1_def, n2_def, n3_def, safmin,        &
                         zero, half, one, two, five, seven, ten, c30, c100, c1000,    &
                         id_cmp, qr_cmp, ortho_gen_qr, norm, arth, reorder
!   
#ifdef _MATMUL
    use Utilities, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Utilities, only : transpose=>transpose2
#endif
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT.
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=c1000, def_conda=50000._stnd
!
    character(len=*), parameter :: name_proc='id_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err4, err, eps, ulp, tmp,   &
                                               shift, tol, conda, norma, normr
    real(stnd), allocatable, dimension(:)   :: eigval, singval, diagr, beta
    real(stnd), allocatable, dimension(:,:) :: a, leftvec, rightvec, rightvect, res, t, c, v
!
    integer      :: iok
    integer(i4b) :: i, j, k, l, n, nr, nc, m, p, krank0, nid, nt, idep, nover, blk_size, success
    integer(i4b), allocatable, dimension(:) :: ip
!   
    logical(lgl) :: details, test_dep, random_qr
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SPECIFY IF A RANDOMIZED OR DETERMINISTIC COLUMN ID ALGORITHM IS USED.
!
    random_qr = true
!
!   CHOOSE TUNING PARAMETERS FOR THE RANDOMIZED ID ALGORITHM.
!
!   DETERMINE THE BLOCKSIZE PARAMETER IN THE RANDOMIZED COLUMN ID ALGORITHM.
!
    blk_size = 20_i4b
!
!   DETERMINE THE OVERSAMPLING SIZE PARAMETER IN THE RANDOMIZED COLUMN ID ALGORITHM.
!
    nover = 10_i4b
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( ulp )
    eps    = fudge*ulp
!
!   SPECIFY EVENTUALLY A SHIFT TO AVOID ZERO SINGULAR VALUE
!   OR SHIFT THE SINGULAR VALUES DISTRIBUTION.
!
    shift = 0.01_stnd
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       DETERMINE THE TARGET RANK OF THE COLUMN INTERPOLATIVE DECOMPOSITION, WHICH IS SOUGHT.
!
        nid = n/2_i4b
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigval(n), singval(n),       &
                  diagr(n), beta(n), stat=iok  )
!
        if ( iok/=0 ) then
!
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
!
            cycle
!
        end if
!
        do k = 1_i4b, 5_i4b
!
            select case( k )
!
                case( 1_i4b )
!
!                   TEST 1: ID DECOMPOSITION OF a(:nr,:nc) WITH nr=nc .
!
                    nr = n
                    nc = n
!
                case( 2_i4b )
!
!                   TEST 2: ID DECOMPOSITION OF a(:nr,:nc) WITH nr>=nc AND nr NOT MUCH LARGER THAN nc .
!
                    nr = n + 1_i4b
                    nc = n
!
                case( 3_i4b )
!
!                   TEST 3: ID DECOMPOSITION OF a(:nr,:nc) WITH nr>=nc AND nr MUCH LARGER THAN nc .
!
                    nr = n + n
                    nc = n
!
                case( 4_i4b )
!
!                   TEST 4: ID DECOMPOSITION  OF a(:nr,:nc) WITH nc>=nr AND nc NOT MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + 1_i4b
!
                case( 5_i4b )
!
!                   TEST 5: ID DECOMPOSITION OF a(:nr,:nc) WITH nc>=nr AND nc MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + n
!
            end select
!
!           ALLOCATE WORK ARRAYS.
!
            nt  = nc - nid
!
            j = max( nr, nc )
!
            allocate( a(nr,j), res(nr,j), t(nid,nt), c(nr,nid), v(nid,nc),   &
                     ip(nc), leftvec(nr,n), rightvec(nc,n), rightvect(n,nc), &
                     stat=iok )
!
            if ( iok/=0 ) then
!
                if ( allocated(a) )         deallocate( a )
                if ( allocated(res) )       deallocate( res )
                if ( allocated(t) )         deallocate( t )
                if ( allocated(c) )         deallocate( c )
                if ( allocated(v) )         deallocate( v )
                if ( allocated(ip) )        deallocate( ip )
                if ( allocated(leftvec) )   deallocate( leftvec )
                if ( allocated(rightvec) )  deallocate( rightvec )
                if ( allocated(rightvect) ) deallocate( rightvect )
!
                cycle
!
            end if
!
!           GENERATE A RANDOM COEFFICIENT MATRIX leftvec(:nr,:n) .
!
            call random_number( leftvec(:nr,:n) )
!
            leftvec(:nr,:n) = leftvec(:nr,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                leftvec(:nr,:n) = leftvec(:nr,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( leftvec(:nr,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( leftvec(:nr,:n), beta(:n) )
!
!           GENERATE A RANDOM COEFFICIENT MATRIX rightvec(:nc,:n) .
!
            call random_number( rightvec(:nc,:n) )
!
            rightvec(:nc,:n) = rightvec(:nc,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                rightvec(:nc,:n) = rightvec(:nc,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( rightvec(:nc,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( rightvec(:nc,:n), beta(:n) )
!
            do j = 1_i4b, 25_i4b
!
                select case( j )
!
                    case( 1_i4b )
!
!                       TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b)   = one
                        eigval(2_i4b:n) = ulp
!
                    case( 2_i4b )
!
!                       TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = one
                        eigval(n)             = ulp
!
                    case( 3_i4b )
!
!                       TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                        eigval(n)             = one
!
                    case( 4_i4b )
!
!                       TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                        eigval(1_i4b)         = ulp
                        eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                        eigval(n)             = two
!
                    case( 5_i4b )
!
!                       TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
                    case( 6_i4b )
!
!                       TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                        err = real( n-1_i4b, stnd )
!
                        do l = 1_i4b, n
                            eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                        end do
!
                    case( 7_i4b )
!
!                       TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                        eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
                    case( 8_i4b )
!
!                       TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                        eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
                    case( 9_i4b )
!
!                       TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
                    case( 10_i4b )
!
!                       TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
                    case( 11_i4b )
!
!                       TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                        call random_number( eigval(:n) )
!
                        eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
                    case( 12_i4b )
!
!                       TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
                    case( 13_i4b )
!
!                       TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + one
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 14_i4b )
!
!                       TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + one
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 15_i4b )
!
!                       TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + one
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 16_i4b )
!
!                       TEST 16: CLUSTER OF EIGENVALUES AT 1 WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        eigval(1_i4b)   = one/conda
                        eigval(2_i4b:n) = one
!
                    case( 17_i4b )
!
!                       TEST 17: CLUSTER OF EIGENVALUES AT 1/conda WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        eigval(1_i4b)   = one
                        eigval(2_i4b:n) = one/conda
!
                    case( 18_i4b )
!
!                       TEST 18: ARITHMETIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        tmp = real( n-1_i4b, stnd )
                        err = one - one/conda
!
                        do l = 1_i4b, n
                            eigval(l) = one - ( real(l-1_i4b,stnd)/tmp )*err
                        end do
!
                    case( 19_i4b )
!
!                       TEST 19: GEOMETRIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        tmp = real( n-1_i4b, stnd )
!
                        do l = 1_i4b, n
                            eigval(l) = conda**( -real(l-1_i4b,stnd)/tmp )
                        end do
!
                    case( 20_i4b )
!
!                       TEST 20: RANDOM DISTRIBUTION OF EIGENVALUES IN THE INTERVAL 1/conda TO 1
!                                WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        call random_number( beta(:n) )
!
                        tmp = log( one/conda )

                        beta(:n) = tmp + beta(:n)*abs( tmp )
                        eigval(:n) = exp( beta(:n) )
!
                    case( 21_i4b )
!
!                       TEST 21: VERY SLOW DECAY OF EIGENVALUES.
!
                        conda = real( n - 1_i4b, stnd )
!
                        do l = 1_i4b, n
!
                            tmp = real( l-1_i4b, stnd )
!
                            eigval(l) = exp( -tmp/conda )
!        
                        end do
!
                    case( 22_i4b )
!
!                       TEST 22: SLOW DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = one/( tmp*tmp )
!        
                        end do
!
                    case( 23_i4b )
!
!                       TEST 23: FAST DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = exp( -tmp/seven )
!        
                        end do
!
                    case( 24_i4b )
!
!                       TEST 24: S-SHAPED DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
                        end do
!
                    case default
!
!                       DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                        call random_number( eigval(:n) )
!
                        call random_number( tmp )
!
                        if ( tmp>safmin ) then
                             eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                        end if
!
                end select
!
#ifdef _F2003
                if ( ieee_support_datatype( eigval ) ) then
                    if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
                end if
#endif
!
                m = m + 1_i4b
!
!               COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!
                singval(:n) = sqrt( eigval(:n) )
!
                singval(:n) = singval(:n) + shift
!
                krank0 = count( singval(:n)/=zero )
!
!               FORM A MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
                rightvect(:n,:nc) = transpose( rightvec(:nc,:n) )
!
                do i = 1_i4b, nc
!
                    rightvect(:n,i) = rightvect(:n,i)*singval(:n)
!
                end do
!
                a(:nr,:nc) = matmul( leftvec(:nr,:n), rightvect(:n,:nc) )
!
                test_dep = nc==n .and. n>5_i4b
!
                if ( test_dep ) then
!
!                   GENERATE A DEFICIENT COEFFICIENT MATRIX a(:m,:n) WITH RANK
!                   AT MOST n - 1_i4b .
!
                    idep = min( nc, 5_i4b ) 
                    a(:nr,idep) = a(:nr,1_i4b) + a(:nr,nc)
!
!                   COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
                    norma = norm( a(:nr,:nc) )
!
                else
!
!                   COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
                    norma = norm( singval(:n) )
!
                end if
!
!               MAKE A COPY OF DATA MATRIX.
!
                res(:nr,:nc) = a(:nr,:nc)
!
!               COMPUTE RANDOMIZED OR DETERMINISTIC ID DECOMPOSITION OF RANDOM DATA MATRIX.
!
                random_qr = .not.random_qr
!
!               SET TOLERANCE FOR CHECKING THE RANK OF THE PARTIAL ID APPROXIMATION IN THE SUBROUTINE.
!
                tol = real( n, stnd)*ulp
!
                call id_cmp( a(:nr,:nc), ip(:nc), t(:nid,:nt), c=c(:nr,:nid),      &
                             v=v(:nid,:nc), beta=beta(:nid), rnorm=normr, tol=tol, &
                             random_qr=random_qr, blk_size=blk_size, nover=nover   )
!
!               COMPUTE RELATIVE ERROR OF THE COLUMN ID APPROXIMATION.
!
                err1 = normr/norma
!
!               CHECK ACCURACY OF THE FROBENIUS NORM OF THE RESIDUAL MATRIX.
!
                res(:nr,:nc) = res(:nr,:nc) - matmul( c(:nr,:nid), v(:nid,:nc) )
!
                if ( normr<=one ) then
!
                    err2 = abs( norm( res(:nr,:nc) ) - normr )
!
                else
!
                    err2 = abs( norm( res(:nr,:nc) )/normr - one )
!
                end if
!
!               GENERATE ORTHOGONAL MATRIX Q OF PARTIAL QR DECOMPOSITION OF DATA MATRIX a .
!
                call ortho_gen_qr( a(:nr,:nr), beta(:nid) )
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION c(:nr,:nid) - Q(:nr,:nid)*(Q(:nr,:nid)'*c(:nr,:nid)).
!
                t(:nid,:nid) = matmul( transpose(a(:nr,:nid)), c(:nr,:nid) )
!
                res(:nr,:nid) = abs( c(:nr,:nid) - matmul( a(:nr,:nid), t(:nid,:nid) ) )
!
                err3 = maxval( res(:nr,:nid) )/real(nr,stnd)
!
!               CHECK ORTHOGONALITY OF c(:nr,:nid) AND ITS ORTHOGONAL COMPLEMENT Q(:nr,nid+1:nc).
!
                if ( nr>nid ) then
!
                    res(:nid,nid+1_i4b:nr) = matmul( transpose(c(:nr,:nid)), a(:nr,nid+1_i4b:nr) )
!
                    err4 = maxval( abs( res(:nid,nid+1_i4b:nr) ) )/real(nr,stnd)
!
                else
!
                    err4 = zero
!
                end if
!
                err = max( err2, err3, err4 )
!
                if ( test_dep ) then
!
!                   CHECK IF LINEARLY DEPENDANT COLUMNS HAVE BEEN IDENTIFIED.
!
                    do l = nid+1_i4b, nc
!
                        p = ip(l)
!
                        if ( p==idep .or. p==1_i4b .or. p==nc ) exit
!
                    end do
!
                    test_dep = l==nc+1_i4b
!
                end if
!
                if ( err<=eps .and. .not.test_dep ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',m,' is correct'
!
                        write (prtunit,*) 'nr= ', nr,' , nc= ', nc,' , case= ', j,       &
                                          ' , krank0= ', krank0,' , nid= ', nid,         &
                                          ' , random_qr= ', random_qr,                   &
                                          ' , blk_size= ', blk_size, ' , nover= ', nover
!        
                        write (prtunit,*) 'Relative error of the column ID decomposition  &
                                           &||A - C*V||_F/||A||_F             = ', err1
!        
                        write (prtunit,*) 'Accuracy of the Frobenius norm of the  &
                                          &residual matrix                           = ', err2
!        
                        write (prtunit,*) 'Accuracy of the range of the column ID &
                                          &approximation                             = ', err3
!
                        if ( nr>nid ) then
                            write (prtunit,*) 'Orthogonality of the range of the ID approximation&
                                              & and its orthogonal complement = ', err4
                        end if
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',m,' is incorrect'
!
                    write (prtunit,*) '=> ',err,' > ',eps,' , nr= ', nr,' , nc= ', nc,       &
                                      ' , case= ', j,' , krank0= ', krank0,' , nid= ', nid,  &
                                      ' , test_dep= ', test_dep,' , random_qr= ', random_qr, &
                                      ' , blk_size= ', blk_size, ' , nover= ', nover
!        
                    write (prtunit,*) 'Relative error of the column ID decomposition  &
                                      &||A - C*V||_F/||A||_F             = ', err1
!        
                    write (prtunit,*) 'Accuracy of the Frobenius norm of the  &
                                      &residual matrix                           = ', err2
!        
                    write (prtunit,*) 'Accuracy of the range of the column ID &
                                      &approximation                             = ', err3
!
                    if ( nr>nid ) then
                        write (prtunit,*) 'Orthogonality of the range of the ID approximation&
                                          & and its orthogonal complement = ', err4
                    end if
!
                    write (prtunit,*)
!
                end if
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, leftvec, rightvec, rightvect, res, t, c, v, ip )
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, singval, diagr, beta )
!
    end do
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_id_cmp
! ==========================
!
end program test_id_cmp
