program test_rqb_cmp
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine RQB_CMP in module Random.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for checking
!    the accuracy of subroutine RQB_CMP.
!
! LATEST REVISION : 18/06/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def, true, false,        &
                         zero, half, one, two, five, seven, ten, c30, c100,          &
                         qr_cmp, ortho_gen_qr, singval_sort, rqb_cmp, norm, arth,    &
                         lamch, random_seed_, random_number_
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, nqb IS THE TARGET RANK OF THE APPROXIMATE QB,
! niter IS THE NUMBER OF RANDOM POWER OR SUBSPACE ITERATIONS PERFORMED,
! def_conda IS THE APPROXIMATE CONDITION NUMBER OF THE GENERATED MATRIX IN SOME CASES.
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, nqb=10, niter=20
!   
    real(stnd), parameter  :: fudge=ten, def_conda=50000._stnd
!
    character(len=*), parameter :: name_proc='rqb_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, eps, ulp, sqrt_ulp, &
                                               safmin, tmp, norma, conda, normb, relerr, relerr2
    real(stnd), allocatable, dimension(:)   :: eigval, singval, diagr, beta
    real(stnd), allocatable, dimension(:,:) :: a, leftvec, rightvec, rightvect, q, b, res
!
    integer      :: iok
    integer(i4b) :: i, j, k, l, n, nr, nc, m, success
    integer(i4b), allocatable, dimension(:) :: ip
!   
    logical(lgl) :: details, ortho, comp_qr, pivoting, min_norm
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   DETERMINE IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE (POWER) SUBSPACE ITERATIONS, TO AVOID LOSS OF ACCURACY
!   DUE TO ROUNDING ERRORS.
!
    ortho = true
!
!   DETERMINE IF A QB OR QR FACTORIZATION IS COMPUTED.
!
    comp_qr = true
!
!   DETERMINE IF A QR FACTORIZATION WITH COLUMN PIVOTING IS COMPUTED.
!
    pivoting = true
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp      = epsilon( err )
    sqrt_ulp = sqrt ( ulp )
    safmin   = lamch( 'S' )
!
!   SET THE TOLERANCE FOR THE ERROR OF THE RELATIVE ERROR IN FROBENIUS NORM.
!
    eps = 0.01_stnd
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigval(n), singval(n), diagr(n), beta(n), stat=iok )
!
        if ( iok/=0 ) then
!
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
!
            cycle
!
        end if
!
        do k = 1_i4b, 5_i4b
!
            select case( k )
!
                case( 1_i4b )
!
!                   TEST 1: SVD OF a(:nr,:nc) WITH nr=nc .
!
                    nr = n
                    nc = n
!
                case( 2_i4b )
!
!                   TEST 2: SVD OF a(:nr,:nc) WITH nr>=nc AND nr NOT MUCH LARGER THAN nc .
!
                    nr = n + 1_i4b
                    nc = n
!
                case( 3_i4b )
!
!                   TEST 3: SVD OF a(:nr,:nc) WITH nr>=nc AND nr MUCH LARGER THAN nc .
!
                    nr = n + n
                    nc = n
!
                case( 4_i4b )
!
!                   TEST 4: SVD OF a(:nr,:nc) WITH nc>=nr AND nc NOT MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + 1_i4b
!
                case( 5_i4b )
!
!                   TEST 5: SVD OF a(:nr,:nc) WITH nc>=nr AND nc MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + n
!
            end select
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( a(nr,nc), res(nqb,nqb),         &
                      leftvec(nr,n), rightvec(nc,n),  &
                      q(nr,nqb), b(nqb,nc), ip(nc),   &
                      rightvect(n,nc), stat=iok  )
!
            if ( iok/=0 ) then
!
                if ( allocated(a) )         deallocate( a )
                if ( allocated(res) )       deallocate( res )
                if ( allocated(leftvec) )   deallocate( leftvec )
                if ( allocated(rightvec) )  deallocate( rightvec )
                if ( allocated(rightvect) ) deallocate( rightvect )
                if ( allocated(q) )         deallocate( q )
                if ( allocated(b) )         deallocate( b )
                if ( allocated(ip) )        deallocate( ip )
!
                cycle
!
            end if
!
!           GENERATE A RANDOM COEFFICIENT MATRIX leftvec(:nr,:n) .
!
            call random_number_( leftvec(:nr,:n) )
!
            leftvec(:nr,:n) = leftvec(:nr,:n) - half
!
            call random_number_( tmp )
!
            if ( tmp>safmin ) then
                leftvec(:nr,:n) = leftvec(:nr,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( leftvec(:nr,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( leftvec(:nr,:n), beta(:n) )
!
!           GENERATE A RANDOM COEFFICIENT MATRIX rightvec(:nc,:n) .
!
            call random_number_( rightvec(:nc,:n) )
!
            rightvec(:nc,:n) = rightvec(:nc,:n) - half
!
            call random_number_( tmp )
!
            if ( tmp>safmin ) then
                rightvec(:nc,:n) = rightvec(:nc,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( rightvec(:nc,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( rightvec(:nc,:n), beta(:n) )
!
            do j = 1_i4b, 25_i4b
!
                select case( j )
!
                    case( 1_i4b )
!
!                       TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b)   = one
                        eigval(2_i4b:n) = ulp
!
                    case( 2_i4b )
!
!                       TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = one
                        eigval(n)             = ulp
!
                    case( 3_i4b )
!
!                       TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                        eigval(n)             = one
!
                    case( 4_i4b )
!
!                       TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                        eigval(1_i4b)         = ulp
                        eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                        eigval(n)             = two
!
                    case( 5_i4b )
!
!                       TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
                    case( 6_i4b )
!
!                       TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                        err = real( n-1_i4b, stnd )
!
                        do l = 1_i4b, n
                            eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                        end do
!
                    case( 7_i4b )
!
!                       TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                        eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
                    case( 8_i4b )
!
!                       TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                        eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
                    case( 9_i4b )
!
!                       TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
                    case( 10_i4b )
!
!                       TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
                    case( 11_i4b )
!
!                       TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                        call random_number_( eigval(:n) )
!
                        eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
                    case( 12_i4b )
!
!                       TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
                    case( 13_i4b )
!
!                       TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + one
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 14_i4b )
!
!                       TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + one
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 15_i4b )
!
!                       TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + one
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 16_i4b )
!
!                       TEST 16: CLUSTER OF EIGENVALUES AT 1 WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        eigval(1_i4b)   = one/conda
                        eigval(2_i4b:n) = one
!
                    case( 17_i4b )
!
!                       TEST 17: CLUSTER OF EIGENVALUES AT 1/conda WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        eigval(1_i4b)   = one
                        eigval(2_i4b:n) = one/conda
!
                    case( 18_i4b )
!
!                       TEST 18: ARITHMETIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        tmp = real( n-1_i4b, stnd )
                        err = one - one/conda
!
                        do l = 1_i4b, n
                            eigval(l) = one - ( real(l-1_i4b,stnd)/tmp )*err
                        end do
!
                    case( 19_i4b )
!
!                       TEST 19: GEOMETRIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        tmp = real( n-1_i4b, stnd )
!
                        do l = 1_i4b, n
                            eigval(l) = conda**( -real(l-1_i4b,stnd)/tmp )
                        end do
!
                    case( 20_i4b )
!
!                       TEST 20: RANDOM DISTRIBUTION OF EIGENVALUES IN THE INTERVAL 1/conda TO 1
!                                WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        call random_number_( beta(:n) )
!
                        tmp = log( one/conda )

                        beta(:n)   = tmp + beta(:n)*abs( tmp )
                        eigval(:n) = exp( beta(:n) )
!
                    case( 21_i4b )
!
!                       TEST 21: VERY SLOW DECAY OF EIGENVALUES.
!
                        conda = real( n - 1_i4b, stnd )
!
                        do l = 1_i4b, n
!
                            tmp = real( l-1_i4b, stnd )
!
                            eigval(l) = exp( -tmp/conda )
!        
                        end do
!
                    case( 22_i4b )
!
!                       TEST 22: SLOW DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = one/( tmp*tmp )
!        
                        end do
!
                    case( 23_i4b )
!
!                       TEST 23: FAST DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = exp( -tmp/seven )
!        
                        end do
!
                    case( 24_i4b )
!
!                       TEST 24: S-SHAPE DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
                        end do
!
                    case default
!
!                       DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                        call random_number_( eigval(:n) )
!
                        call random_number_( tmp )
!
                        if ( tmp>safmin ) then
                             eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                        end if
!
                end select
!
#ifdef _F2003
                if ( ieee_support_datatype( eigval ) ) then
                    if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
                end if
#endif
!
!               COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!
                singval(:n) = sqrt( eigval(:n) )
!
!               SORT THE SINGULAR VALUES.
!
                call singval_sort( sort, singval(:n) )
!
!               FORM A MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
                rightvect(:n,:nc) = transpose( rightvec(:nc,:n) )
!
                do i = 1_i4b, nc
!
                    rightvect(:n,i) = rightvect(:n,i)*singval(:n)
!
                end do
!
                a(:nr,:nc) = matmul( leftvec(:nr,:n), rightvect(:n,:nc) )
!
!               COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
!                norma = norm( a(:nr,:nc) )
                norma = norm( singval(:n) )
!
!               COMPUTE THE TRUE RELATIVE ERROR.
!
                relerr = sqrt(sum( (singval(nqb+1_i4b:n)/norma)**2 ))
!
!               COMPUTE APPROXIMATE PARTIAL QB OR QR FACTORIZATION OF RANDOM DATA MATRIX.
!
                comp_qr = .not.comp_qr
!
                if ( comp_qr ) then
!
                    pivoting = .not.pivoting
!
                    if ( pivoting ) then
!
                        min_norm = false
!
                        call rqb_cmp( a(:nr,:nc), q(:nr,:nqb), b(:nqb,:nc), niter=niter, ortho=ortho, ip=ip(:nc) )
!
                    else
!
!                       A COMPLETE ORTHOGONAL FACTORIZATION IS COMPUTED HERE, BUT WITHOUT PIVOTING.
!
                        min_norm = true
!
                        call rqb_cmp( a(:nr,:nc), q(:nr,:nqb), b(:nqb,:nc), niter=niter, ortho=ortho, tau=beta(:nqb) )
!
                    end if
!
                else
!
                    min_norm = false
!
                    call rqb_cmp( a(:nr,:nc), q(:nr,:nqb), b(:nqb,:nc), niter=niter, ortho=ortho, comp_qr=comp_qr )
!
                end if
!
!               COMPUTE THE ESTIMATED RELATIVE ERROR.
!
                if ( min_norm ) then
                    normb = norm( b(:nqb,:nqb) )
                else
                    normb = norm( b(:nqb,:nc) )
                end if
!
                tmp = one - (normb/norma)**2
                relerr2 = sqrt( max( tmp, zero ) )
!
!               COMPUTE ERROR BETWEEN THE TRUE AND ESTIMATED RELATIVE ERRORS.
!
                err = abs( relerr - relerr2 )
!
                m = m + 1_i4b
!
                if ( details .or. err>eps ) then
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q**(t)*q
!                   WHERE q IS THE Q MATRIX IN THE PARTIAL QB FACTORIZATION.
!
                    res(:nqb,:nqb) = matmul( transpose( q(:nr,:nqb) ), q(:nr,:nqb) )
!
                    do l = 1_i4b, nqb
                        res(l,l) = one - res(l,l)
                    end do
!
                    err1 = maxval( norm(res(:nqb,:nqb), dim=2_i4b) )/real(nr,stnd)
!
                end if
!
                if ( err<=eps ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',m,' is correct'
!
                        write (prtunit,*) 'nr= ', nr,' , nc= ', nc,' , case= ', j,          &
                                          ' , nqb= ', nqb, ' , niter= ', niter,             &
                                          ' , ortho= ', ortho,' , comp_qr= ', comp_qr,      &
                                          ' , pivoting= ',pivoting,' , min_norm= ',min_norm

!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q**(t)*q
!                       WHERE q IS THE Q MATRIX IN THE PARTIAL QB OR QR FACTORIZATION.
!
                        write (prtunit,*) 'Orthogonality of the computed q matrix                   = ', &
                                          err1/ulp, ' (must be less than :', fudge, ' )'
!
!                       PRINT RELATIVE ACCURACY OF THE QB OR QR APPROXIMATION IN FROBENIUS NORM.
!
                        write (prtunit,*) 'Relative accuracy of the computed QB or QR approximation = ', &
                                          relerr2, ' (must be of the order of :',relerr, ' )'
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',m,' is incorrect'
!
                    write (prtunit,*) '=> ',err,' > ',eps,' , nr= ', nr,' , nc= ', nc,' , case= ', j, &
                                      ' , nqb= ', nqb,' , niter= ', niter,' , ortho= ', ortho,        &
                                      ' , comp_qr= ', comp_qr,' , pivoting= ',pivoting,               &
                                      ' , min_norm= ',min_norm
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q**(t)*q
!                   WHERE q IS THE Q MATRIX IN THE PARTIAL QB OR QR FACTORIZATION.
!
                    write (prtunit,*) 'Orthogonality of the computed q matrix                   = ', &
                                      err1/ulp, ' (must be less than :', fudge, ' )'
!
!                   PRINT RELATIVE ACCURACY OF THE QB OR QR APPROXIMATION IN FROBENIUS NORM.
!
                    write (prtunit,*) 'Relative accuracy of the computed QB or QR approximation = ', &
                                      relerr2, ' (must be of the order of :',relerr, ' )'
!
                    write (prtunit,*)
!
                end if
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, leftvec, rightvec, rightvect, q, b, res, ip )
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, singval, diagr, beta )
!
    end do
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_rqb_cmp
! ===========================
!
end program test_rqb_cmp
