program test_rqlp_cmp
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine RQLP_CMP in module SVD_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for checking
!    the accuracy of subroutine RQLP_CMP.
!
! LATEST REVISION : 21/09/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, true, i4b, stnd, n1_def, n2_def, n3_def,      &
                         zero, half, one, two, five, seven, eight, ten,     &
                         c30, c100, safmin, rqlp_cmp, qr_cmp, ortho_gen_qr, &
                         norm, arth, svd_sort
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, nqlp=15_i4b
!   
    real(stnd), parameter  :: fudge=ten, def_conda=50000._stnd
!
    character(len=*), parameter :: name_proc='rqlp_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err4, eps, eps1, eps2, ulp, &
                                               tmp, tmp2, shift, tol, conda, norma, normr
    real(stnd), allocatable, dimension(:)   :: eigval, singval, diagr, beta
    real(stnd), allocatable, dimension(:,:) :: a, leftvec, rightvec, rightvect, q, p, r
!
    integer      :: iok
    integer(i4b) :: i, j, k, l, n, nr, nc, m, o, krank0, niter, niter_qrql, nqlp2, success
!   
    logical(lgl) :: details, ortho
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   CHOOSE TUNING PARAMETERS FOR THE RANDOMIZED QLP ALGORITHM.
!
!   SPECIFY IF ORTHONORMALISATION IS CARRIED OUT BETWEEN EACH SUBSPACE ITERATION.
!
    ortho = true
!
!   DETERMINE THE NUMBER OF SUBSPACE ITERATIONS TO BE PERFORMED IN THE ALGORITHM.
!
    niter = 2_i4b
!
!   DETERMINE THE NUMBER OF QR-QL ITERATIONS PERFORMED ON THE L FACTOR AFTER THE INITIAL
!   QLP FACTORIZATION FOR IMPROVING THE L-VALUES.
!
    niter_qrql = 60_i4b
!
!   eps1 IS THE RELATIVE ACCURACY OF THE QLP APPROXIMATION COMPARED TO THE OPTIMAL SVD APPROXIMATION.
!   eps2 IS THE RELATIVE ACCURACY OF THE L-VALUES AS ESTIMATES OF THE SINGULAR VALUES.
!
    eps1   = 0.04_stnd
    eps2   = 0.55_stnd
!
!   SET THE REQUIRED PRECISION OF THE RESULTS FOR THE ORTHOGONAL MATRICES Q AND P.
!
    ulp    = epsilon( ulp )
    eps    = fudge*ulp
!
!   SPECIFY EVENTUALLY A MULTIPLICATIVE SHIFT TO CREATE A GAP
!   IN THE SINGULAR VALUES DISTRIBUTION. shift*singval(1) WILL BE
!   ADDED TO HALF OF THE (LARGEST) SINGULAR VALUES.
!
    shift = 8.5_stnd
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigval(n), singval(n), diagr(n), beta(n), stat=iok )
!
        if ( iok/=0 ) then
!
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
!
            cycle
!
        end if
!
        do k = 1_i4b, 5_i4b
!
            select case( k )
!
                case( 1_i4b )
!
!                   TEST 1: QR DECOMPOSITION OF a(:nr,:nc) WITH nr=nc .
!
                    nr = n
                    nc = n
!
                case( 2_i4b )
!
!                   TEST 2: QR DECOMPOSITION OF a(:nr,:nc) WITH nr>=nc AND nr NOT MUCH LARGER THAN nc .
!
                    nr = n + 1_i4b
                    nc = n
!
                case( 3_i4b )
!
!                   TEST 3: QR DECOMPOSITION OF a(:nr,:nc) WITH nr>=nc AND nr MUCH LARGER THAN nc .
!
                    nr = n + n
                    nc = n
!
                case( 4_i4b )
!
!                   TEST 4: QR DECOMPOSITION  OF a(:nr,:nc) WITH nc>=nr AND nc NOT MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + 1_i4b
!
                case( 5_i4b )
!
!                   TEST 5: QR DECOMPOSITION OF a(:nr,:nc) WITH nc>=nr AND nc MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + n
!
            end select
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( a(nr,nc), r(nqlp,nqlp), q(nr,nqlp), p(nqlp,nc),  &
                      leftvec(nr,n), rightvec(nc,n), rightvect(n,nc),  &
                      stat=iok )
!
            if ( iok/=0 ) then
!
                if ( allocated(a) )         deallocate( a )
                if ( allocated(r) )         deallocate( r )
                if ( allocated(q) )         deallocate( q )
                if ( allocated(p) )         deallocate( p )
                if ( allocated(leftvec) )   deallocate( leftvec )
                if ( allocated(rightvec) )  deallocate( rightvec )
                if ( allocated(rightvect) ) deallocate( rightvect )
!
                cycle
!
            end if
!
!           GENERATE A RANDOM COEFFICIENT MATRIX leftvec(:nr,:n) .
!
            call random_number( leftvec(:nr,:n) )
!
            leftvec(:nr,:n) = leftvec(:nr,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                leftvec(:nr,:n) = leftvec(:nr,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( leftvec(:nr,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( leftvec(:nr,:n), beta(:n) )
!
!           GENERATE A RANDOM COEFFICIENT MATRIX rightvec(:nc,:n) .
!
            call random_number( rightvec(:nc,:n) )
!
            rightvec(:nc,:n) = rightvec(:nc,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                rightvec(:nc,:n) = rightvec(:nc,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( rightvec(:nc,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( rightvec(:nc,:n), beta(:n) )
!
            do j = 1_i4b, 25_i4b
!
                select case( j )
!
                    case( 1_i4b )
!
!                       TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b)   = one
                        eigval(2_i4b:n) = ulp
!
                    case( 2_i4b )
!
!                       TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = one
                        eigval(n)             = ulp
!
                    case( 3_i4b )
!
!                       TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                        eigval(n)             = one
!
                    case( 4_i4b )
!
!                       TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                        eigval(1_i4b)         = ulp
                        eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                        eigval(n)             = two
!
                    case( 5_i4b )
!
!                       TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
                    case( 6_i4b )
!
!                       TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                        tmp = real( n-1_i4b, stnd )
!
                        do l = 1_i4b, n
                            eigval(l) = ulp**(real(l-1_i4b,stnd)/tmp)
                        end do
!
                    case( 7_i4b )
!
!                       TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                        eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
                    case( 8_i4b )
!
!                       TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                        eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
                    case( 9_i4b )
!
!                       TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
                    case( 10_i4b )
!
!                       TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
                    case( 11_i4b )
!
!                       TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                        call random_number( eigval(:n) )
!
                        eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
                    case( 12_i4b )
!
!                       TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
                    case( 13_i4b )
!
!                       TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + one
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 14_i4b )
!
!                       TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + one
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 15_i4b )
!
!                       TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + one
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 16_i4b )
!
!                       TEST 16: CLUSTER OF EIGENVALUES AT 1 WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        eigval(1_i4b)   = one/conda
                        eigval(2_i4b:n) = one
!
                    case( 17_i4b )
!
!                       TEST 17: CLUSTER OF EIGENVALUES AT 1/conda WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        eigval(1_i4b)   = one
                        eigval(2_i4b:n) = one/conda
!
                    case( 18_i4b )
!
!                       TEST 18: ARITHMETIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        tmp  = real( n-1_i4b, stnd )
                        tmp2 = one - one/conda
!
                        do l = 1_i4b, n
                            eigval(l) = one - ( real(l-1_i4b,stnd)/tmp )*tmp2
                        end do
!
                    case( 19_i4b )
!
!                       TEST 19: GEOMETRIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        tmp = real( n-1_i4b, stnd )
!
                        do l = 1_i4b, n
                            eigval(l) = conda**( -real(l-1_i4b,stnd)/tmp )
                        end do
!
                    case( 20_i4b )
!
!                       TEST 20: RANDOM DISTRIBUTION OF EIGENVALUES IN THE INTERVAL 1/conda TO 1
!                                WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        call random_number( beta(:n) )
!
                        tmp = log( one/conda )

                        beta(:n) = tmp + beta(:n)*abs( tmp )
                        eigval(:n) = exp( beta(:n) )
!
                    case( 21_i4b )
!
!                       TEST 21: VERY SLOW DECAY OF EIGENVALUES.
!
                        conda = real( n - 1_i4b, stnd )
!
                        do l = 1_i4b, n
!
                            tmp = real( l-1_i4b, stnd )
!
                            eigval(l) = exp( -tmp/conda )
!        
                        end do
!
                    case( 22_i4b )
!
!                       TEST 22: SLOW DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = one/( tmp*tmp )
!        
                        end do
!
                    case( 23_i4b )
!
!                       TEST 23: FAST DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = exp( -tmp/seven )
!        
                        end do
!
                    case( 24_i4b )
!
!                       TEST 24: S-SHAPE DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
                        end do
!
                    case default
!
!                       DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                        call random_number( eigval(:n) )
!
                        call random_number( tmp )
!
                        if ( tmp>safmin ) then
                             eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                        end if
!
                end select
!
#ifdef _F2003
                if ( ieee_support_datatype( eigval ) ) then
                    if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
                end if
#endif
!
                m = m + 1_i4b
!
!               COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!
                singval(:n) = sqrt( eigval(:n) )
!
                tol = real( n, stnd)*sqrt( ulp )
!
                where( singval(:n)<=tol ) singval(:n) = zero
!
!               SORT THE SINGULAR VALUES IN DECREASING ORDER
!
                call svd_sort( 'd', singval(:n) )
!
!               CREATE A GAP IN THE SINGULAR VALUES DISTRIBUTION.
!
                singval(:n/2) = singval(:n/2) + shift*singval(1)
!
!               DETERMINE THE RANK OF THE GENERATED MATRIX.
!
                krank0 = count( singval(:n)/=zero )
!
!               FORM THE MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
                rightvect(:n,:nc) = transpose( rightvec(:nc,:n) )
!
                do i = 1_i4b, nc
!
                    rightvect(:n,i) = rightvect(:n,i)*singval(:n)
!
                end do
!
                a(:nr,:nc) = matmul( leftvec(:nr,:n), rightvect(:n,:nc) )
!
!               COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
                norma = norm( singval(:krank0) )
!
!               COMPUTE PARTIAL RANDOMIZED QLP DECOMPOSITION OF RANDOM DATA MATRIX.
!
                nqlp2 = min( nqlp, krank0 )
!
                call rqlp_cmp( a(:nr,:nc), lmat=r(:nqlp2,:nqlp2), qmat=q(:nr,:nqlp2), pmat=p(:nqlp2,:nc),  &
                               niter=niter, ortho=ortho, niter_qrql=niter_qrql )
!
!               CHECK THE RELATIVE ACCURACY OF THE QLP APPROXIMATION COMPARED TO THE OPTIMAL SVD APPROXIMATION
!               OF RANK nqlp2.
!
                normr = norm( r(:nqlp2,:nqlp2) )
                tmp   = norm( singval(:nqlp2) )

                err1 = abs( tmp  - normr )/norma
!
!               CHECK THE RELATIVE ACCURACY OF THE L-VALUES.
!
                do l = 1_i4b, nqlp2
                    diagr(l) = abs( r(l,l) )
                end do
!
                err2 = maxval( abs( diagr(:nqlp2) - singval(:nqlp2) )/singval(:nqlp2) )
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q**(t)*q
!               WHERE q IS THE ORTHOGONAL MATRIX Q FROM THE QLP DECOMPOSITION.
!
                r(:nqlp2,:nqlp2) = matmul( transpose( q(:nr,:nqlp2) ), q(:nr,:nqlp2) )
!
                do l = 1_i4b, nqlp2
                    r(l,l) = one - r(l,l)
                end do
!
                err3 = maxval( norm(r(:nqlp2,:nqlp2), dim=2_i4b) )/real(n,stnd)
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - p*p**(t)
!               WHERE p IS THE ORTHOGONAL MATRIX P FROM THE QLP DECOMPOSITION.
!
                r(:nqlp2,:nqlp2) = matmul( p(:nqlp2,:nc), transpose( p(:nqlp2,:nc) ) )
!
                do l = 1_i4b, nqlp2
                    r(l,l) = one - r(l,l)
                end do
!
                err4 = maxval( norm(r(:nqlp2,:nqlp2), dim=2_i4b) )/real(n,stnd)
!
                if ( max( err3, err4 )<=eps .and. err1 < eps1 .and. err2 < eps2 ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',m,' is correct'
!
                        write (prtunit,*) 'nr= ', nr,' , nc= ', nc,' , case= ', j,     &
                                          ' , krank0= ', krank0,' , nqlp = ',nqlp,     &
                                          ' , niter = ',niter,' ,ortho= ', ortho,      &
                                          ' , niter_qrql = ',niter_qrql
!
!                       CHECK THE RELATIVE ACCURACY OF THE QLP APPROXIMATION COMPARED TO THE OPTIMAL
!                       SVD APPROXIMATION OF a.
!
                        write (prtunit,*) 'Relative Accuracy of the QLP approximation = ', &
                                          err1, ' (must be less than :', eps1, ' )'
!
!                       CHECK THE RELATIVE ACCURACY OF THE L-VALUES AS ESTIMATES OF THE SINGULAR VALUES
!                       OF a WHERE THE L-VALUES ARE THE DIAGONAL ELEMENTS OF THE L FACTOR FROM THE QLP
!                       DECOMPOSITION OF a.
!
                        write (prtunit,*) 'Relative Accuracy of the L-values          = ', &
                                          err2, ' (must be less than :', eps2, ' )'
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q**(t)*q
!                       WHERE q IS THE ORTHOGONAL MATRIX Q FROM THE QLP DECOMPOSITION OF a .
!
                        write (prtunit,*) 'Orthogonality of the computed Q factor     = ', &
                                          err3/ulp, ' (must be less than :', fudge, ' )'
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - p*p**(t)
!                       WHERE p IS THE ORTHOGONAL MATRIX P FROM THE QLP DECOMPOSITION OF a .
!
                        write (prtunit,*) 'Orthogonality of the computed P factor     = ', &
                                          err4/ulp, ' (must be less than :', fudge, ' )'
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',m,' is incorrect'
!
                    write (prtunit,*) 'nr= ', nr,' , nc= ', nc,' , case= ', j,     &
                                      ' , krank0= ', krank0,' , nqlp = ',nqlp,     &
                                      ' , niter = ',niter,' ,ortho= ', ortho,      &
                                      ' , niter_qrql = ',niter_qrql
!
!                   CHECK THE RELATIVE ACCURACY OF THE QLP APPROXIMATION COMPARED TO THE OPTIMAL
!                   SVD APPROXIMATION OF a.
!
                    write (prtunit,*) 'Relative Accuracy of the QLP approximation = ', &
                                      err1, ' (must be less than :', eps1, ' )'
!
!                   CHECK THE RELATIVE ACCURACY OF THE L-VALUES AS ESTIMATES OF THE SINGULAR VALUES
!                   OF a WHERE THE L-VALUES ARE THE DIAGONAL ELEMENTS OF THE L FACTOR FROM THE QLP
!                   DECOMPOSITION OF a.
!
                    write (prtunit,*) 'Relative Accuracy of the L-values          = ', &
                                      err2, ' (must be less than :', eps2, ' )'
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q**(t)*q
!                   WHERE q IS THE ORTHOGONAL MATRIX Q FROM THE QLP DECOMPOSITION OF a .
!
                    write (prtunit,*) 'Orthogonality of the computed Q factor     = ', &
                                      err3/ulp, ' (must be less than :', fudge, ' )'
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - p*p**(t)
!                   WHERE p IS THE ORTHOGONAL MATRIX P FROM THE QLP DECOMPOSITION OF a .
!
                    write (prtunit,*) 'Orthogonality of the computed P factor     = ', &
                                      err4/ulp, ' (must be less than :', fudge, ' )'
!
                    write (prtunit,*)
!
                end if
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, leftvec, rightvec, rightvect, r, q, p )
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, singval, diagr, beta )
!
    end do
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_rqlp_cmp
! ============================
!
end program test_rqlp_cmp
